# SID: 1.31
#
# DATEMODIFIED: 01/10/17 14:35:21
#
# ASITECListenerSvc.ksh
#
# Goal: Configure MVS Host and\or TEC event enabler using the ASITECListenerSVC.ksh tool
#
# usage: see ASITECListenerSvc.ksh -?

num_params=$#;

# default values
bInstall=0
bRemove=0
bDisable=0
bEnable=0
TMP_LOCALDIR=$(regfind -p "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0" -n "InstallDirectory" 2>nul: | grep InstallDirectory | sed 's/[ ]*InstallDirectory = //' | sed 's/\\/\\\\\\\\/g' | sed 's/\\\\\\\\$//g')
local_host="$(print "$COMPUTERNAME" | awk '{print toupper($0);}')";
mvs_host="$local_host"

usage   ()
{
  print ''
  print 'Usage: '$0' -M<mvs_name> | -T<event_enabler> [-i|-r] [options...]'
  print ''
  print '  Required Parameter(s):'
  print '  -M<mvs_name>		-	Name of MVS Instance'
  print ''
  print '  -T<event_enabler>	-	The TEC client name. Specify hostname or IP address'
  print ''
  print '	-i		-	Install MVS instance and\or TEC Event Enabler'
  print '	-r		-	Remove MVS instance and\or TEC Event Enabler'
  print ''
  print '  -n<netid>		-	VTAM Netid for the MVS Instance.'
  print '				The Netid can be obtained by the Network System Programmer.'
  print '				Required only with -M.'
  print ''
  print '  Optional Parameter(s):'
  print '  -A<local_dir>		-	Local Directory Name.  Default=['$TMP_LOCALDIR']'
  print ''
  print '  -p<port_number>	-	The port number for the TEC client. If the port is not'
  print '				not specified, port 4030 is used. The port number must match the'
  print '				port number used by the Tivoli BSM Event Enablement Service.'	
  print '' 	
  print '  -l<log_level>		-	log level values used by TBSM [0 - 5]'
  print ''
  print '  -c<cell_size>		-	The maximum size of a TBSM TEC entry element placed on the'
  print '				TBSM queue. The default is 1024.'
  print ''	
  print '  -x<max_entries>	-	The maximum number of TBSM TEC entries that the queue can'
  print '				contain at a given time. The default is 32000.'
  print ''
  print '  -o<retry_count> 	-	The number of connection attempts to the TEC [EE] before'
  print '				terminating. When 0 is used, the service will retry forever.'
  print '				The default is 4294967295 which is the largest DWORD value.'
  print ''
  print '  -w<retry_wait>	-	The time interval between connection attempts in seconds'
  print '				[ 1 - 86400 ]. The default is 30.'
  print ''
  print '  -d			-	Disable TEC client for -T<event_enabler>'
  print '  -e			-	Enable TEC client for -T<event_enabler>'
  print ''
  print '  -u<debug_collect>	-	Debug Collect values [ 0 | 1 ] for specified TEC. -T is required'
  print '				with -u. The default is 0'
  print ''
  print '  Examples:'
  print ''
  print '  1) Install MVS Instance for an OS named E003'
  print '		'$0' -ME003 -nUSASI01.NJ1TM390 -i'
  print ''
  print '  2) Install MVS Instance for an OS named E003 and setup TEC Event Enabler using port 4030'
  print '		'$0' -ME003 -nUSASI01.NJ1TM390 -TTEC7 -p4030 -i'
  print ''
  print '  3) Delete MVS Instance for an OS named CPUA'
  print '		'$0' -MCPUA -r'
  print ''
  print '  4) Change log level from 2 to 0 to enable highest level of diagnostics'
  print '		'$0' -l0'
  print ''
}

while getopts "M:T:A:n:p:l:c:x:o:w:u:irde?" c; do
  case $c in
  M)  MVSNAME=$(echo $OPTARG | awk '{print toupper($1);}');;
  T)  EVENTENABLER=$OPTARG;;
  A)  LOCALDIR=$OPTARG;;
  i)  bInstall=1;;
  r)  bRemove=1;;
  n)  netid=$OPTARG;;
  p)  port=$OPTARG;;
  l)  loglevel=$OPTARG;;
  c)  cellsize=$OPTARG;;
  x)  maxentries=$OPTARG;;
  o)  retrycount=$OPTARG;;
  w)  retrywait=$OPTARG;;
  d)  bDisable=1;;
  e)  bEnable=1;;
  u)  bDCollect=$OPTARG;;
  \?) usage; exit 1;;
  *)  print >&2 Error with option $OPTOPT; usage; exit 1;;
	esac
done
shift $OPTIND-1

check_nt_res_kit()
{
  filename="${1?"ERROR: check_nt_res_kit() missing required field <filename>"}"
    
  # check for utility
  rc=`which "$filename"`; rc=$?
  if [ $rc -ne 0 ]; then;
  echo "ERROR: Could not find "$filename".  Please install Windows NT Resource Kit Supplement 3" >&2
  echo "       or later and ensure "$filename" is in the path." >&2
    exit 1
  fi
}

check_for_required_parameters()
{
	if [ -z "$MVSNAME" -a -z "EVENTENABLER" ] ; then
		echo "ERROR: Missing required parameters -M<mvs_name> | -T<event_enabler>" >&2
		usage >&2
		exit 2
	fi

	if [ -n "$MVSNAME" ] ; then
		if [ $bInstall -eq 0 -a $bRemove -eq 0 ] ; then
			echo "ERROR: Either option must be specified. [ -i | -r ]" >&2
			usage >&2
			exit 3
		fi
		if [ $bInstall -eq 1 ] ; then
			if [ -z "$netid" ] ; then
				echo "ERROR: Netid [-n] is required when using -M with the -i option" >&2
				usage >&2
				exit 4
			fi
		fi
	fi

	if [ -n "$EVENTENABLER" ] ; then
		if [ $bInstall -eq 0 -a $bRemove -eq 0 -a $bDisable -eq 0 -a $bEnable -eq 0 ] ; then
		    if [ -n $EVENTENABLER -a -n bDCollect ] ; then
			   return
			fi
		    echo "ERROR: Either of these options must be specified. [ -i | -r | -d | -e ]" >&2
			usage >&2
			exit 5
		fi
		if [ $bInstall -eq 1 ] ; then
			if [ -z "$port" ] ; then
				port=4030
			fi
		fi
	fi

	if [ $num_params -lt 1 ]; then
      echo "ERROR: Missing required parameter(s)" >&2
      usage >&2
      exit 7
    fi

	# if local directory not specified, look in the registry of host <mvs_host>
	if [ -z "$LOCALDIR" ]; then;
		LOCALDIR=$(REGFIND -m '\\'"$mvs_host" -p "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0" -n "InstallDirectory" 2>nul: | grep InstallDirectory | sed 's/[ ]*InstallDirectory = //' | sed 's/\\/\\\\\\\\/g' | sed 's/\\\\\\\\$//g')
	fi
  
	# get rid of potential extra backslash ('\')
	LOCALDIR=$(print "$LOCALDIR" | sed 's/\\\\/\\/g')
	
	check_nt_res_kit REGFIND.EXE
	check_nt_res_kit REGINI.EXE
    check_nt_res_kit REG.EXE
}

Install_MVS_HOST()
{
	NETIDstr=$(regfind -p "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Instances\\$MVSNAME" -n "Netid" 2>nul: | grep Netid) 
	if [ -n "$NETIDstr" ] ; then
		echo "Registry key for MVS Host ("$MVSNAME") is already defined"
		return
	fi

	rm -f "$0.reg" 2> nul:

	if [ -z "$LOCALDIR" ] ; then
	   LOCALDIR=$TMP_LOCALDIR
	fi
	echo "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Instances
        $MVSNAME
			Netid	  = "${netid}"
			QueuePath = "${LOCALDIR}\\Data\\Queues\\$MVSNAME.que" 
	" >> "$0.reg" ;
	echo "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings
		Log
			LogHeading	= "%Y/%m/%d %H:%M:%S\|%t\|%O\|%o\|TECLI\|%i\|%F\|%L\|"
			Prefix		= "TECLI"
			LogLevel	= REG_DWORD 2
		Queue
			CellSize	= REG_DWORD  1024		
			MaxEntries	= REG_DWORD	 32000
	" >> "$0.reg" ;

	# apply the generated file
	if [ -f "$0.reg" ]; then;
		regini -m \\\\$COMPUTERNAME "$0.reg" > nul:
		rm -f "$0.reg" 2> nul:
	fi

	echo "MVS instance ("$MVSNAME") added to registry"
}

Remove_MVS_HOST()
{
	rc=$(echo 'y' | REG DEL 'HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Instances'\\"${MVSNAME}" 2>&1)
	if $(print "$rc" | grep -q "Permanentely delete"); then ;
		print "Successfully deleted MVS Host name ("$MVSNAME") from registry"
	elif $(print "$rc" | grep -q "The system was unable to find the specified registry key"); then;
		print "MVS Host name ("$MVSNAME") is not defined in registry"
	else
		print "ERROR: $rc"
	fi
}

Install_TEC_Event_Enabler()
{
	EEstr=$(regfind -p "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\EventEnablers\\$EVENTENABLER" -n "Enabled" 2>nul: | grep Enabled) 
	if [ -n "$EEstr" ] ; then
		echo "Registry key for TEC Event Enabler ("$EVENTENABLER") is already defined"
		return
	fi

	rm -f "$0.reg" 2> nul:

    echo "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\EventEnablers
	    $EVENTENABLER
			DebugCollect	=	REG_DWORD   0
			Enabled			=	REG_DWORD   1
			Port			=	REG_DWORD   "${port}"
	" >> "$0.reg" ;
	echo "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings
		Retry
			NumMaxRetries	=   REG_DWORD	0xffffffff
			RetryWait		=	REG_DWORD	30000
	" >> "$0.reg" ;

	# apply the generated file
	if [ -f "$0.reg" ]; then;
		regini -m \\\\$COMPUTERNAME "$0.reg" > nul:
		rm -f "$0.reg" 2> nul:
	fi

	echo "TEC Event Enabler ("$EVENTENABLER") added to registry"
}

Remove_TEC_Event_Enabler()
{
	rc=$(echo 'y' | REG DEL 'HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\EventEnablers'\\"${EVENTENABLER}" 2>&1)
	if $(print "$rc" | grep -q "Permanentely delete"); then ;
		print "Successfully deleted TEC Event Enabler ("$EVENTENABLER") from registry"
	elif $(print "$rc" | grep -q "The system was unable to find the specified registry key"); then;
		print "TEC Event Enabler ("$EVENTENABLER") is not defined in registry"
	else
		print "ERROR: $rc"
	fi
}

Update_Log_Level()
{
	LGVLstr=$(regfind -p "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\Log" -n "LogLevel" 2>nul: | grep LogLevel) 
	if [ -z "$LGVLstr" ] ; then
		echo "Cannot update Registry key. The ASITECListenerSvc registry key does not exist"
		return
	fi

	rm -f "$0.reg" 2> nul:

    echo "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\Log
        LogLevel	= REG_DWORD "${loglevel}" 
	" >> "$0.reg" ;
	
	# apply the generated file
	if [ -f "$0.reg" ]; then;
		regini -m \\\\$COMPUTERNAME "$0.reg" > nul:
		rm -f "$0.reg" 2> nul:
	fi

	echo "LogLevel ("$loglevel") updated for ASITECListenerSvc"
}

Update_Cell_Size()
{
	CSizestr=$(regfind -p "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\Queue" -n "CellSize" 2>nul: | grep CellSize) 
	if [ -z "$CSizestr" ] ; then
		echo "Cannot update Registry key. The ASITECListenerSvc Queue key does not exist"
		return
	fi

	rm -f "$0.reg" 2> nul:

	echo "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\Queue
        CellSize = REG_DWORD "${cellsize}" 
	" >> "$0.reg" ;
	
	# apply the generated file
	if [ -f "$0.reg" ]; then;
		regini -m \\\\$COMPUTERNAME "$0.reg" > nul:
		rm -f "$0.reg" 2> nul:
	fi

	echo "CellSize ("$cellsize") updated for ASITECListenerSvc Queue"
}

Update_Max_Entries()
{
	MaxEntstr=$(regfind -p "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\Queue" -n "MaxEntries" 2>nul: | grep MaxEntries) 
	if [ -z "$MaxEntstr" ] ; then
		echo "Cannot update Registry key. The ASITECListenerSvc Queue key does not exist"
		return
	fi

	rm -f "$0.reg" 2> nul:

	echo "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\Queue
        MaxEntries = REG_DWORD "${maxentries}" 
	" >> "$0.reg" ;
	
	# apply the generated file
	if [ -f "$0.reg" ]; then;
		regini -m \\\\$COMPUTERNAME "$0.reg" > nul:
		rm -f "$0.reg" 2> nul:
	fi

	echo "MaxEntries ("$maxentries") updated for ASITECListenerSvc Queue"
}

Update_Retry_Count()
{
	Retrycntstr=$(regfind -p "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\Retry" -n "NumMaxRetries" 2>nul: | grep NumMaxRetries) 
	if [ -z "$Retrycntstr" ] ; then
		echo "Cannot update Registry key. The ASITECListenerSvc Retry key does not exist"
		return
	fi

	rm -f "$0.reg" 2> nul:

	echo "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\Retry
        NumMaxRetries = REG_DWORD "${retrycount}" 
	" >> "$0.reg" ;
	
	# apply the generated file
	if [ -f "$0.reg" ]; then;
		regini -m \\\\$COMPUTERNAME "$0.reg" > nul:
		rm -f "$0.reg" 2> nul:
	fi

	echo "NumMaxRetries ("$retrycount") updated for ASITECListenerSvc Retry count"
}

Update_Retry_Wait()
{
	Retrywaitstr=$(regfind -p "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\Retry" -n "RetryWait" 2>nul: | grep RetryWait) 
	if [ -z "$Retrywaitstr" ] ; then
		echo "Cannot update Registry key. The ASITECListenerSvc Retry key does not exist"
		return
	fi

	rm -f "$0.reg" 2> nul:

	echo "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\Retry
        RetryWait = REG_DWORD  `expr "$retrywait" \* 1000`
	" >> "$0.reg" ;
	
	# apply the generated file
	if [ -f "$0.reg" ]; then;
		regini -m \\\\$COMPUTERNAME "$0.reg" > nul:
		rm -f "$0.reg" 2> nul:
	fi

	echo "RetryWait ("$retrywait") updated for ASITECListenerSvc Retry interval (seconds)"
}

Disable_EE()
{
	EEstr=$(regfind -p "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\EventEnablers\\$EVENTENABLER" -n "Enabled" 2>nul: | grep Enabled) 
	if [ -z "$EEstr" ] ; then
		echo "Registry key for Event Enabler ("$EVENTENABLER") is not defined"
		return
	fi

	rm -f "$0.reg" 2> nul:

    echo "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\EventEnablers
		$EVENTENABLER
			Enabled		= REG_DWORD 0 
	" >> "$0.reg" ;
	
	# apply the generated file
	if [ -f "$0.reg" ]; then;
		regini -m \\\\$COMPUTERNAME "$0.reg" > nul:
		rm -f "$0.reg" 2> nul:
	fi

	echo "Event Enabler ("$EVENTENABLER") is set for disabled"
}

Enable_EE()
{
	EEstr=$(regfind -p "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\EventEnablers\\$EVENTENABLER" -n "Enabled" 2>nul: | grep Enabled) 
	if [ -z "$EEstr" ] ; then
		echo "Registry key for Event Enabler ("$EVENTENABLER") is not defined"
		return
	fi

	rm -f "$0.reg" 2> nul:

    echo "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\EventEnablers
		$EVENTENABLER
			Enabled		= REG_DWORD 1 
	" >> "$0.reg" ;
	
	# apply the generated file
	if [ -f "$0.reg" ]; then;
		regini -m \\\\$COMPUTERNAME "$0.reg" > nul:
		rm -f "$0.reg" 2> nul:
	fi

	echo "Event Enabler ("$EVENTENABLER") is set for enabled"
}

create_service()	
{
        svc_name="${1?"ERROR: create_service() missing required param <svc_name>"}"
	
        rc=$(sc query $svc_name | awk ' BEGIN { result = 0 } END {print result} /SERVICE_NAME/ { result = 1 } // { next } ')
        if [ $rc -eq 0 ]; then;
		   echo 'Creating service ['$svc_name']...'
		   if [ -z "$LOCALDIR" ] ; then
	          LOCALDIR=$TMP_LOCALDIR
		   fi
		   "$LOCALDIR"'\\bin\\'"$svc_name" '/service'
		else
           echo '['$svc_name'] service already installed'
        fi
}

Update_Debug_Collect()
{
    DCstr=$(regfind -p "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\EventEnablers""\\$EVENTENABLER" -n "DebugCollect" 2>nul: | grep DebugCollect) 
	if [ -n "$DCstr" ] ; then
		rm -f "$0.reg" 2> nul:

		echo "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0\Components\ASITECListenerSvc\Settings\EventEnablers
		   $EVENTENABLER
		      DebugCollect =  REG_DWORD $bDCollect 
		" >> "$0.reg" ;
	
		# apply the generated file
		if [ -f "$0.reg" ]; then
			regini.exe -m \\\\$COMPUTERNAME "$0.reg" > nul:
			rm -f "$0.reg" 2> nul:
		fi

		echo "DebugCollect ("$bDCollect") updated for TEC ["$EVENTENABLER"]"
	else
	    echo "TEC ["$EVENTENABLER"] is not installed"
	fi
}

main()
{
	check_for_required_parameters

	# install or remove an MVS instance
	if	[ -n "$MVSNAME" -a $bInstall -eq 1 ] ; then
		Install_MVS_HOST
	fi
	if	[ -n "$MVSNAME" -a $bRemove -eq 1 ] ; then
		Remove_MVS_HOST
	fi

	# install or remove TEC Event Enabler
	if	[ -n "$EVENTENABLER" -a $bInstall -eq 1 ] ; then
		Install_TEC_Event_Enabler
	fi		
	if	[ -n "$EVENTENABLER" -a $bRemove -eq 1 ] ; then
		Remove_TEC_Event_Enabler
	fi

	# create service for all install (-i) requests
	if	[ $bInstall -eq 1 ] ; then
		create_service "ASITECListenerSvc"
	fi		
	
	# process optional operands

	# change loglevel setting
	if	[ -n "$loglevel" ] ; then
		if [ "$loglevel" -ge 0 -a "$loglevel" -le 5 ] ; then 
			Update_Log_Level
		else
			echo "ERROR: Invalid loglevel option" >&2
			usage >&2
			exit 100
		fi
	fi

	# Request override Queue parameters
	if [ -n "$cellsize" ] ; then
		if [ "$cellsize" -gt 0 ] ; then
			Update_Cell_Size
		else
			echo "ERROR: Invalid CellSize option" >&2
			usage >&2
			exit 101
		fi
	fi
	if [ -n "$maxentries" ] ; then
		if [ "$maxentries" -gt 0 ] ; then
			Update_Max_Entries
		else
			echo "ERROR: Invalid MaxEntries option" >&2
			usage >&2
			exit 102
		fi
	fi

	# Request override retry parameters
	if [ -n "$retrycount" ] ; then
		if [ "$retrycount" -gt 0 ] ; then 
			Update_Retry_Count
		else 
			echo "ERROR: Invalid RetryCount option" >&2
			usage >&2
			exit 103
		fi
	fi
	if [ -n "$retrywait" ] ; then
		if [ "$retrywait" -gt 0 -a "$retrywait" -le 86400 ] ; then
			Update_Retry_Wait
		else 
			echo "ERROR: Invalid RetryWait option" >&2
			usage >&2
			exit 104
		fi
	fi

	# enable or disable TEC
	if	[ $bDisable -eq 1 -a $bRemove -eq 0 ] ; then
		if [ -n "$EVENTENABLER" ] ; then
			Disable_EE
		fi
	fi
	if	[ $bEnable -eq 1 -a $bRemove -eq 0 ] ; then
		if [ -n "$EVENTENABLER" ] ; then
			Enable_EE
		fi
	fi

	# Turn on | off the Debug_Collect Option
	if	[ -n "$bDCollect" ] ; then
		if [ "$bDCollect" != 0 -a "$bDCollect" != 1 ] ; then
		   echo "ERROR: Invalid Debug_Collect [-u] option" >&2
		   usage >&2
		   exit 105
		fi 
		if [ -z "$EVENTENABLER" ] ; then
		   echo "ERROR: TEC [-T] is required with -u option" >&2
		   usage >&2
		   exit 106
	    fi 
		Update_Debug_Collect
	fi
}

main

























