# archive/scripts/opcparsedailyplan.ksh   
# SID: 1.28
# Date Modified: 01/10/31 08:42:43

# OPCParseDailyPlan.ksh
#
# Parser for OPC Daily Plan
#

TM_LANG=${TM_LANG:-EN_US}

Usage()
{
  print "USAGE:"
  print "  OPCParseDailyPlan options... <daily-plan-file>"
  print "  The options are:"
  print "    -d<date_format>  Optional parameter that specifies an alternate date"
  print "                     format in the plan file. Must use MM for month, DD for"
  print "                     day of month, and YY or YYYY for year."
  print "    -l<language>     Optional parameter that specifies an alternate language"
  print "                     to use for keyword parsing. The default is $TM_LANG."
  print "    -ncb             Skip checking the input file for binary characters."
  print "    -nm              Skip the mapping of the loaded data."
  print "    -np              Skip the parsing of the data."
  print "    -nl              No load of the parsed data is performed."
  print "    -t               Produce trace messages for diagnostics (implies -v)."
  print "    -v               Run in verbose mode."
  print "    -E               Use a trusted database connection"
  print "    -U<user>         User name for the database login"
  print "    -P<password>     Password for the database login"
  print "    -S<server>       Database server"
  print "    -C<codepage>     Optional parameter to specify code page of daily plan file"
  print "    -e               Central server: Use a trusted database connection"
  print "    -u<user>         Central server: User name for the database login"
  print "    -p<password>     Central server: Password for the database login"
  print "    -s<server>       Central server: Database server"
  print ""
  print "  <daily-plan-file> is the name of file containing an OPC daily plan report"
  print "  to be processed"
  print ""
  print "  Normal processing flow:"
  print "    1. daily plan file is parsed. Produces the files opc_application.bcp,"
  print "       opc_operation.bcp, and opc_predecessor.bcp. This step is skipped"
  print "       when the -np option is specified."
  print "    2. bcp files are loaded into the database. This step is skipped when"
  print "       the -nl option is specified."
  print "    3. the loaded data is mapped to objects in the TM OS/390 Object"
  print "       database. This step is skipped when the -nm option is specified."
  print ""
}

typeset -i bCheckForBinary=1
typeset -i bNoLoadAll=0
typeset -i bNoLoadApplications=0
typeset -i bNoLoadOperations=0
typeset -i bNoLoadPredecessors=0
typeset -i bNoMap=0
typeset -i bNoParse=0
typeset -i bTrusted=0
typeset -i bTrace=0
typeset -i bVerbose=0
typeset -i bCentralTrusted=0

user_opt=
password_opt=-P
server_opt=

central_user=
central_password=
central_server=

codepage_opt=

#
# main
#

# Process command line
while getopts hd:i:l:EU:S:P:eu:s:p:n:vC:t c; do
  if [ "X$c" = XC ]
  then
    codepage_opt="-C$OPTARG"
  elif [ "X$c" = Xi ]
  then
    Input=$OPTARG
  elif [ "X$c" = Xl ]
  then
    TM_LANG="$OPTARG"
  elif [ "X$c" = Xd ]
  then
    DATE_FORMAT="$OPTARG"
  elif [ "X$c" = XE ]
  then
    bTrusted=1
  elif [ "X$c" = XU ]
  then
    user_opt="-U$OPTARG"
  elif [ "X$c" = XP ]
  then
    password_opt="-P$OPTARG"
  elif [ "X$c" = XS ]
  then
    server_opt="-S$OPTARG"
  elif [ "X$c" = Xe ]
  then
    bCentralTrusted=1
  elif [ "X$c" = Xu ]
  then
    central_user="$OPTARG"
  elif [ "X$c" = Xp ]
  then
    central_password="$OPTARG"
  elif [ "X$c" = Xs ]
  then
    central_server="$OPTARG"
  elif [ "X$c" = Xn ]
  then
    case $OPTARG in
      l)   bNoLoadAll=1;;
      la)  bNoLoadApplications=1;;
      lo)  bNoLoadOperations=1;;
      lp)  bNoLoadPredecessors=1;;
      p)  bNoParse=1;;
      m)  bNoMap=1;;
      cb) bCheckForBinary=0;;
    esac
  elif [ "X$c" = Xv ]
  then
    bVerbose=1
  elif [ "X$c" = Xt ]
  then
    bTrace=1; bVerbose=1;
  elif [ "X$c" = Xh -o "X$c" = X? ]
  then
    Usage; exit 0
  else
    Usage >&2; exit 1
  fi
done

shift $OPTIND-1
daily_plan_file=$1

if [ $bCheckForBinary = 1 ]; then
  # Scrub the binary data out of the daily plan file, as follows:
  #   ASCII   REPLACED WITH
  #    000         .
  #
  if [ $bVerbose = 1 ]; then
    echo "Checking for binary data in $daily_plan_file"
  fi
  cat $daily_plan_file | tr '\000' '.' > $daily_plan_file.$$
  err=$?
  if [ $err != 0 ]; then
    print >&2 "Aborting due to file conversion error"
    exit $err
  fi

  mv -f $daily_plan_file $daily_plan_file.save
  mv -f $daily_plan_file.$$ $daily_plan_file
fi

# Lookup the installation directory from the registry
INSTALLDIR=$(regfind -p "HKEY_LOCAL_MACHINE\SOFTWARE\Accessible Software, Inc.\Access1\1.0" -n "InstallDirectory" 2>nul: | grep InstallDirectory | sed 's/[ ]*InstallDirectory = //' | sed 's?\\?/?g')
if [ ! -d "$INSTALLDIR" ]; then
  print >&2 "Invalid installation directory: $INSTALLDIR"
  exit 1
fi

TM_LANG="${INSTALLDIR}/lang/$TM_LANG"

# Figure out the database options for ISQL
isqlopts="-n -w2048"
bcpopts="-F2 -b1000 $codepage_opt"
if [ $bTrusted = 1 ]; then
  isqlopts="$isqlopts -E $server_opt"
  bcpopts="$bcpopts -T $server_opt"
else
  isqlopts="$isqlopts $user_opt $password_opt $server_opt"
  bcpopts="$bcpopts $user_opt $password_opt $server_opt"
fi

map_schedules()
{
  print "Starting mapping"

  if [ "X$central_server" != X ]
  then
    centralOpts=", @centralServer=\"$central_server\""
    if [ $bCentralTrusted = 1 ]
    then
      centralOpts="$centralOpts, @centralTrusted=1"
    else
      centralOpts="$centralOpts, @centralUser=\"$central_user\", @centralPw=\"$central_password\""
    fi
  else
    centralOpts=""
  fi

  isql $isqlopts -Q"EXEC OPCLoad.._LoadOPCPlan BUSC, 1, \"$daily_plan_file\"$centralOpts"
}

if [ $bNoMap = 0 -a $bNoParse = 1 -a $bNoLoadAll = 1 ]; then
  map_schedules
  exit
fi

# Perform a first pass to determine the level of indentation for the
# application and operation specifications
indent_assignment=$(awk  -v bVerbose=$bVerbose -v bNoParse=$bNoParse -v bcpopts="$bcpopts" -v isqlopts="$isqlopts" -v keywordfile="$TM_LANG/OPC.KEY" 'BEGIN {
  while (getline < keywordfile) {
    n=split($0, k, "=");
    if (n==0) {
      print "BAD KEYWORD FILE \"" keywordfile "\"";
      exit 1;
    }
    kw=k[1];
    sub("[ ]*$", "", kw);
    kv=k[2]
    keyword[kw]=kv;
  }
  close(keywordfile);
}
$0 ~ keyword["APPLICATION_ID"] {
  expr=keyword["APPLICATION_ID"] ".*";
  sub(expr, "", $0);
  indent=$0;
  print "indent=\"" indent "\"";
  exit(0);
}' $daily_plan_file)

eval $indent_assignment

awk -v DATE_FORMAT="$DATE_FORMAT" -v indent="$indent" -v bTrace=$bTrace -v bVerbose=$bVerbose -v bNoParse=$bNoParse -v bcpopts="$bcpopts" -v isqlopts="$isqlopts" -v keywordfile="$TM_LANG/OPC.KEY" 'function ltrim(str)
{
  sub("^[ \t]*", "", str);
  return str;
}
function rtrim(str)
{
  sub("[ \t]*$", "", str);
  return str;
}
function convert_date(date)
{
  year=substr(date, year_pos, 2);
  month=substr(date, month_pos, 2);
  day=substr(date, day_pos, 2);
  return sprintf("%s%s%s", year, month, day);
}
function convert_time(time)
{
  if (bTrace)
    print "TRACE: Line " NR ": convert_time(" time ")"
  sub("[.]", "", time);
  time=rtrim(time);
  return time;
}
function canonical_datetime(date, time)
{
  if (length(date) == 6)
  {
    y=substr(date, 1, 2);
    if (y < 90)
      date="20" date;
    else
      date="19" date;
  }
  datetime = substr(date,1,4) "-" substr(date,5,2) "-" substr(date,7,2) " " substr(time,1,2) ":" substr(time,3,2)
  return datetime;
}
function parse_day_mod(val)
{
  if (bTrace)
    print "TRACE: Line " NR ": parse_day_mod(" val ")"
  if (val == "<")
    return -1;
  else if (val == ">")
    return +1;
  else
    return 0;
}
function exec_select(id_column, query)
{
  print query > "tmp.sql";
  close("tmp.sql");
  cmd=sprintf("isql %s -dOPCLoad -itmp.sql | awk '\''NR==1 {getline;getline;sub(\" *\",\"\");print $0;exit;} /NR>1/ {print $0>>\"OPCParseDailyPlan.log\";}'\''", isqlopts, id_column);
#print "cmd=" cmd;
  cmd | getline result;
#print "result=" result;
  close(cmd);
  if (result != "")
    sub(" *$", "", result);
  return result;
}
function OutputFileHeader()
{
  ;
}
function OutputFileFooter()
{
}
function OutputDailyPlan(start_date, start_time, end_date, end_time)
{
  # Convert date/time values to canonical format: YYYY-MM-DD hh:mm
  start_date=convert_date(start_date);
  start_time=convert_time(start_time);
  start_date=canonical_datetime(start_date, start_time);

  end_date=convert_date(end_date);
  end_time=convert_time(end_time);
  end_date = canonical_datetime(end_date, end_time);

  fileparam=FILENAME
#  gsub("'\''", "'\'\''", fileparam);
#print "fileparam=" fileparam;
  if (bVerbose)
    print "DAILY PLAN: \"" FILENAME "\"  START: \"" start_date "\"  END: \"" end_date"\"";
  query="asisp_reset_opc_daily_plan \"" fileparam "\", \"" start_date "\", \"" end_date"\"";
#print "query=" query
  plan_id=exec_select("id", query);
  if (plan_id == "")
  {
    print "Could not create opc_daily_plan record. Query was:";
    print query;
    exit(1);
  }

  if (bNoParse==1) {
    exit(0);
  }
  close("opc_operation.bcp");
  close("opc_application.bcp");
  close("opc_predecessor.bcp");
  system("rm -f opc_operation.bcp opc_application.bcp opc_predecessor.bcp");
  close("rm -f opc_operation.bcp opc_application.bcp opc_predecessor.bcp");
  print "plan_id|app_id|app_name|app_owner|start_day|start_day_mod|start_hour|stop_day|stop_day_mod|stop_hour|app_desc">>"opc_application.bcp";
  print "plan_id|app_id|app_name|op_id|op_name|job_name|start_day|start_day_mod|start_hour|duration|stop_day|stop_day_mod|stop_hour|op_desc|0|0|||0|0|0">>"opc_operation.bcp";
  print "plan_id|succ_op_id|pred_app_name|pred_op_name">>"opc_predecessor.bcp";
}
function OutputApplication(app_id, app_name, app_instance, app_desc, app_owner, priority, status, start_day, start_day_mod, start_hour, stop_day, stop_day_mod, stop_hour)
{
  if (bVerbose)
    print napps " " app_name " " start_day " " start_time;

  # Make sure that the output delimiter is not in the data
  delimiter="|"
  pattern_delimiter = "\\" delimiter
  gsub( pattern_delimiter, "", plan_id       )
  gsub( pattern_delimiter, "", app_id        )
  gsub( pattern_delimiter, "", app_name      )
  gsub( pattern_delimiter, "", app_owner     )
  gsub( pattern_delimiter, "", start_day     )
  gsub( pattern_delimiter, "", start_day_mod )
  gsub( pattern_delimiter, "", start_hour    )
  gsub( pattern_delimiter, "", stop_day      )
  gsub( pattern_delimiter, "", stop_day_mod  )
  gsub( pattern_delimiter, "", stop_hour     )
  gsub( pattern_delimiter, "", app_desc      )

  print plan_id delimiter app_id delimiter app_name delimiter app_owner delimiter start_day delimiter start_day_mod delimiter start_hour delimiter stop_day delimiter stop_day_mod delimiter stop_hour delimiter app_desc >> "opc_application.bcp";
  app_start_day=start_day;
  app_start_hour=input_time;
}
function OutputOperation(app_id, op_id, op_name, job_name, status, start_day, start_day_mod, start_hour, duration, stop_day, stop_day_mod, stop_hour, op_desc, instanceno)
{
  if (bVerbose)
    print "    " op_id " " app_name " " operation " " jobname ;

  # Make sure that the output delimiter is not in the data
  delimiter="|"
  pattern_delimiter = "\\" delimiter
  gsub( pattern_delimiter, "", plan_id       )
  gsub( pattern_delimiter, "", app_id        )
  gsub( pattern_delimiter, "", app_name      )
  gsub( pattern_delimiter, "", op_id         )
  gsub( pattern_delimiter, "", op_name       )
  gsub( pattern_delimiter, "", job_name      )
  gsub( pattern_delimiter, "", status        )
  gsub( pattern_delimiter, "", start_day     )
  gsub( pattern_delimiter, "", start_day_mod )
  gsub( pattern_delimiter, "", start_hour    )
  gsub( pattern_delimiter, "", duration      )
  gsub( pattern_delimiter, "", stop_day      )
  gsub( pattern_delimiter, "", stop_day_mod  )
  gsub( pattern_delimiter, "", stop_hour     )
  gsub( pattern_delimiter, "", op_desc       )
  gsub( pattern_delimiter, "", instanceno    )

  print plan_id delimiter app_id delimiter app_name delimiter op_id delimiter op_name delimiter job_name delimiter status delimiter start_day delimiter start_day_mod delimiter start_hour delimiter duration delimiter stop_day delimiter stop_day_mod delimiter stop_hour delimiter op_desc delimiter instanceno >> "opc_operation.bcp";
}
function OutputPredecessor(app_name, app_instance, operation, ext_app, predecessor, app_id, op_id)
{

  # Make sure that the output delimiter is not in the data
  delimiter="|"
  pattern_delimiter = "\\" delimiter
  gsub( pattern_delimiter, "", plan_id     )
  gsub( pattern_delimiter, "", op_id       )
  gsub( pattern_delimiter, "", ext_app     )
  gsub( pattern_delimiter, "", predecessor )

#   Note: Duplicate predecessor records are being eliminated here.  The
#   current design only allows for an operation to distinguish its
#   predecessor by application and operation name.  It is possible that
#   an operation might have predecessors for two instances of the same
#   application and operation (different input arrival times).  Currently,
#   the input arrival time of the predecessor is being ignored from the
#   daily plan report.  In the future, we may wish to redesign the
#   table and processing to retain the input arrival times.  For now, we
#   will just eliminate any "duplicate" predecessors and realize that we 
#   have some loss of data.

  if (!predecessors[plan_id delimiter op_id delimiter ext_app delimiter predecessor]) {
    print plan_id delimiter op_id delimiter ext_app delimiter predecessor >> "opc_predecessor.bcp";
    predecessors[plan_id delimiter op_id delimiter ext_app delimiter predecessor] = 1
  }
}
#These are the RuleEngine versions of the output functions
#function OutputFileHeader()
#{
#  printf("(definstances daily_plan\n") > "daily_plan.clp";
#}
#function OutputFileFooter()
#{
#  printf(")\n") >> "daily_plan.clp";
#}
#function OutputDailyPlan(start_date, start_time, end_date, end_time)
#{
#  start_date=convert_date(start_date);
#  start_time=convert_time(start_time);
#  end_date=convert_date(end_date);
#  end_time=convert_time(end_time);
#  printf("  (PLAN_PERIOD of OPC_PLAN_PERIOD (start_date %d) (start_time %d) (end_date %d) (end_time %d))\n",
#         start_date, start_time, end_date, end_time) >> "daily_plan.clp";
#}
#function OutputApplication(app_name, app_instance, description, owner, priority, status, start_day, start_day_mod, input_time, deadline_day, deadline_time)
#{
#  printf("  (%s.%d of OPC_APPLICATION (app_name \"%s\") (app_id %d) (description \"%s\") (owner \"%s\") (priority %d) (status %s) (start_day %d) (start_day_mod \"%s\") (input_time %d) (deadline_day %d) (deadline_time %d))\n",
#          app_name, app_instance, app_name, app_id, description, owner, priority, status, start_day, start_day_mod, input_time, deadline_day, deadline_time) >> "daily_plan.clp";
#}
#function OutputOperation(app_name, app_instance, operation, app_id, op_id, jobname, status, start_day, start_day_mod, start_time, duration, deadline_day, deadline_time)
#{
#  if (start_day_mod=="")
#    start_day_mod="_";
#  printf("    (%s.%d.%s of OPC_OPERATION (app_id %d) (op_id %d) (operation %s) (jobname \"%s\") (status %s) (start_day %d) (start_mod %c) (start_time %d) (duration %d) (deadline_day %d) (deadline_time %d))\n",
#         app_name, app_instance, operation, app_id, op_id, operation, jobname, status, start_day, start_day_mod, start_time, duration, deadline_day, deadline_time) >> "daily_plan.clp";
#}
#function OutputPredecessor(app_name, app_instance, operation, ext_app, predecessor, app_id, op_id)
#{
#  printf("      (%s.%d.%s.%s.%s of OPC_PREDECESSOR (app_id %d) (succ_op_id %d) (pred_app_name \"%s\") (pred_op_name %s))\n",
#         app_name, app_instance, operation, ext_app, predecessor, app_id, op_id, ext_app, predecessor) >> "daily_plan.clp";
#}
BEGIN {
  while (getline < keywordfile) {
    n=split($0, k, "=");
    if (n==0) {
      print "BAD KEYWORD FILE \"" keywordfile "\"";
      exit 1;
    }
    kw=k[1];
    sub("[ ]*$", "", kw);
    kv=k[2]
    # Make variable assignment, checking for overrides from command line
    # Currently the only supported override is -d<DATE_FORMAT> but if you
    # add others keep them consistent with this style.
    if (kw == "DATE_FORMAT" && DATE_FORMAT != "")
      kv = DATE_FORMAT;

    keyword[kw]=kv;
  }
  close(keywordfile);
  date_format=keyword["DATE_FORMAT"];
  day_pos=index(date_format, "D");
  month_pos=index(date_format, "M");
  year_pos=index(date_format, "Y");

  APPLICATION_EXPR="^" indent "[^ ]*[ ][ ]*[0-9] [A-Z ][A-Z][^A-Za-z0-9][0-9][0-9][ <>][0-9][0-9].[0-9][0-9]";
  OPERATION_EXPR="^" indent "[ ][A-Z0-9#]";
  APP_BREAK_EXPR="^" indent "[-][-][-]";
  offset=length(indent);

  # The seenPlanStart flag will allow us to determine if we are
  # processing the actual daily plan report.  The flag is not
  # turned on until we hit the start of the report.  This way, we 
  # can ignore irrelevant data thay may precede the report - such 
  # as JES messages from the host batch job that produced the 
  # report.

  seenPlanStart = 0

#  print "APPLICATION_EXPR=\"" APPLICATION_EXPR "\""
#  print "OPERATION_EXPR=\"" OPERATION_EXPR "\""
  OutputFileHeader();
}
$0 ~ keyword["REQUESTED_PLANNING_PERIOD"] {
  pat=keyword["REQUESTED_PLANNING_PERIOD"];
  sub(pat, "", $0);
  OutputDailyPlan($2, $3, $5, $6);

  seenPlanStart = 1
}
seenPlanStart && NF>1 && $0 ~ APPLICATION_EXPR {
  ++napps;
  if (last_app_name == $1)
    app_instance++;
  else {
    app_instance=1;
    last_app_name=$1;
  }
  app_name=$1;
  priority=$2;
  status=$3;
  start_day=rtrim(substr($0, 27+offset, 2));
  start_day_mod=parse_day_mod(substr($0, 29+offset, 1));
  input_time=convert_time(substr($0, 30+offset, 5));
  deadline_day_mod=parse_day_mod(substr($0, 41+offset, 1));
  deadline_day=rtrim(substr($0, 42+offset, 2));
  deadline_time=convert_time(substr($0, 45+offset, 5));
  description=rtrim(substr($0, 51+offset, 25));
  # Owner is specified on the next line
  getline;
  owner=ltrim(rtrim(substr($0,1+offset,24)));
  ++app_id;
  OutputApplication(app_id, app_name, app_instance, description, owner, priority, status, start_day, start_day_mod, input_time, deadline_day, deadline_day_mod, deadline_time);
}
seenPlanStart && NF>1 && $0 ~ OPERATION_EXPR && $1!=keyword["OP_ID"] {
  while ($0 !~ APP_BREAK_EXPR) {
    if ($0 ~ /^1/) {  # skip page break
      page=$NF;
      for (i = 0; i < 7; i++)
        if (getline==0)
          exit;
    }
    if ($0~keyword["APPLICATION_OWNER"] ||
        $1==keyword["OP_ID"]) {
      break;
    }
    if ($0 ~ OPERATION_EXPR) {     # Operation definition
      operation=rtrim(substr($0, 2+offset, 8));
      jobname=rtrim(substr($0, 11+offset, 8));
      status=substr($0, 25+offset, 1);
      start_day=rtrim(substr($0, 27+offset, 2));  
      if (start_day == "")  # Not really an operation definition -- mostly likely an application owner after a page break
        break;
      if (status == "C") # Completed job - we do not load these
        break;
      start_day_mod=parse_day_mod(substr($0, 29+offset, 1));
      start_time=convert_time(substr($0, 30+offset, 5));
      duration=convert_time(substr($0, 36+offset, 5));
      deadline_day_mod=parse_day_mod(substr($0, 41+offset, 1));
      deadline_day=rtrim(substr($0, 42+offset, 2));
      deadline_time=convert_time(substr($0, 45+offset, 5));
      description=rtrim(substr($0, 52+offset, 24));
      op_id++;
      if (status == " ") status = "";
      # Compute instance number of this record. We group all operations by app_name, operation, jobname, start_time.
      # The first time we see one with a given set it is instance 0, incremented by 1 thereafter.
      instanceno=num_instances[app_name,operation,jobname,start_time]++;
      OutputOperation(app_id, op_id, operation, jobname, status, start_day, start_day_mod, start_time, duration, deadline_day, deadline_day_mod, deadline_time, description, instanceno);
    }
    # Check for predecessor definition
    ext_app=rtrim(substr($0, 77+offset, 17));
    predecessor=rtrim(substr($0, 94+offset, 8));
    if (predecessor != "") {
      if (ext_app=="")
        ext_app=app_name;
      OutputPredecessor(app_name, app_instance, operation, ext_app, predecessor, app_id, op_id);
    }
    if (getline==0)
      exit;
  }
}
$0 ~ keyword["END_OF_DAILY_OPERATING_PLAN"] { print $0; exit(0); }
END {
  OutputFileFooter();
}' $daily_plan_file

if [ $bCheckForBinary = 1 ]; then
  # Restore original file to preserve timestamp
  mv -f $daily_plan_file.save $daily_plan_file 
fi

err=$?
if [ $err != 0 ]; then
  exit $err
fi


bcp_table()
{
  table=$1
  cmd="bcp OPCLoad..$table in ../$table.bcp $bcpopts -f $table.fmt"
  echo $cmd
  eval $cmd
}

# Load tables from bcp files
bcp_tables()
{

if [ $bNoLoadApplications = 0 ]; then
  # Create BCP format file for opc_application
  cat <<- END > opc_application.fmt
6.0
11
1       SQLCHAR       0       12      "|"       1       plan_id
2       SQLCHAR       0       12      "|"       2       app_id
3       SQLCHAR       0       20      "|"       3       app_name
4       SQLCHAR       0       24      "|"       4       app_owner
5       SQLCHAR       0       2       "|"       5       start_day
6       SQLCHAR       0       1       "|"       6       start_day_mod
7       SQLCHAR       0       4       "|"       7       start_hour
8       SQLCHAR       0       2       "|"       8       stop_day
9       SQLCHAR       0       1       "|"       9       stop_day_mod
10      SQLCHAR       0       4       "|"       10      stop_hour
11      SQLCHAR       0       24     "\r\n"     11      app_desc
END
  bcp_table opc_application
fi

if [ $bNoLoadOperations = 0 ]; then
  # Create BCP format file for opc_operation
  cat <<- END > opc_operation.fmt
7.0
16
1       SQLCHAR       0       20      "|"       1       plan_id
2       SQLCHAR       0       20      "|"       2       app_id
3       SQLCHAR       0       20      "|"       3       app_name
4       SQLCHAR       0       20      "|"       4       op_id
5       SQLCHAR       0       20      "|"       5       op_name
6       SQLCHAR       0       8       "|"       6       job_name
7       SQLCHAR       0       1       "|"       7       status
8       SQLCHAR       0       2       "|"       8       start_day
9       SQLCHAR       0       1       "|"       9       start_day_mod
10      SQLCHAR       0       4       "|"       10      start_hour
11      SQLCHAR       0       4       "|"       11      duration
12      SQLCHAR       0       2       "|"       12      stop_day
13      SQLCHAR       0       1       "|"       13      stop_day_mod
14      SQLCHAR       0       4       "|"       14      stop_hour
15      SQLCHAR       0       24      "|"       15      op_desc
16      SQLCHAR       0       4       "\r\n"    23      instanceno
END
  bcp_table opc_operation
fi

if [ $bNoLoadPredecessors = 0 ]; then
  # Create BCP format file for opc_predecessor
  cat <<- END > opc_predecessor.fmt
6.0
4
1       SQLCHAR       0       12      "|"       1       plan_id
2       SQLCHAR       0       12      "|"       2       succ_op_id
3       SQLCHAR       0       20      "|"       3       pred_app_name
4       SQLCHAR       0       8       "\r\n"    4       pred_op_name
END
  bcp_table opc_predecessor
fi

}

main()
{
if [ $bNoLoadAll = 0 ]; then
  print "Starting load"
  bcp_tables
fi

if [ $bNoMap = 0 ]; then
  map_schedules
  exit
fi
}

mkdir -p ./logs/
(cd logs; main > "$0.$daily_plan_file"."$(date +""%Y""%m""%d"_"%H""%M""%S"")".$$.out.txt)
