-- File: archive/db/sql/_discoverimsresources.sql
-- SID: 1.2
-- Date Modified: 02/01/16 19:24:46
--
-- Module Description
--   Defines the _DiscoverIMSResources stored procedure

Use Object
GO

-------------------------------------------------------------------------
-- _DiscoverIMSResources
PRINT 'Creating procedure _DiscoverIMSResources'
GO

IF EXISTS ( SELECT * FROM sysobjects WHERE name = '_DiscoverIMSResources' AND type = 'P')
  DROP PROCEDURE _DiscoverIMSResources
GO

-- File: archive/db/sql/_discoverimsresources.sql
-- SID: 1.2
-- Date Modified: 02/01/16 19:24:46
-- As shipped in patch 24 - forthman 
-- Works with DiscoveryBatch, IMSDiscovery, ProcessedIMSDiscovery tables to discover IMS objects
--   Will rediscover any pre-existing objects and create new objects for any that it does not find.
CREATE PROCEDURE _DiscoverIMSResources
  @cid          ClassID,
  @id           ObjID,
  @batchid      INT = NULL,
  @incremental  TINYINT = 0,
  @verbose      TINYINT = 0
AS
  DECLARE @err INT, @rc INT

  SET NOCOUNT ON

  RAISERROR( '_DiscoverIMSResources @cid = ''%s'', @id = %d, @batchid = %d, @incremental = %d , @verbose = %d', 0, 1, @cid, @id, @batchid, @incremental, @verbose)

  -- HACK
  SELECT GETDATE()

  -- Initialization FOR the timings of steps
  DECLARE @startdate DATETIME
  DECLARE @nowdate   DATETIME
  DECLARE @stepdate  DATETIME
  SELECT @nowdate    = GETDATE()
  SELECT @stepdate   = @nowdate
  SELECT @startdate  = @stepdate
  DECLARE @stepname  NVARCHAR(20)
  SELECT @stepname = NULL
  DECLARE @steptime INT
  DECLARE @totaltime INT
  DECLARE @debug INT

  SELECT @debug = 1

  -- Timing header
  IF @debug >= 1
  BEGIN
    RAISERROR( 'Timings: Stepname             ( s_step/ s_total)', 0, 1)
  END

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'INIT'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END


  -- IF the batchid was NOT specified THEN pick the oldest one that IS IN a LOADED state
  IF @batchid IS NULL
  BEGIN
    SELECT @batchid = min(batchid) 
    FROM DiscoveryBatch AS b JOIN DiscoveryBatchPStatVal AS v ON v.pstat = b.pstat
    WHERE
      FormatType = 18
--      AND ActionType = 1
      AND ( ActionType = 1 OR ActionType = 12) -- HACK ActionType 12, remove 12 after fixing data upstream
      AND v.name = 'LOADED'
    IF @batchid IS NULL
    BEGIN
      PRINT 'No IMS DiscoveryBatch IS IN a LOADED state, nothing to do!'
      RETURN 0
    END
  END

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Get Batchid'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

/* This garbage collect should be done as a daily job, and does not need to be done here 
   All queries check to see if the link table is references are deleted = 0
  -- Garbage collection IF NOT running IN incremental mode
--  IF @incremental = 0
--  BEGIN
    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Starting garbage collection'
    DECLARE linkc CURSOR FOR
      SELECT dol.link_type, RTRIM(dol.src_cid), RTRIM(dol.dst_cid), dst.cname
      FROM derived_obj_link AS dol
        JOIN obj_class AS src ON src.cid = dol.src_cid
        JOIN obj_class AS dst ON dst.cid = dol.dst_cid
        JOIN isa_chain AS dic ON dic.cid = dol.dst_cid
        JOIN link_type AS lt ON lt.link_type = dol.link_type
      WHERE
        src.flag_abstract = 0
        AND dst.flag_abstract = 0
        AND dic.base_cid = 'IMBA'
        AND lt.flag_containment = 1
    OPEN linkc
    DECLARE @link_type LinkType, @src_cid ClassID, @dst_cid ClassID, @dst_cname NVARCHAR(24)
    FETCH NEXT FROM linkc INTO @link_type, @src_cid, @dst_cid, @dst_cname
    WHILE @@FETCH_STATUS <> -1
    BEGIN
      -- DELETE link records
      DECLARE @table sysname
      SET @table = @src_cid + '_' + @dst_cid + '_' + @link_type
      IF @verbose = 1
        PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Doing garbage collection ON ' + @table
      EXEC("
      DELETE FROM il FROM " + @table + " AS il JOIN link AS l ON
        l.link_type = '" + @link_type + "'
        AND l.src_cid = '" + @src_cid + "'
        AND l.src_id = il.src_id
        AND l.dst_cid = '" + @dst_cid + "'
        AND l.dst_id = il.dst_id
      WHERE
        l.deleted = 1
      ")
      SET @rc = @@ROWCOUNT
      IF @verbose = 1
        PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + '  ROWCOUNT=' + CAST(@rc AS NVARCHAR)
    
      -- DELETE instance records
      IF @verbose = 1
        PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Doing garbage collection ON ' + @dst_cname + '_C'
      EXEC("
      DELETE FROM " + @dst_cname + "_C WHERE id NOT IN (SELECT l.dst_id FROM link AS l WHERE l.dst_cid='" + @dst_cid + "' AND l.deleted=0)
      ")
      SET @rc = @@ROWCOUNT
      IF @verbose = 1
        PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + '  ROWCOUNT=' + CAST(@rc AS NVARCHAR)

      FETCH NEXT FROM linkc INTO @link_type, @src_cid, @dst_cid, @dst_cname
    END
    DEALLOCATE linkc

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'GC link and instance'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

    -- Garbage-collect the directly contained instances.
   
    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Doing garbage collection ON IMSProgram_C'
    DELETE FROM child FROM IMSProgram_C AS child WHERE NOT EXISTS
    (
      SELECT * FROM IMSSubsystem_C AS parent WHERE parent.cid = child._ParentCID AND parent.id = child._ParentID
    )
    SET @rc = @@ROWCOUNT
    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + '  ROWCOUNT=' + CAST(@rc AS NVARCHAR)

    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Doing garbage collection ON IMSDatabase_C'
    DELETE FROM child FROM IMSDatabase_C AS child WHERE NOT EXISTS
    (
      SELECT * FROM IMSProgram_C AS parent WHERE parent.cid = child._ParentCID AND parent.id = child._ParentID
    )
    SET @rc = @@ROWCOUNT
    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + '  ROWCOUNT=' + CAST(@rc AS NVARCHAR)

    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Doing garbage collection ON IMSTransaction_C'
    DELETE FROM child FROM IMSTransaction_C AS child WHERE NOT EXISTS
    (
      SELECT * FROM IMSProgram_C AS parent WHERE parent.cid = child._ParentCID AND parent.id = child._ParentID
    )
    SET @rc = @@ROWCOUNT
    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + '  ROWCOUNT=' + CAST(@rc AS NVARCHAR)

    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Doing garbage collection ON IMSArea_C'
    DELETE FROM child FROM IMSArea_C AS child WHERE NOT EXISTS
    (
      SELECT * FROM IMSDatabase_C AS parent WHERE parent.cid = child._ParentCID AND parent.id = child._ParentID
    )
    SET @rc = @@ROWCOUNT
    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + '  ROWCOUNT=' + CAST(@rc AS NVARCHAR)
*/
  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'GC Direct Containment'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

--  END -- END of garbage collection

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Garbage Collection'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  -- Take care of any mismatches between discovery data obj_cid AND obj_class.cid
  UPDATE IMSDiscovery SET obj_cid = 'IMQA' WHERE obj_cid = 'IMCQ'

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Update cid in IMSDiscovery'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  -- Rediscover IMSSubsystem objects
  IF @verbose = 1
    PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Mapping SystemId to os_id'
  UPDATE child SET
    os_id = p.id
  FROM
    IMSDiscovery AS child
    JOIN OS_V AS p ON p._SystemId = child.SystemId

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Map SystemId to os_id'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END


  -- IF we could NOT map any of the records FROM the batch to an operating system THEN
  -- there's something wrong.
  IF NOT EXISTS
  (
    SELECT * FROM IMSDiscovery WHERE batchid = @batchid AND os_id IS NOT NULL
  )
  BEGIN
    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: batch ' + CAST(@batchid AS NVARCHAR) + ' cannot be resolved to any operating system. The SystemId field of the discovery records does NOT match any known operating systems.'
    RETURN -30000
  END

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Error on no _SystemId'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  -- Compute the path of all objects. We can do this now because all operating systems have been looked up.
  DECLARE osc CURSOR FOR
    SELECT DISTINCT os_id FROM IMSDiscovery WHERE batchid = @batchid
  OPEN osc
  DECLARE @os_id ObjID, @Path NVARCHAR(255)
  FETCH NEXT FROM osc INTO @os_id
  WHILE @@FETCH_STATUS <> -1
  BEGIN
    EXEC asisp_fullpath OS, @os_id, @Path OUTPUT
    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Updating paths FOR objects under ' + @Path
    UPDATE IMSDiscovery SET obj_path = 
      @Path
      + '/' + IMS_SUBSYSTEM
      + CASE WHEN imsc_cid IS NOT NULL THEN '/' + IMS_SUBCOMPONENT END
      + CASE WHEN imsp_cid IS NOT NULL THEN '/' + IMS_SUBPROGRAM END
      + CASE WHEN obj_cid = 'IMAR' THEN '/' + IMS_AREA END
    WHERE
      batchid = @batchid
      AND os_id = @os_id
    FETCH NEXT FROM osc INTO @os_id
  END
  DEALLOCATE osc

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Compute Path of batch'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  DECLARE @flagDiscovery INT

  IF EXISTS ( SELECT * FROM IMSDiscovery AS child WHERE batchid = @batchid AND IMS_SUBSYSTEM IS NOT NULL ) 
    SELECT @flagDiscovery = 1
  ELSE
    SELECT @flagDiscovery = 0

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'flagDiscovery IMSS/TM/DB'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END


  IF ( @flagDiscovery = 1 )
  BEGIN
  -- Rediscover Subsystems
  DECLARE imsu_cid CURSOR FOR
    SELECT DISTINCT imsu_cid 
    FROM IMSDiscovery AS child
    WHERE 
          child.batchid = @batchid 
      AND child.IMS_SUBSYSTEM IS NOT NULL
  OPEN imsu_cid 
  DECLARE @imsu_cid ClassID
  DECLARE @imsu_table NVARCHAR(80)
  DECLARE @sql NVARCHAR(1000)
  FETCH NEXT FROM imsu_cid INTO @imsu_cid
  WHILE @@FETCH_STATUS <> -1 
  BEGIN
    SELECT @imsu_table = cname + '_C' FROM obj_class WHERE cid = @imsu_cid
    SELECT @sql = '
  UPDATE child SET
    imsu_id = obj.id,
    rediscovered = 1
  FROM
    IMSDiscovery AS child
    JOIN ' + @imsu_table + ' AS obj ON          -- Find any existing IMSSubsystem objects by class,name
      obj.cid = child.obj_cid
      AND obj.name = child.IMS_SUBSYSTEM
    JOIN link AS pc ON                     -- Lookup FOR parent links of subsystems to make sure
      pc.src_cid = ''OS''                    -- we are under the right operating system
      AND pc.src_id = child.os_id 
      AND pc.dst_cid = obj.cid
      AND pc.dst_id = obj.id
      AND pc.deleted = 0                   -- USE only Active objects
  WHERE
        child.batchid = ' + CONVERT( NVARCHAR(20), @batchid ) + '
    AND child.IMS_SUBSYSTEM IS NOT NULL        -- Only attempt to rediscovery IMS_SUBSYSTEM with actual names'

--    PRINT @sql

    EXEC (@sql)
    FETCH NEXT FROM imsu_cid INTO @imsu_cid 
  END
  CLOSE imsu_cid 
  DEALLOCATE imsu_cid 

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Rd IMSS/TM/DB'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  -- HACK
--  SELECT rediscovered, * FROM IMSDiscovery WHERE batchid = @batchid
  
  -- CREATE any required subsystems
  IF EXISTS (
  SELECT * FROM IMSDiscovery AS d
    INNER JOIN isa_chain AS isa ON d.obj_cid = isa.cid
    WHERE batchid = @batchid 
      AND imsu_id IS NULL 
      AND isa.base_cid = 'IMSU'
  )
  BEGIN
    EXEC @rc = _CreateDiscoveredObjects
      @cid,
      @id,
      @batchid         = @batchid,
      @DiscoveryTable  = 'IMSDiscovery',
      @LinkTypeColumn  = '''PHYC''',
      @ParentCIDColumn = '''OS''',
      @ParentIDColumn  = 'os_id',
      @ChildCIDColumn  = 'imsu_cid',
      @ChildIDColumn   = 'imsu_id',
      @ChildNameColumn = 'IMS_SUBSYSTEM',
      @WhereClause     = 'JOIN isa_chain AS ic ON ic.cid = d.obj_cid WHERE ic.base_cid = ''IMSU'' AND d.os_id IS NOT NULL AND d.imsu_id IS NULL',
      @AdditionalDiscoveryColumns = ', obj_path',
      @AdditionalAttributeColumns = ', _Path',
      @verbose         = @verbose
    IF @rc > 0
    BEGIN
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ':Discovered ' + CAST(@rc AS NVARCHAR) + ' subsystems'
    END
  END

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Cr IMSS/TM/DB'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  -- UPDATE obj_id for all the IMS subsystem objects
  UPDATE IMSDiscovery SET obj_id = imsu_id WHERE obj_cid = imsu_cid

  -- UPDATE the subsystem ids of all objects
  UPDATE d SET imsu_id = imsud.obj_id
  FROM
    IMSDiscovery AS d
    JOIN IMSDiscovery AS imsud ON
      imsud.obj_cid = d.imsu_cid
      AND imsud.obj_name = d.IMS_SUBSYSTEM
  WHERE
    d.obj_cid NOT IN (SELECT cid FROM isa_chain WHERE base_cid = 'IMSU')
      AND d.batchid = @batchid
      AND imsud.obj_id IS NOT NULL

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Up IMSS/TM/DB'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  IF @verbose = 1
    SELECT 'IMSSubsystem', * FROM IMSDiscovery AS d JOIN isa_chain AS i ON i.cid = d.obj_cid
    WHERE batchid = @batchid AND i.base_cid = 'IMSS'

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Di IMSS/TM/DB'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  END


  IF EXISTS ( SELECT * FROM IMSDiscovery AS child WHERE batchid = @batchid AND IMS_SUBCOMPONENT IS NOT NULL AND obj_cid <> 'IMPG') 
    SELECT @flagDiscovery = 1
  ELSE
    SELECT @flagDiscovery = 0

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'flagDiscovery IMSC no IMPG'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END


  IF ( @flagDiscovery = 1 )
  BEGIN
  -- Rediscover objects that live under the subsystems
  IF @verbose = 1
    PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Rediscovering subcomponents under subsystems'
  UPDATE child SET
    imsc_link_type = dol.link_type
  FROM
    IMSDiscovery AS child
    JOIN derived_obj_link AS dol ON    -- Lookup link_type FOR relationship
      dol.src_cid = child.imsu_cid
      AND dol.dst_cid = child.obj_cid

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Rd IMSC, Step 1'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  DECLARE imsc_cid CURSOR FOR
    SELECT DISTINCT imsc_cid 
    FROM IMSDiscovery AS child
    WHERE 
          child.batchid = @batchid 
      AND child.IMS_SUBCOMPONENT IS NOT NULL
  OPEN imsc_cid 
  DECLARE @imsc_cid ClassID
  DECLARE @imsc_table NVARCHAR(80)
  FETCH NEXT FROM imsc_cid INTO @imsc_cid
  WHILE @@FETCH_STATUS <> -1 
  BEGIN
    SELECT @imsc_table = cname + '_C' FROM obj_class WHERE cid = @imsc_cid
    SELECT @sql = '

  UPDATE child SET
    imsc_id = obj.id,
    rediscovered = 1
  FROM
    IMSDiscovery AS child
    JOIN ' + @imsc_table + ' AS obj ON
      obj.cid = child.obj_cid
      AND obj.name = child.IMS_SUBCOMPONENT
    JOIN link AS pc ON                 -- Make sure we are identifying the right subcomponent
      pc.src_cid = child.imsu_cid      -- under the right subsystem.
      AND pc.src_id = child.imsu_id
      AND pc.dst_cid = obj.cid
      AND pc.dst_id = obj.id
      AND pc.link_type = child.imsc_link_type
      AND pc.deleted = 0                   -- USE only Active objects
  WHERE
    child.batchid = ' + CONVERT( NVARCHAR(20), @batchid ) + '
    AND child.IMS_SUBCOMPONENT IS NOT NULL
'

--    PRINT @sql

    EXEC (@sql)
    FETCH NEXT FROM imsc_cid INTO @imsc_cid
  END
  CLOSE imsc_cid 
  DEALLOCATE imsc_cid 

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Rd IMSC, Step 2'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  -- HACK
--  SELECT rediscovered, * FROM IMSDiscovery WHERE batchid = @batchid

  -- CREATE any required subcomponents
  IF EXISTS (SELECT * FROM IMSDiscovery WHERE batchid = @batchid AND imsu_id IS NOT NULL AND imsc_cid IS NOT NULL AND imsc_cid <> 'IMPG' AND imsc_id IS NULL)
  BEGIN
    EXEC @rc = _CreateDiscoveredObjects
      @cid,
      @id,
      @batchid         = @batchid,
      @DiscoveryTable  = 'IMSDiscovery',
      @LinkTypeColumn  = 'imsc_link_type',
      @ParentCIDColumn = 'imsu_cid',
      @ParentIDColumn  = 'imsu_id',
      @ChildCIDColumn  = 'imsc_cid',
      @ChildIDColumn   = 'imsc_id',
      @ChildNameColumn = 'IMS_SUBCOMPONENT',
      @WhereClause     = 'JOIN isa_chain AS ic ON ic.cid = d.obj_cid WHERE ic.base_cid IN ( ''IMBM'', ''IMSC'' ) AND d.imsu_id IS NOT NULL AND d.imsc_id IS NULL',
      @AdditionalDiscoveryColumns = ',  DD, DSN, obj_path',
      @AdditionalAttributeColumns = ',  _DDName, _Dataset, _Path',
      @verbose         = @verbose
    IF @rc > 0
    BEGIN
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ':Discovered ' + CAST(@rc AS NVARCHAR) + ' subcomponents'
    END
  END

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Cr IMSC'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  END



  IF EXISTS ( SELECT * FROM IMSDiscovery AS child WHERE batchid = @batchid AND IMS_SUBCOMPONENT IS NOT NULL AND obj_cid = 'IMPG') 
    SELECT @flagDiscovery = 1
  ELSE
    SELECT @flagDiscovery = 0

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'flagDiscovery IMPG'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END


  IF ( @flagDiscovery = 1 )
  BEGIN
  -- Rediscover programs under the subsystems. These are directly contained AND therefore
  -- NOT handled like the other subcomponents above.
  IF @verbose = 1
    PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Rediscovering programs under subsystems'
  UPDATE child SET
    imsc_id = obj.id,
    rediscovered = 1
  FROM
    IMSDiscovery AS child
    JOIN IMSProgram_C AS obj ON
      obj.cid = child.obj_cid
      AND obj.name = child.IMS_SUBCOMPONENT
      AND obj._ParentCID = child.imsu_cid
      AND obj._ParentID = child.imsu_id
      AND obj.deleted = 0
  WHERE
        child.imsc_cid = 'IMPG'
    AND child.IMS_SUBCOMPONENT IS NOT NULL

  -- HACK
--  SELECT rediscovered, * FROM IMSDiscovery WHERE batchid = @batchid


  -- CREATE any required programs
  IF EXISTS (SELECT * FROM IMSDiscovery WHERE batchid = @batchid AND imsu_id IS NOT NULL AND imsc_cid = 'IMPG' AND imsc_id IS NULL)
  BEGIN
    EXEC @rc = _CreateDiscoveredObjects
      @cid,
      @id,
      @batchid         = @batchid,
      @DiscoveryTable  = 'IMSDiscovery',
      @LinkTypeColumn  = 'NULL',
      @ParentCIDColumn = 'imsu_cid',
      @ParentIDColumn  = 'imsu_id',
      @ChildCIDColumn  = 'imsc_cid',
      @ChildIDColumn   = 'imsc_id',
      @ChildNameColumn = 'IMS_SUBCOMPONENT',
      @WhereClause     = 'WHERE d.obj_cid = ''IMPG'' AND d.imsu_id IS NOT NULL AND d.imsc_id IS NULL',
      @AdditionalDiscoveryColumns = ', imsu_cid, imsu_id, obj_path',
      @AdditionalAttributeColumns = ', _ParentCID, _ParentID, _Path',
      @verbose         = @verbose
    IF @rc > 0
    BEGIN
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ':Discovered ' + CAST(@rc AS NVARCHAR) + ' programs'
    END
  END

  END



  IF EXISTS ( SELECT * FROM IMSDiscovery AS child WHERE batchid = @batchid AND IMS_SUBCOMPONENT IS NOT NULL ) 
    SELECT @flagDiscovery = 1
  ELSE
    SELECT @flagDiscovery = 0

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'flagDiscovery IMSC'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END


  IF ( @flagDiscovery = 1 )
  BEGIN

  -- UPDATE obj_id for all the IMS subcomponent objects
  UPDATE IMSDiscovery SET obj_id = imsc_id WHERE obj_cid = imsc_cid

  -- UPDATE the subcomponent ids of all objects
  UPDATE d SET imsc_id = imscd.obj_id
  FROM
    IMSDiscovery AS d
    JOIN IMSDiscovery AS imscd ON
          imscd.obj_cid  = d.imsc_cid
      AND imscd.obj_name = d.IMS_SUBCOMPONENT
      AND imscd.imsu_cid = d.imsu_cid
      AND imscd.imsu_id  = d.imsu_id
  WHERE
    d.obj_cid IN (SELECT cid FROM isa_chain WHERE base_cid IN ( 'IMSP', 'IMAR' )) 
      AND d.batchid = @batchid
      AND imscd.obj_id IS NOT NULL


  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Up IMSC'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  END


  IF EXISTS ( SELECT * FROM IMSDiscovery AS child WHERE batchid = @batchid AND IMS_SUBPROGRAM IS NOT NULL ) 
    SELECT @flagDiscovery = 1
  ELSE
    SELECT @flagDiscovery = 0

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'flagDiscovery IMSP'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END


  IF ( @flagDiscovery = 1 )
  BEGIN
  IF @verbose = 1
    PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Rediscovering subprograms'
-- wrap the following in a cursor
  -- Rediscover objects that live under programs

  DECLARE imsp_cid CURSOR FOR
    SELECT DISTINCT imsp_cid 
    FROM IMSDiscovery AS child
    WHERE 
          child.batchid = @batchid 
      AND child.IMS_SUBPROGRAM IS NOT NULL
  OPEN imsp_cid 
  DECLARE @imsp_cid ClassID
  DECLARE @imsp_table NVARCHAR(80)
  FETCH NEXT FROM imsp_cid INTO @imsp_cid
  WHILE @@FETCH_STATUS <> -1 
  BEGIN
    SELECT @imsp_table = cname + '_C' FROM obj_class WHERE cid = @imsp_cid
    SELECT @sql = '

  UPDATE child SET
    imsp_id = obj.id,
    rediscovered = 1
  FROM
    IMSDiscovery AS child
    JOIN ' + @imsp_table + ' AS obj ON
      obj.cid = child.obj_cid
      AND obj.name = child.IMS_SUBPROGRAM
      AND obj._ParentCID = child.imsc_cid
      AND obj._ParentID = child.imsc_id
      AND obj.deleted = 0
  WHERE
        child.batchid = ' + CONVERT( NVARCHAR(20), @batchid ) + '
    AND child.IMS_SUBPROGRAM IS NOT NULL
'

--    PRINT @sql

    EXEC (@sql)
    FETCH NEXT FROM imsp_cid INTO @imsp_cid
  END
  CLOSE imsp_cid 
  DEALLOCATE imsp_cid 


  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Rd IMSP'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  -- HACK
--  SELECT rediscovered, * FROM IMSDiscovery WHERE batchid = @batchid


  -- CREATE any required subprograms
  IF EXISTS (SELECT * FROM IMSDiscovery WHERE batchid = @batchid AND imsc_cid = 'IMPG' AND imsc_id IS NOT NULL AND imsp_cid IS NOT NULL AND imsp_id IS NULL)
  BEGIN
    EXEC @rc = _CreateDiscoveredObjects
      @cid,
      @id,
      @batchid         = @batchid,
      @DiscoveryTable  = 'IMSDiscovery',
      @LinkTypeColumn  = 'NULL',
      @ParentCIDColumn = '''IMPG''',
      @ParentIDColumn  = 'imsc_id',
      @ChildCIDColumn  = 'imsp_cid',
      @ChildIDColumn   = 'imsp_id',
      @ChildNameColumn = 'IMS_SUBPROGRAM',
      @WhereClause     = 'JOIN isa_chain AS ic ON ic.cid = d.obj_cid WHERE ic.base_cid = ''IMSP'' AND d.imsc_cid = ''IMPG'' AND d.imsc_id IS NOT NULL AND d.imsp_id IS NULL',
      @AdditionalDiscoveryColumns = ', DD, DSN, ''IMPG'', imsc_id, obj_path',
      @AdditionalAttributeColumns = ', _DDName, _Dataset, _ParentCID, _ParentID, _Path',
      @verbose         = @verbose
    IF @rc > 0
    BEGIN
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ':Discovered ' + CAST(@rc AS NVARCHAR) + ' subprograms'
    END
  END

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Cr IMSP'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  -- UPDATE obj_id for all the IMS subprogram objects
  UPDATE IMSDiscovery SET obj_id = imsp_id WHERE obj_cid = imsp_cid

  -- UPDATE the subprogram ids of all objects
  UPDATE d SET imsp_id = imspd.obj_id
  FROM
    IMSDiscovery AS d
    JOIN IMSDiscovery AS imspd ON
          imspd.obj_cid  = d.imsp_cid
      AND imspd.obj_name = d.IMS_SUBPROGRAM
      AND imspd.imsu_cid = d.imsu_cid 
      AND imspd.imsu_id  = d.imsu_id  
      AND imspd.imsc_cid = d.imsc_cid 
      AND imspd.imsc_id  = d.imsc_id  
  WHERE
    d.obj_cid IN ( 'IMAR' ) 
      AND d.batchid = @batchid
      AND imspd.obj_id IS NOT NULL

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Up IMSP'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END


  END


  IF EXISTS ( SELECT * FROM IMSDiscovery AS child WHERE batchid = @batchid AND IMS_AREA IS NOT NULL ) 
    SELECT @flagDiscovery = 1
  ELSE
    SELECT @flagDiscovery = 0

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'flagDiscovery IMAR'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END


  IF ( @flagDiscovery = 1 )
  BEGIN
  -- Rediscover database areas
  IF @verbose = 1
    PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Rediscovering IMSAreas'
  UPDATE child SET
    imar_id = obj.id,
    rediscovered = 1
  FROM
    IMSDiscovery AS child
    JOIN IMSArea_C AS obj ON
      obj.cid = child.obj_cid
      AND obj.name = child.IMS_AREA
      AND obj._ParentCID = child.imsp_cid
      AND obj._ParentID = child.imsp_id
      AND obj.deleted = 0
  WHERE
    child.batchid = @batchid
    AND child.obj_cid = 'IMAR'  -- child IS IMSArea

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Rd IMAR'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  -- HACK
--  SELECT rediscovered, * FROM IMSDiscovery WHERE batchid = @batchid

  -- CREATE any required areas
  IF EXISTS (SELECT * FROM IMSDiscovery WHERE batchid = @batchid AND imsp_cid = 'IMDA' AND imsp_id IS NOT NULL AND obj_cid = 'IMAR' AND imar_id IS NULL)
  BEGIN
    EXEC @rc = _CreateDiscoveredObjects
      @cid,
      @id,
      @batchid         = @batchid,
      @DiscoveryTable  = 'IMSDiscovery',
      @LinkTypeColumn  = 'NULL',
      @ParentCIDColumn = '''IMDA''',
      @ParentIDColumn  = 'imsp_id',
      @ChildCIDColumn  = '''IMAR''',
      @ChildIDColumn   = 'imar_id',
      @ChildNameColumn = 'IMS_AREA',
      @WhereClause     = 'WHERE d.obj_cid = ''IMAR'' AND d.imsp_cid = ''IMDA'' AND d.imsp_id IS NOT NULL AND d.imar_id IS NULL',
      @AdditionalDiscoveryColumns = ', DD, DSN, ''IMDA'', imsp_id, obj_path',
      @AdditionalAttributeColumns = ', _DDName, _Dataset, _ParentCID, _ParentID, _Path',      @verbose         = @verbose
    IF @rc > 0
    BEGIN
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ':Discovered ' + CAST(@rc AS NVARCHAR) + ' areas'
    END
  END

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Cr IMAR'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  -- UPDATE obj_id for all the IMS areas 
  UPDATE IMSDiscovery SET obj_id = imar_id WHERE obj_cid = 'IMAR'

  -- UPDATE the IMS areas ids of all objects, this case does not exist so nothing to do


  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Up IMAR'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  END

GOTO EndOfProc
  -- IF NOT running IN incremental mode THEN DELETE objects that were NOT rediscovered.
  IF @incremental = 0
  BEGIN
    -- TABLE of objects to be deleted
    CREATE TABLE #DELETE
    (
      reason      NVARCHAR(60),
      link_type   LinkType NULL,
      src_cid     ClassID,
      src_id      ObjID,
      dst_cid     ClassID,
      dst_id      ObjID
    )

    -- Subsystems under the OS's IN the discovery batch that were NOT rediscovered
    IF @verbose = 1
    BEGIN
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Unrediscovered subsystems'
      SELECT DISTINCT 'Unrediscovered subsystems',
        l.link_type, l.src_cid, l.src_id, l.dst_cid, l.dst_id
      FROM link AS l
        JOIN IMSDiscovery AS dis ON l.src_cid = 'OS' AND dis.os_id = l.src_id
        JOIN link_type AS lt ON lt.link_type = l.link_type
      WHERE
        dis.batchid = @batchid
        AND lt.flag_containment = 1
        AND NOT EXISTS
        (
          SELECT * FROM IMSDiscovery AS redis
          WHERE
            redis.batchid = dis.batchid
            AND redis.os_id = dis.os_id
            AND redis.imsu_cid = dis.imsu_cid
            AND redis.imsu_id = dis.imsu_id
        )
    END
PRINT 'SKIPPING'
GOTO EndOfProc
    INSERT INTO #DELETE
      SELECT DISTINCT 'Unrediscovered subsystems',
        l.link_type, l.src_cid, l.src_id, l.dst_cid, l.dst_id
      FROM link AS l
        JOIN IMSDiscovery AS dis ON l.src_cid = 'OS' AND dis.os_id = l.src_id
        JOIN link_type AS lt ON lt.link_type = l.link_type
      WHERE
        dis.batchid = @batchid
        AND lt.flag_containment = 1
        AND NOT EXISTS
        (
          SELECT * FROM IMSDiscovery AS redis
          WHERE
            redis.batchid = dis.batchid
            AND redis.os_id = dis.os_id
            AND redis.imsu_cid = dis.imsu_cid
            AND redis.imsu_id = dis.imsu_id
        )
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('_DiscoverIMSResources: ERROR DELETING UNREDISCOVERED SUBSYSTEMS', 3, 1)
      GOTO EndOfProc
    END

    -- All programs under the unrediscovered subsystems
    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Programs under unrediscovered subsystems'
    INSERT INTO #DELETE
      SELECT DISTINCT 'Programs under unrediscovered subsystems',
        NULL, _ParentCID, _ParentID, cid, id
      FROM
        #DELETE AS d
        JOIN IMSProgram_C AS p ON
          p._ParentCID = d.dst_cid
          AND p._ParentID = d.dst_id
      WHERE
        d.src_cid = 'OS'
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('_DiscoverIMSResources: ERROR DELETING PROGRAMS UNDER UNREDISCOVERED SUBSYSTEMS', 3, 1)
      GOTO EndOfProc
    END

    -- Previously known subcomponents that were NOT rediscovered
    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Previously known subcomponents NOT rediscovered'
    INSERT INTO #DELETE
      SELECT DISTINCT 'Previously known subcomponents NOT rediscovered',
        l.link_type, l.src_cid, l.src_id, l.dst_cid, l.dst_id
      FROM link AS l
        JOIN IMSDiscovery AS dis ON l.src_cid = dis.imsu_cid AND dis.imsu_id = l.src_id
        JOIN isa_chain AS ic ON ic.dst_cid = l.dst_cid
        JOIN link_type AS lt ON lt.link_type = l.link_type
      WHERE
        dis.batchid = @batchid
        AND lt.flag_containment = 1
        AND ic.base_cid =  'IMSC' -- we only want links to subcomponents
        AND dis.imsc_cid <> 'IMPG'  -- exclude programs since these are directly contained
        AND NOT EXISTS
        (
          SELECT * FROM IMSDiscovery AS redis
          WHERE
            redis.batchid = dis.batchid
            AND redis.imsu_cid = dis.imsu_cid
            AND redis.imsu_cid = dis.imsu_cid
            AND redis.imsc_cid = dis.imsc_cid
            AND redis.imsc_id = dis.imsc_id
            AND redis.imsc_id = l.dst_id
        )
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('_DiscoverIMSResources: ERROR DELETING PREVIOUSLY KNOWN SUBCOMPONENTS NOT REDISCOVERED', 3, 1)
      GOTO EndOfProc
    END

    -- Unrediscovered subcomponents
    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Unrediscovered subcomponents under rediscovered subsystems'
    INSERT INTO #DELETE
      SELECT DISTINCT 'Unrediscovered subcomponents under rediscovered subsystems',
        l.link_type, l.src_cid, l.src_id, l.dst_cid, l.dst_id
      FROM
        #DELETE AS d
        JOIN link AS l ON l.src_cid = d.dst_cid AND l.src_id = d.dst_id
        JOIN link_type AS lt ON lt.link_type = l.link_type
      WHERE
        lt.flag_containment = 1
        AND NOT EXISTS
        (
          SELECT * FROM IMSDiscovery AS redis
          WHERE
            redis.batchid = d.batchid
            AND redis.os_id = d.os_id
            AND redis.imsu_cid = d.imsu_cid
            AND redis.imsu_id = d.imsu_id
        )
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('_DiscoverIMSResources: ERROR DELETING UNREDISCOVERED SUBCOMPONENTS UNDER REDISCOVERED SUBSYSTEMS', 3, 1)
      GOTO EndOfProc
    END

    -- Previously known programs that were NOT rediscovered under subsystems that have programs IN
    -- the discovery batch.
    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Previously known programs under subsystems with programs'
    INSERT INTO #DELETE
      SELECT DISTINCT 'Previously known programs under subsystems with programs',
        NULL, _ParentCID, _ParentID, cid, id
      FROM
        IMSProgram_C AS obj
        JOIN IMSDiscovery AS pd ON
          pd.imsu_cid = obj._ParentCID
          AND pd.imsu_id = obj._ParentID
          AND pd.imsc_cid = 'IMPG'
      WHERE
        pd.batchid = @batchid
        AND NOT EXISTS
        (
          SELECT * FROM IMSDiscovery AS redis
          WHERE
            redis.batchid = pd.batchid
            AND redis.os_id = pd.os_id
            AND redis.imsu_cid = pd.imsu_cid
            AND redis.imsu_id = pd.imsu_id
            AND redis.imsc_cid = pd.imsc_cid
            AND redis.imsc_id = pd.imsc_id
            AND redis.imsc_id = obj.id
        )
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('_DiscoverIMSResources: ERROR DELETING PREVIOUSLY KNOWN PROGRAMS UNDER SUBSYSTEMS WITH PROGRAMS', 3, 1)
      GOTO EndOfProc
    END

    -- SubPrograms under the unrediscovered programs
    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Subprograms under unrediscovered programs'
    INSERT INTO #DELETE
      SELECT DISTINCT 'Subprograms under unrediscovered programs',
        NULL, _ParentCID, _ParentID, cid, id
      FROM
        #DELETE AS d
        JOIN IMSSubProgram_C AS sp ON
          sp._ParentCID = d.dst_cid
          AND sp._ParentID = d.dst_id
      WHERE
        d.dst_cid = 'IMPG'
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('_DiscoverIMSResources: ERROR DELETING SUBPROGRAMS UNDER UNREDISCOVERED PROGRAMS', 3, 1)
      GOTO EndOfProc
    END

    -- Previously known subprograms that were NOT rediscovered under programs that have subprograms IN
    -- the discovery batch.
    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Previously known subprograms NOT rediscovered'
    INSERT INTO #DELETE
      SELECT DISTINCT 'Previously known subprograms NOT rediscovered',
        NULL, _ParentCID, _ParentID, cid, id
      FROM
        #DELETE AS d
        JOIN IMSSubProgram_C AS sp ON
          sp._ParentCID = d.dst_cid
          AND sp._ParentID = d.dst_id
        JOIN IMSDiscovery AS pd ON
          pd.imsc_cid = sp._ParentCID
          AND pd.imsc_id = sp._ParentID
          AND pd.imsp_cid = sp.cid
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('_DiscoverIMSResources: ERROR DELETING PREVIOUSLY KNOWN SUBPROGRAMS NOT REDISCOVERED', 3, 1)
      GOTO EndOfProc
    END

    -- Areas under the unrediscovered subprograms of type IMSDatabase
    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Areas under unrediscovered databases'
    INSERT INTO #DELETE
      SELECT DISTINCT 'Areas under unrediscovered databases',
        NULL, _ParentCID, _ParentID, cid, id
      FROM
        #DELETE AS d
        JOIN IMSArea_C AS sp ON
          sp._ParentCID = d.dst_cid
          AND sp._ParentID = d.dst_id
      WHERE
        d.dst_cid = 'IMDA'
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('_DiscoverIMSResources: ERROR DELETING AREAS UNDER UNREDISCOVERED DATABASES', 3, 1)
      GOTO EndOfProc
    END

    -- Previously known areas that were NOT rediscovered under databases that have areas IN
    -- the discovery batch.
    IF @verbose = 1
      PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Previously known areas NOT rediscovered'
    INSERT INTO #DELETE
      SELECT DISTINCT 'Previously known areas NOT rediscovered',
        NULL, _ParentCID, _ParentID, cid, id
      FROM
        #DELETE AS d
        JOIN IMSArea_C AS a ON
          a._ParentCID = d.dst_cid
          AND a._ParentID = d.dst_id
        JOIN IMSDiscovery AS pd ON
          pd.imsc_cid = a._ParentCID
          AND pd.imsc_id = a._ParentID
          AND pd.imsp_cid = a.cid
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('_DiscoverIMSResources: ERROR DELETING PREVIOUSLY KNOWN AREAS NOT REDISCOVERED', 3, 1)
      GOTO EndOfProc
    END

    IF @verbose = 1
      SELECT DISTINCT COUNT(*) AS n, link_type, src_cid, src_id, dst_cid, dst_id, reason
      FROM #DELETE
      GROUP BY link_type, src_cid, src_id, dst_cid, dst_id, reason
  END
    
EndOfProc:

  PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Moving records FROM IMSDiscovery INTO ProcessedIMSDiscovery FOR batch ' + CAST(@batchid AS NVARCHAR)
  -- Move records FROM IMSDiscovery INTO ProcessedIMSDiscovery
  INSERT ProcessedIMSDiscovery SELECT GETDATE(), * 
  FROM IMSDiscovery WHERE batchid = @batchid
  IF @@ERROR = 0
    DELETE IMSDiscovery WHERE batchid = @batchid

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'In ProcessedIMSDiscovery'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END


  -- Mark discovery batch AS completed
  UPDATE DiscoveryBatch SET pend = GETDATE(), pstat = v.pstat, error = 0
  FROM DiscoveryBatchPStatVal AS v
  WHERE batchid = @batchid AND v.name = 'COMPLETED'

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Mark batch as complete'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END

  IF @verbose = 1
    PRINT CONVERT(NVARCHAR(40), GETDATE(), 121) + ': _DiscoverIMSResources: Finished'

  -- Timing of Step
  IF @debug >= 1
  BEGIN
    SELECT @stepname = 'Exit _DiscoverIMSResources'
    SELECT @nowdate    = GETDATE()
    SELECT @steptime = DATEDIFF( s, @stepdate, @nowdate)
    SELECT @totaltime = DATEDIFF( s, @startdate, @nowdate)
    RAISERROR( 'Timings: %-20s (%7d/%8d)', 0, 1, @stepname, @steptime, @totaltime)
    SELECT @stepdate   = @nowdate
  END
GO