-- File: archive/db/sql/_loadopcplan.sql
-- SID: 1.7
-- Date Modified: 01/11/26 15:02:00
--
-- Module Description
--   Defines the _LoadOPCPlan stored procedure

USE OPCLoad
GO


---------------------------------------------------------------------------------------------------
-- _LoadOPCPlan
PRINT 'Creating procedure _LoadOPCPlan'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = '_LoadOPCPlan' AND type = 'P')
  DROP PROCEDURE _LoadOPCPlan
GO

-- File: archive/db/sql/_loadopcplan.sql
-- SID: 1.7
-- Date Modified: 01/11/26 15:02:00
-- Loads the OPC application/operation data from the opc_application/opc_operation tables for the
-- specified OPC daily plan instance. BatchCycles (BCYS) and BatchCycle (BTCY) aggregates
-- are created if needed, AS well AS the batch jobs corresponding to the operations.
CREATE PROCEDURE _LoadOPCPlan
  @cid        ClassID,
  @id         ObjID,
  @planName   NVARCHAR(40),
  @planStart  DATETIME = NULL,
  @verbose    TINYINT = 0,
  @centralServer NVARCHAR(40)=NULL,
  @centralUser   NVARCHAR(40)=NULL,
  @centralPw     NVARCHAR(40)=NULL,
  @centralTrusted TINYINT=0
AS
  SET NOCOUNT ON

  DECLARE @chkpt INT, @msg NVARCHAR(255), @r INT
  DECLARE @cmd NVARCHAR(1024)
  DECLARE @rc INT

  SELECT @chkpt = 1

  DECLARE @tcount INT
  SELECT @tcount = @@TRANCOUNT
  IF @tcount <> 0 AND @centralServer IS NOT NULL
  BEGIN
    RAISERROR('Cannot invoke _LoadOPCPlan from a remote server from inside a transaction', 3, 1)
    RETURN -30000
  END

  DECLARE @verboseFlag NVARCHAR(2)
  IF @verbose <> 0
    SET @verboseFlag = '-v'
  ELSE
    SET @verboseFlag = ''

  -- Lookup the specified plan
  DECLARE @plan_id INT
  IF @planStart IS NULL
    SELECT @plan_id = MAX(id) FROM opc_daily_plan WHERE name = @planName
  ELSE
    SELECT @plan_id = id FROM opc_daily_plan WHERE name = @planName AND start_date = @planStart

  IF @plan_id IS NULL
  BEGIN
    DECLARE @d NVARCHAR(40)
    IF @planStart IS NULL
      SELECT @d = '(NULL)'
    ELSE
      SELECT @d = CONVERT(NVARCHAR(40), @planStart)
    RAISERROR('Could not find plan matching criteria: name=''%s'', start=''%s''', 3, 1, @planName, @d)
    RETURN -30000
  END

  IF @centralServer IS NOT NULL
  BEGIN
    IF @centralTrusted <> 0
    BEGIN
      IF @centralUser IS NOT NULL OR @centralPw IS NOT NULL
      BEGIN
        RAISERROR('Do not specify @centralUser or @centralPw with @centralTrusted', 3, 1)
        RETURN -3000
      END
    END
    ELSE
    BEGIN -- @not using a trustedconnection
      IF @centralUser IS NULL OR @centralPw is NULL
      BEGIN
        RAISERROR('If you do not use the @centralTrusted flag, you must specify both @centralUser and @centralPw', 3, 1)
        RETURN -3000
      END
    END

    IF @centralTrusted <> 0
      EXEC master..xp_asisprintf @cmd OUTPUT, 'sh TbsmOpcCopyFromCentral -s %s -t %s', @centralServer, @verboseFlag
    ELSE
      EXEC master..xp_asisprintf @cmd OUTPUT, 'sh TbsmOpcCopyFromCentral -s %s -u %s -p %s %s', @centralServer, @centralUser, 
                                              @centralPw, @verboseFlag

    PRINT @cmd
    EXEC @rc = master..xp_cmdshell @cmd
    IF @rc <> 0
    BEGIN
      RAISERROR('Central to remote pre-copy failed', 3, 1)
      RETURN -3000
    END
  END
  ELSE
  BEGIN -- @centralServer IS NULL
    TRUNCATE TABLE dyn_obj_path
    TRUNCATE TABLE dyn_obj_path_detail
    TRUNCATE TABLE opc_app_pattern
    TRUNCATE TABLE opc_op_pattern

    -- execute AS dynamic sql so we don't depend on the Object database or its contents
    EXEC("
    INSERT INTO dyn_obj_path
      SELECT * FROM Object..dyn_obj_path

    INSERT INTO dyn_obj_path_detail
      SELECT * FROM Object..dyn_obj_path_detail

    INSERT INTO opc_app_pattern
      SELECT * FROM Object..opc_app_pattern

    INSERT INTO opc_op_pattern
      SELECT * FROM Object..opc_op_pattern
    ")
  END

  -- Delete any previous records for the specified plan from the worktables
--  DELETE opc_app_path_detail WHERE plan_id = @plan_id
--  DELETE opc_app_aggr WHERE plan_id = @plan_id

  TRUNCATE TABLE opc_app_path_detail
  TRUNCATE TABLE opc_app_aggr
  TRUNCATE TABLE opc_app_new_instance

  -- Change all NULL values to blank strings in pattern-matchable fields
  UPDATE opc_application SET app_desc = '' WHERE plan_id = @plan_id AND app_desc IS NULL
  UPDATE opc_application SET app_owner = '' WHERE plan_id = @plan_id AND app_owner IS NULL
  UPDATE opc_application SET app_name = '' WHERE plan_id = @plan_id AND app_name IS NULL

  -- Clean opc_application app_owner and app_desc with the following criteria:
  -- if the path_id and app_name match then make sure that the app_owner and app_desc match.
  --   if app_owner or app_desc are not consistent the use the data values from the first based on app_id

  IF @verbose = 1
    EXEC asisp_checkpointmessage 'Cleaning application owner and description fields', 1, @chkpt
  SELECT @chkpt = @chkpt + 1
  UPDATE opc_application 
    SET 
      app_owner = a3.app_owner,
      app_desc  = a3.app_desc
    FROM (
      SELECT a1.* FROM opc_application AS a1
      INNER JOIN opc_application AS a2 ON a1.plan_id = a2.plan_id AND a1.app_name = a2.app_name
      WHERE 
            a1.plan_id = @plan_id
        AND (    a1.app_owner <> a2.app_owner
              OR a1.app_desc  <> a2.app_desc
            )
        AND (    a1.app_id    < a2.app_id
            )
      ) AS a3
    WHERE 
          opc_application.plan_id  = a3.plan_id
      AND opc_application.app_name = a3.app_name

  -- Group the applications using the opc_app_pattern table. We go through the opc_app_pattern
  -- entries in descending priority order.
  IF @verbose = 1
    EXEC asisp_checkpointmessage 'Aggregating applications', 1, @chkpt
  SELECT @chkpt = @chkpt + 1
  DECLARE patc INSENSITIVE CURSOR FOR SELECT DISTINCT priority FROM opc_app_pattern ORDER BY priority DESC
  OPEN patc
  DECLARE @pri INT
  FETCH NEXT FROM patc INTO @pri
  WHILE @@FETCH_STATUS <> -1
  BEGIN
    INSERT opc_app_aggr(plan_id, app_name, app_owner, app_desc, path_id)
      SELECT DISTINCT @plan_id, a.app_name, a.app_owner, a.app_desc, pat.path_id
      FROM
        opc_application AS a,
        opc_daily_plan  AS p,
        opc_app_pattern AS pat
      WHERE
        p.id = @plan_id
        AND pat.priority = @pri
        AND a.plan_id = @plan_id
        AND 
        (
          p.name LIKE pat.plan_name_pat
          AND a.app_desc LIKE pat.app_desc_pat
          AND a.app_name LIKE pat.app_name_pat
          AND a.app_owner LIKE pat.app_owner_pat
        )
        AND NOT EXISTS (SELECT * FROM opc_app_aggr already WHERE already.plan_id = @plan_id AND already.app_name = a.app_name)
    FETCH NEXT FROM patc INTO @pri
  END
  DEALLOCATE patc

  IF @verbose = 1
    SELECT 'FRONTEND opc_app_aggr' AS 'FRONTEND opc_app_aggr', * FROM opc_app_aggr

  -- Populate opc_app_path_detail by joining opc_app_aggr to dyn_obj_path_detail
  IF @verbose = 1
    EXEC asisp_checkpointmessage 'Populating opc_app_path_detail by joining opc_app_aggr to dyn_obj_path_detail', 1, @chkpt
  SELECT @chkpt = @chkpt + 1
  INSERT opc_app_path_detail(plan_id, app_name, path_id, gen, obj_cid, obj_name, obj_desc)
    SELECT g.plan_id, g.app_name, g.path_id, d.gen, d.obj_cid, d.obj_name, d.obj_desc
    FROM opc_app_aggr AS g JOIN dyn_obj_path_detail AS d ON g.path_id = d.path_id
    WHERE g.plan_id = @plan_id

  IF @verbose = 1
    SELECT 'FRONTEND opc_app_path_detail' AS 'FRONTEND opc_app_path_detail', * FROM opc_app_path_detail

  -- Perform substitutions in the obj_name and obj_desc fields of opc_app_path_detail using the
  -- app_name, app_owner, and app_desc columns from opc_application
  DECLARE patc CURSOR FOR
    SELECT DISTINCT g.app_name, g.app_owner, g.app_desc, d.obj_name, d.obj_desc
    FROM 
      opc_app_aggr AS g,
      opc_app_path_detail AS d
    WHERE 
      g.plan_id = @plan_id
      AND d.plan_id = @plan_id
      AND d.app_name = g.app_name
      AND (d.obj_name LIKE '%[%][A-z]%[%]%' OR d.obj_desc LIKE '%[%][A-z]%[%]%')
  OPEN patc
  -- These cannot be NVARCHAR fields in 1.1.1
  DECLARE @app_name VARCHAR(20), @app_owner VARCHAR(40), @app_desc VARCHAR(255), @obj_name VARCHAR(255), @obj_desc VARCHAR(255)
  FETCH NEXT FROM patc INTO @app_name, @app_owner, @app_desc, @obj_name, @obj_desc
  WHILE @@FETCH_STATUS <> -1
  BEGIN
    -- These cannot be NVARCHAR fields in 1.1.1
    DECLARE @resolved_obj_name VARCHAR(255), @resolved_obj_desc VARCHAR(255)
    SELECT @resolved_obj_name = @obj_name, @resolved_obj_desc = @obj_desc
    IF @obj_name LIKE '%[%][A-z]%[%]%'
    BEGIN
      IF @verbose = 1
      BEGIN
        PRINT 'Resolving name pattern [' + @obj_name + ']'
        PRINT '  app_name  = [' + @app_name + ']'
        PRINT '  app_owner = [' + @app_owner + ']'
        PRINT '  app_desc  = [' + @app_desc + ']'
      END
      EXEC master..xp_replacefields @resolved_obj_name OUTPUT, @app_name=@app_name, @app_owner=@app_owner, @app_desc=@app_desc
      IF @verbose = 1
        RAISERROR('Resolved object name pattern ''%s'' to ''%s''', 0, 1, @obj_name, @resolved_obj_name)
    END
    IF @obj_desc LIKE '%[%][A-z]%[%]%'
    BEGIN
      IF @verbose = 1
        PRINT 'Resolving description pattern [' + @obj_desc + ']'
      EXEC master..xp_replacefields @resolved_obj_desc OUTPUT, @app_name=@app_name, @app_owner=@app_owner, @app_desc=@app_desc
    END
    UPDATE opc_app_path_detail SET obj_name = @resolved_obj_name, obj_desc = @resolved_obj_desc 
      WHERE plan_id = @plan_id AND app_name = @app_name AND obj_name = @obj_name AND obj_desc = @obj_desc
    FETCH NEXT FROM patc INTO @app_name, @app_owner, @app_desc, @obj_name, @obj_desc
  END
  DEALLOCATE patc

  /*  jcp: creation of parents moved to _LoadOPCPlanBackend */

/*
  -- Allocate variable for 'discovery' time values we will set
  DECLARE @disctime DATETIME
  SELECT @disctime = GETDATE()

  -- Update the opc_app_instance.rtime column for all the opc applications in this plan that we have rediscovered
  IF @verbose = 1
  BEGIN
    SELECT @msg = 'Updating opc_app_instance.rtime column for all rediscovered applications'
    EXEC asisp_checkpointmessage @msg, 1, @chkpt
  END
  SELECT @chkpt = @chkpt + 1
  UPDATE opc_app_instance SET rtime = @disctime WHERE plan_name = @planName
*/

  UPDATE opc_app_instance SET existing = 1

  /* jcp: the following used to be part of _LoadOPCPlan_CreateBatchCycles */

  -- Now that the aggregate objects exist, we make sure the OPC applications IN this plan have been 
  -- mapped to BatchCycle instances.
  IF @verbose = 1
  BEGIN
    SELECT @msg = 'Creating new opc_app_instance records'
    EXEC asisp_checkpointmessage @msg, 1, @chkpt
  END
  SELECT @chkpt = @chkpt + 1
  INSERT opc_app_instance(plan_name, app_name, ctime)
    SELECT DISTINCT @planName, a.app_name, GETDATE() FROM opc_application a
    WHERE plan_id = @plan_id AND NOT EXISTS 
    (
      SELECT * FROM opc_app_instance i WHERE i.plan_name = @planName AND i.app_name = a.app_name
    )

  -- Any new opc_app_instance records for this plan indicate BatchCycle instances
  -- that have to be created
  IF @verbose = 1
  BEGIN
    EXEC asisp_checkpointmessage 'Determining what required applications need to be created', 1, @chkpt
    SELECT 'opc_app_instance', * FROM opc_app_instance WHERE plan_name = @planName
  END
  SELECT @chkpt = @chkpt + 1

  IF EXISTS (SELECT * FROM opc_app_instance WHERE plan_name = @planName AND existing = 0)
  BEGIN
    -- Before insert to opc_app_new_instance cleanup opc_application.app_desc so that
    -- each app_name only has one description, otherwise there will be primary key problems
    UPDATE a2
      SET
        a2.app_desc = a1.app_desc
      FROM opc_app_instance AS i
        JOIN opc_application AS a1 ON a1.app_name = i.app_name
        JOIN opc_application AS a2 ON a2.app_name = i.app_name
      WHERE
        a1.plan_id = @plan_id
        AND i.plan_name = @planName
/*        AND i.btcy_id IS NULL */
        AND i.existing = 0
        AND a1.app_desc <> a2.app_desc
        AND a1.app_id   <  a2.app_id

    INSERT INTO opc_app_new_instance
      SELECT DISTINCT a.app_name, ISNULL(a.app_desc,'')
      FROM opc_app_instance AS i 
        JOIN opc_application AS a ON a.app_name = i.app_name
      WHERE 
        a.plan_id = @plan_id 
        AND i.plan_name = @planName
        AND i.existing = 0
  END

  /* jcp: the preceding used to be part of _LoadOPCPlan_CreateBatchCycles */

  -- All required OPC applications from the plan have now been mapped to BTCY instances.
  -- We now do the same for the links.

  -- opc_app_path_aggr holds the records from opc_app_path_detail for the objects
  -- that the OPC applications are to be aggregated into.
  DELETE FROM opc_app_path_aggr WHERE plan_id = @plan_id
  INSERT INTO opc_app_path_aggr (plan_id, app_name, path_id, gen, obj_cid, obj_name)
  SELECT DISTINCT
    plan_id,
    app_name,
    path_id,
    gen,
    obj_cid,
    obj_name
  FROM opc_app_path_detail d
  WHERE
    d.plan_id = @plan_id 
    AND d.gen = 
    (
      SELECT MAX(m.gen) 
      FROM opc_app_path_detail m 
      WHERE m.plan_id = @plan_id AND m.path_id = d.path_id AND m.app_name = d.app_name
    )

  IF @verbose = 1
    SELECT 'opc_app_path_aggr', * FROM opc_app_path_aggr

  /* containment record logic moved to _LoadOPCPlanBackend */

  /* LOB logic moved to _LoadOPCPlanBackend */

  -- Calculate the estimated start/stop times for the applications. The daily plan specifies
  -- input arrival DAY, which is the DAY of the MONTH corresponding to the start date of the
  -- planning period. There are boundary conditions surrounding the beginning/end of MONTH
  -- that have to be taken into account.
  IF @verbose = 1
  BEGIN
    EXEC asisp_checkpointmessage 'Calculating estimated start/stop times for applications', 1, @chkpt
    SELECT @chkpt = @chkpt + 1
  END

  UPDATE a SET
    a.est_start = CONVERT(NVARCHAR(10),p.start_date,102),
    a.est_stop  = CONVERT(NVARCHAR(10),p.start_date,102) -- yes, p.start_date, *NOT* p.stop_date
  FROM
    opc_application a,
    opc_daily_plan p
  WHERE
    a.plan_id = @plan_id
    AND p.id = @plan_id

  DECLARE @plan_start_dom INT
  SELECT @plan_start_dom = DATEPART(MONTH, start_date) FROM opc_daily_plan WHERE id = @plan_id

  -- Catch any start/stop hour fields set to 2400. Change these to 2359.
  UPDATE opc_application SET start_hour = 2359 WHERE start_hour = 2400
  UPDATE opc_application SET stop_hour = 2359 WHERE stop_hour = 2400

  -- Change the estimated start date of applications whose start_day value differs
  -- from that of the planning period within the same MONTH (AS specified by 
  -- opc_application.start_day_mod=' ').
  IF @verbose = 1
  BEGIN
    EXEC asisp_checkpointmessage 'Calculating estimated start times for applications within plan MONTH', 1, @chkpt
    SELECT @chkpt = @chkpt + 1
  END
  UPDATE a SET
    a.est_start =
      CONVERT(NVARCHAR(10), DATEADD(DAY, start_day - DATEPART(DAY,est_start), est_start), 102) +
      ' ' +
      REPLICATE('0', 2-LEN(CONVERT(NVARCHAR(4),start_hour/100))) + CONVERT(NVARCHAR(4),start_hour/100) +
      ':' +
      REPLICATE('0', 2-LEN(CONVERT(NVARCHAR(4),start_hour%100))) + CONVERT(NVARCHAR(4),start_hour%100)
  FROM
    opc_application a
  WHERE
    a.plan_id = @plan_id
    AND a.start_day_mod = 0
    AND a.start_day <> @plan_start_dom

  -- Change the estimated start date of applications whose start_day_mod value is
  -- nonzero indicating that the application's start date is in the previous
  -- or next MONTH from the planning period's.
  IF @verbose = 1
  BEGIN
    EXEC asisp_checkpointmessage 'Calculating estimated start times for applications outside of plan MONTH', 1, @chkpt
    SELECT @chkpt = @chkpt + 1
  END
  UPDATE a SET
    a.est_start = 
      CONVERT(DATETIME,
        CONVERT(NVARCHAR(4), DATEPART(YEAR, DATEADD(MONTH, a.start_day_mod, a.est_start))) + '-' +
        CONVERT(NVARCHAR(2), DATEPART(MONTH, DATEADD(MONTH, a.start_day_mod, a.est_start))) + '-' +
        CONVERT(NVARCHAR(2), a.start_day)
      )
  FROM
    opc_application a
  WHERE
    a.plan_id = @plan_id
    AND a.start_day_mod <> 0

  -- Change the estimated stop date of applications whose stop_day value differs
  -- from that of the planning period's start date within the same MONTH (AS specified by 
  -- opc_application.stop_day_mod=' ').
  IF @verbose = 1
  BEGIN
    EXEC asisp_checkpointmessage 'Calculating estimated completion times for applications within plan MONTH', 1, @chkpt
    SELECT @chkpt = @chkpt + 1
  END
  UPDATE a SET
    a.est_stop =
      CONVERT(NVARCHAR(10), DATEADD(DAY, stop_day - DATEPART(DAY,est_stop), est_stop), 102) +
      ' ' +
      REPLICATE('0', 2-LEN(CONVERT(NVARCHAR(4),stop_hour/100))) + CONVERT(NVARCHAR(4),stop_hour/100) +
      ':' +
      REPLICATE('0', 2-LEN(CONVERT(NVARCHAR(4),stop_hour%100))) + CONVERT(NVARCHAR(4),stop_hour%100)
  FROM
    opc_application a
  WHERE
    a.plan_id = @plan_id
    AND a.stop_day_mod = 0
    AND a.stop_day <> @plan_start_dom

  -- Change the estimated stop date of applications whose stop_day_mod value is
  -- nonzero indicating that the application's stop date is in the previous
  -- or next MONTH from the planning period's start date.
  IF @verbose = 1
  BEGIN
    EXEC asisp_checkpointmessage 'Calculating estimated completion times for applications outside of plan MONTH', 1, @chkpt
    SELECT @chkpt = @chkpt + 1
  END
  UPDATE a SET
    a.est_stop = 
      CONVERT(DATETIME,
        CONVERT(NVARCHAR(4), DATEPART(YEAR, DATEADD(MONTH, a.stop_day_mod, a.est_stop))) + '-' +
        CONVERT(NVARCHAR(2), DATEPART(MONTH, DATEADD(MONTH, a.stop_day_mod, a.est_stop))) + '-' +
        CONVERT(NVARCHAR(2), a.stop_day)
      )
  FROM
    opc_application a
  WHERE
    a.plan_id = @plan_id
    AND a.stop_day_mod <> 0

  -- Calculate the estimated start/stop times for jobs. These times come from the 'input arrival time'
  -- field of the operation in the daily plan report. Since this value is inherited from the application
  -- unless the operation is defined as 'time dependent' and the daily plan does not tell us
  -- whether a given operation is time dependent, we can use these times only as indicators of the
  -- EARLIEST expected start times of the jobs unless they differ from the application's start time.

  -- Catch any start/stop hour fields SET to 2400. Change these to 2359.
  UPDATE opc_operation SET start_hour = 2359 WHERE start_hour = 2400
  UPDATE opc_operation SET stop_hour = 2359 WHERE stop_hour = 2400

  UPDATE opc_operation SET
    est_start = CONVERT(NVARCHAR(10),p.start_date,102)
  FROM
    opc_daily_plan p
  WHERE
    plan_id = @plan_id
    AND p.id = @plan_id

  IF @verbose = 1
    SELECT 'opc_operation after est_start update', * FROM opc_operation WHERE plan_id = @plan_id

  -- Change the estimated start date of operations whose start_day value differs
  -- from that of the planning period within the same month (as specified by 
  -- opc_operation.start_day_mod=' ').
  IF @verbose = 1
  BEGIN
    EXEC asisp_checkpointmessage 'Calculating estimated start times for operations within plan month', 1, @chkpt
    SELECT @chkpt = @chkpt + 1
  END
  IF @verbose = 1
    SELECT @chkpt AS chkpt,
      o.start_day,
      o.est_start,
      o.start_hour,
      new_est_start =
        CONVERT(NVARCHAR(10), DATEADD(DAY, start_day - DATEPART(DAY,est_start), est_start), 102) +
        ' ' +
        REPLICATE('0', 2-LEN(CONVERT(NVARCHAR(4),start_hour/100))) + CONVERT(NVARCHAR(4),start_hour/100) +
        ':' +
        REPLICATE('0', 2-LEN(CONVERT(NVARCHAR(4),start_hour%100))) + CONVERT(NVARCHAR(4),start_hour%100)
    FROM
      opc_operation o
    WHERE
      o.plan_id = @plan_id
      AND o.start_day_mod = 0
--      AND o.start_day <> @plan_start_dom
 
  UPDATE o SET
    o.est_start =
      CONVERT(NVARCHAR(10), DATEADD(DAY, start_day - DATEPART(DAY,est_start), est_start), 102) +
      ' ' +
      REPLICATE('0', 2-LEN(CONVERT(NVARCHAR(4),start_hour/100))) + CONVERT(NVARCHAR(4),start_hour/100) +
      ':' +
      REPLICATE('0', 2-LEN(CONVERT(NVARCHAR(4),start_hour%100))) + CONVERT(NVARCHAR(4),start_hour%100)
  FROM
    opc_operation o
  WHERE
    o.plan_id = @plan_id
    AND o.start_day_mod = 0
--    AND o.start_day <> @plan_start_dom
  
  -- Change the estimated start date of operations whose start_day_mod value is
  -- nonzero indicating that the operation's start date is in the previous
  -- or next month from the planning period's.
  IF @verbose = 1
  BEGIN
    EXEC asisp_checkpointmessage 'Calculating estimated start times for operations outside of plan month', 1, @chkpt
    SELECT @chkpt = @chkpt + 1
  END
  IF @verbose = 1
    SELECT @chkpt AS chkpt,
      o.start_day_mod,
      o.est_start,
          CONVERT(NVARCHAR(4), DATEPART(YEAR, DATEADD(MONTH, o.start_day_mod, o.est_start))) + '-' +
          CONVERT(NVARCHAR(2), DATEPART(MONTH, DATEADD(MONTH, o.start_day_mod, o.est_start))) + '-' +
          CONVERT(NVARCHAR(2), o.start_day) + ' ' +
          REPLICATE('0', 2-LEN(CONVERT(NVARCHAR(4),o.start_hour/100))) + CONVERT(NVARCHAR(4),o.start_hour/100) +
          ':' +
          REPLICATE('0', 2-LEN(CONVERT(NVARCHAR(4),o.start_hour%100))) + CONVERT(NVARCHAR(4),o.start_hour%100)
      AS new_est_start
    FROM
      opc_operation o
    WHERE
      o.plan_id = @plan_id
      AND o.start_day_mod <> 0

  UPDATE o SET
    o.est_start = 
      CONVERT(DATETIME,
        CONVERT(NVARCHAR(4), DATEPART(YEAR, DATEADD(MONTH, o.start_day_mod, o.est_start))) + '-' +
        CONVERT(NVARCHAR(2), DATEPART(MONTH, DATEADD(MONTH, o.start_day_mod, o.est_start))) + '-' +
        CONVERT(NVARCHAR(2), o.start_day) + ' ' +
        REPLICATE('0', 2-LEN(CONVERT(NVARCHAR(4),o.start_hour/100))) + CONVERT(NVARCHAR(4),o.start_hour/100) +
        ':' +
        REPLICATE('0', 2-LEN(CONVERT(NVARCHAR(4),o.start_hour%100))) + CONVERT(NVARCHAR(4),o.start_hour%100)
      )
  FROM
    opc_operation o
  WHERE
    o.plan_id = @plan_id
    AND o.start_day_mod <> 0

  -- Change the estimated stop date of operations that have a non-null stop day/hour
  IF @verbose = 1
  BEGIN
    EXEC asisp_checkpointmessage 'Calculating estimated completion times for operations within plan month', 1, @chkpt
    SELECT @chkpt = @chkpt + 1
  END
  UPDATE o SET
    o.est_stop = CONVERT(NVARCHAR(11), DATEADD(DAY, CASE WHEN stop_day=start_day THEN 0 ELSE 1 END, est_start), 120) +
                 REPLICATE('0', 2-LEN(CONVERT(NVARCHAR(4),o.stop_hour/100))) + CONVERT(NVARCHAR(4),o.stop_hour/100) +
                 ':' +
                 REPLICATE('0', 2-LEN(CONVERT(NVARCHAR(4),o.stop_hour%100))) + CONVERT(NVARCHAR(4),o.stop_hour%100)
  FROM opc_operation AS o
  WHERE
    o.plan_id = @plan_id
    AND o.stop_day IS NOT NULL
    AND o.stop_hour IS NOT NULL

  -- Discard any operations that are completely outside of the planning period we're interested in
  IF @verbose = 1
  BEGIN
    EXEC asisp_checkpointmessage 'Deleting operations outside of planning period', 1, @chkpt
    SELECT @chkpt = @chkpt + 1
  END
  IF @verbose = 1
    SELECT 'About to delete', job_name,est_start,est_stop,start_day_mod,start_day,start_hour,duration,stop_day,stop_hour
    FROM opc_operation AS o
      JOIN opc_daily_plan AS p ON p.id = o.plan_id
    WHERE
      p.id = @plan_id
     AND ( (o.est_start > p.end_date ) OR ( o.est_stop < p.start_date ) )

  DELETE FROM op FROM opc_predecessor AS op 
    JOIN opc_operation AS o ON o.plan_id = op.plan_id AND o.op_id = op.succ_op_id
    JOIN opc_daily_plan AS p ON p.id = o.plan_id
  WHERE
    p.id = @plan_id
    AND ((o.est_start > p.end_date) OR (o.est_stop < p.start_date))

  IF @verbose = 1
    SELECT 'Operations before frontend delete = ' + CAST(COUNT(*) AS NVARCHAR) FROM opc_operation WHERE plan_id = @plan_id

  DELETE FROM o FROM opc_operation AS o
    JOIN opc_daily_plan AS p ON p.id = o.plan_id
  WHERE
    p.id = @plan_id
    AND ((o.est_start > p.end_date) OR (o.est_stop < p.start_date))

  IF @verbose = 1
    SELECT 'Operations after frontend delete = ' + CAST(COUNT(*) AS NVARCHAR) FROM opc_operation WHERE plan_id = @plan_id

  IF @centralServer IS NOT NULL
  BEGIN
    IF @centralTrusted <> 0
      EXEC master..xp_asisprintf @cmd OUTPUT, 'sh TbsmOpcCopyToCentral -s %s -t -i %d %s', @centralServer, @plan_id, @verboseFlag
    ELSE
      EXEC master..xp_asisprintf @cmd OUTPUT, 
           'sh TbsmOpcCopyToCentral -s %s -u %s -p %s -i %d %s', @centralServer, @centralUser, @centralPw, @plan_id, @verboseFlag

    PRINT @cmd
    EXEC @rc = master..xp_cmdshell @cmd
    IF @rc <> 0
    BEGIN
      RAISERROR('Remote to central mid-copy failed', 3, 1)
      RETURN -3000
    END
  END
  ELSE
  BEGIN
    -- execute AS dynamic sql so we don't depend on the Object database or its contents
    DECLARE @plan_id_str NVARCHAR(9), @verbose_str NVARCHAR(1)
    SET @plan_id_str = CAST(@plan_id AS NVARCHAR)
    SET @verbose_str = CAST(@verbose AS NVARCHAR)
    EXEC("
    DECLARE @plan_id INT, @verbose TINYINT
    SET @plan_id = " + @plan_id_str + "
    SET @verbose = " + @verbose_str + "
    IF @verbose = 1 PRINT 'Repopulating Object..opc_app_new_instance'
    TRUNCATE TABLE Object..opc_app_new_instance
    INSERT INTO Object..opc_app_new_instance
      SELECT * FROM OPCLoad..opc_app_new_instance
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('Error repopulating Object..opc_app_new_instance', 3, 1)
      GOTO AbortLoad
    END

    IF @verbose = 1 PRINT 'Repopulating Object..opc_app_path_aggr'
    TRUNCATE TABLE Object..opc_app_path_aggr
    INSERT INTO Object..opc_app_path_aggr
      SELECT * FROM OPCLoad..opc_app_path_aggr
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('Error repopulating Object..opc_app_path_aggr', 3, 1)
      GOTO AbortLoad
    END

    IF @verbose = 1 PRINT 'Repopulating Object..opc_app_path_detail'
    TRUNCATE TABLE Object..opc_app_path_detail
    INSERT INTO Object..opc_app_path_detail
      SELECT * FROM OPCLoad..opc_app_path_detail
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('Error repopulating Object..opc_app_path_detail', 3, 1)
      GOTO AbortLoad
    END

    IF @verbose = 1 PRINT 'Repopulating Object..opc_application'
    TRUNCATE TABLE Object..opc_application
    INSERT INTO Object..opc_application
      SELECT * FROM OPCLoad..opc_application WHERE plan_id = @plan_id
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('Error repopulating Object..opc_application', 3, 1)
      GOTO AbortLoad
    END

    IF @verbose = 1 PRINT 'Repopulating Object..opc_daily_plan'
    TRUNCATE TABLE Object..opc_daily_plan
    INSERT INTO Object..opc_daily_plan
      SELECT * FROM OPCLoad..opc_daily_plan WHERE id = @plan_id
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('Error repopulating Object..opc_daily_plan', 3, 1)
      GOTO AbortLoad
    END

    IF @verbose = 1 PRINT 'Repopulating Object..opc_operation'
    TRUNCATE TABLE Object..opc_operation
    INSERT INTO Object..opc_operation
      SELECT * FROM OPCLoad..opc_operation WHERE plan_id = @plan_id
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('Error repopulating Object..opc_operation', 3, 1)
      GOTO AbortLoad
    END

    GOTO ContinueLoad
AbortLoad:
    RAISERROR(58004, 11, 1)
ContinueLoad:
    ")
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('Aborting load', 11, 1)
      RETURN -30000
    END
  END

  IF @verbose = 1
  BEGIN
    SELECT @msg = 'OPC frontend finished with plan ' + @planName + ', kicking off backend'
    EXEC asisp_checkpointmessage @msg, 1, @chkpt
  END
  SET @chkpt = @chkpt + 1

  IF @centralServer IS NOT NULL
  BEGIN
    SET @cmd = 'EXEC ' + UPPER(@centralServer) + '.Object.._LoadOPCPlanBackend  ''' + @planName + ''', ' + CAST(@plan_id AS NVARCHAR) + ', @verbose=' + CAST(@verbose AS NVARCHAR)
    IF @verbose = 1
      PRINT @cmd
    EXEC(@cmd)
  END
  ELSE
  BEGIN
    SET @cmd = 'EXEC Object.._LoadOPCPlanBackend  ''' + @planName + ''', ' + CAST(@plan_id AS NVARCHAR) + ', @verbose=' + CAST(@verbose AS NVARCHAR)
    IF @verbose = 1
      PRINT @cmd
    EXEC(@cmd)
  END
GO
