-- File: archive/db/sql/_loadopcplanbackend.sql
-- SID: 1.3
-- Date Modified: 01/11/09 12:41:04
--
-- Module Description
--   Defines the _LoadOPCPlanBackend stored procedure

USE Object
GO

---------------------------------------------------------------------------------------------------
-- _LoadOPCPlanBackend
PRINT 'Creating procedure _LoadOPCPlanBackend'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = '_LoadOPCPlanBackend' AND type = 'P')
  DROP PROCEDURE _LoadOPCPlanBackend
GO

-- File: archive/db/sql/_loadopcplanbackend.sql
-- SID: 1.3
-- Date Modified: 01/11/09 12:41:04
-- Implements the portion of the OPC discovery process that must run on the central Object database.
-- This is called from the _LoadOPCPlan stored procedure that is running in the OPCLoad database,
-- either locally or remotely.
CREATE PROCEDURE _LoadOPCPlanBackend
  @planName   NVARCHAR(40),
  @plan_id    INT,
  @verbose    TINYINT = 0
AS
  DECLARE @chkpt INT, @msg NVARCHAR(255)
  DECLARE @r INT

  SET @chkpt = 0

  SET @msg = 'OPC backend processing plan ' + @planName + ' (' + CAST(@plan_id AS NVARCHAR) + ')'
  EXEC asisp_checkpointmessage @msg, 1, @chkpt
  SELECT @chkpt = @chkpt + 1

  -- opc_app_path_detail now contains resolved values for all object names and descriptions, so
  -- we now traverse through the distinct values at each level of the hierarchy and lookup/create
  -- the objects.

  DECLARE @gen INT, @max_gen INT
  SELECT @max_gen = MAX(gen) FROM opc_app_path_detail WHERE plan_id = @plan_id
  SELECT @gen = 1
  WHILE @gen <= @max_gen
  BEGIN
    IF @verbose = 1
    BEGIN
      SELECT @msg = 'Resolving object names and paths at generation ' + CONVERT(NVARCHAR(9), @gen)
      EXEC Object..asisp_checkpointmessage @msg, 1, @chkpt
    END
    SELECT @chkpt = @chkpt + 1
    DECLARE objc CURSOR FOR
      SELECT DISTINCT
        parent.path_id                        AS path_id,
        parent.obj_cid                        AS parent_cid,
        CONVERT(NVARCHAR(9), parent.obj_id)   AS parent_id,
        child.obj_cid                         AS child_cid,
        CONVERT(NVARCHAR(54), child.obj_name) AS child_name,
        child.obj_desc                        AS child_desc,
        child_class.cname                     AS child_cname,
        child_class.flag_direct_containment   AS child_direct
      FROM
        opc_app_path_detail AS parent,
        opc_app_path_detail AS child
        JOIN obj_class AS child_class ON child_class.cid = child.obj_cid
      WHERE
        parent.plan_id = @plan_id
        AND child.plan_id = @plan_id
        AND parent.app_name = child.app_name
        AND parent.path_id = child.path_id
        AND parent.gen = @gen -1
        AND child.gen = @gen
    OPEN objc
    DECLARE
      @path_id      INT,
      @parent_cid   ClassID, 
      @parent_id    NVARCHAR(9),
      @child_cid    ClassID, 
      @child_name   ObjName, 
      @child_desc   NVARCHAR(255), 
      @child_cname  NVARCHAR(24),
      @child_id     ObjID, 
      @child_direct BIT
    FETCH NEXT FROM objc INTO @path_id, @parent_cid, @parent_id, @child_cid, @child_name, @child_desc, @child_cname, @child_direct
    WHILE @@FETCH_STATUS <> -1
    BEGIN
      DECLARE @sql NVARCHAR(255)
      IF @verbose = 1
        PRINT 'child_name=[' + @child_name + '] child_desc=[' +  @child_desc + ']'
      IF @child_direct = 1
        SELECT @sql = 
          "SELECT o.id FROM " + @child_cname + "_V o " +
          "WHERE o.name = '" + @child_name + "' " + 
          "AND o._ParentCID = '" + @parent_cid + "' " +
          "AND o._ParentID = " + @parent_id + " " +
          "AND o.deleted = 0"
      ELSE
        SELECT @sql =
          "SELECT o.id FROM containment AS c JOIN " + @child_cname + "_V AS o ON " +
          "c.dst_cid = '" + @child_cid + "' " +
          "AND c.dst_id = o.id " + 
          "WHERE o.name = '" + @child_name + "' " + 
          "AND c.src_cid = '" + @parent_cid + "' " + 
          "AND c.src_id = " + @parent_id + " " + 
          "AND c.deleted = 0"
      IF @verbose = 1
      BEGIN
        PRINT @sql
        EXEC(@sql)
      END
      EXEC("
        DECLARE lookupc INSENSITIVE CURSOR FOR " + @sql
        )
      OPEN lookupc
      SELECT @child_id = NULL
      FETCH NEXT FROM lookupc INTO @child_id
      DEALLOCATE lookupc
      IF @child_id IS NULL
      BEGIN
     -- Create object
        IF @verbose = 1
          RAISERROR('Creating %s ''%s'' as child of %s %s', 0, 1, @child_cid, @child_name, @parent_cid, @parent_id)
        EXEC("
        DECLARE @id ObjID
        EXEC @id = asisp_create" + @child_cid + " '" + @child_name + "', '" + @parent_cid + "', " + @parent_id + ", @Desc='" + @child_desc + "'
        DECLARE lookupc INSENSITIVE CURSOR FOR SELECT @id
        ")
        OPEN lookupc
        FETCH NEXT FROM lookupc INTO @child_id
        DEALLOCATE lookupc
      END

      -- Update the object ids for all rows in opc_app_path_detail that match the looked-up/created object.
      -- note that this cannot be done with a where clause of the form "WHERE CURRENT OF objc" because the
      -- objc cursor was declared with a distinct select statement so that we would only do the lookup/create
      -- process once per object in a given path.
      IF @child_id IS NOT NULL BEGIN
        UPDATE child SET obj_id = @child_id 
        FROM
          opc_app_path_detail parent,
          opc_app_path_detail child
        WHERE
          parent.plan_id = @plan_id
          AND parent.path_id = @path_id
          AND child.plan_id = @plan_id
          AND parent.path_id = child.path_id
          AND parent.app_name = child.app_name
          AND parent.gen = @gen - 1
          AND child.gen = @gen
          AND child.obj_name = @child_name

        UPDATE aggr SET obj_id = @child_id 
        FROM
          opc_app_path_detail parent,
          opc_app_path_aggr aggr
        WHERE
          parent.plan_id = @plan_id
          AND parent.path_id = @path_id
          AND aggr.plan_id = @plan_id
          AND parent.path_id = aggr.path_id
          AND parent.app_name = aggr.app_name
          AND parent.gen = @gen - 1
          AND aggr.gen = @gen
          AND aggr.obj_name = @child_name
      END
              
      FETCH NEXT FROM objc INTO @path_id, @parent_cid, @parent_id, @child_cid, @child_name, @child_desc, @child_cname, @child_direct
    END
    DEALLOCATE objc

    SELECT @gen = @gen + 1
  END

  -- create new BatchCycle objects (aka, applications)

  INSERT INTO opc_app_instance (plan_name, app_name)
    SELECT @planName, btcy_name FROM opc_app_new_instance

  DECLARE @nbtcy INT
  SELECT @nbtcy = COUNT(*) FROM opc_app_new_instance
  IF @nbtcy > 0 BEGIN
    IF @verbose = 1
    BEGIN
      SELECT @msg = 'Creating ' + CONVERT(NVARCHAR(9),@nbtcy) + ' BatchCycle object(s)'
      SELECT 'opc_app_new_instance', * FROM opc_app_new_instance
      EXEC Object..asisp_checkpointmessage @msg, 1, @chkpt
    END
    SELECT @chkpt = @chkpt + 1

    -- Allocate the appropriate number of new BTCY records
    DECLARE @base_id INT
    EXEC @base_id = alloc_BTCY @nbtcy

    -- INSERT the class records
    INSERT BatchCycle_C(id,name) SELECT @base_id + row, btcy_name FROM opc_app_new_instance
    IF @@ERROR <> 0
    BEGIN
--      ROLLBACK TRAN _LoadOPCPlan_CreateBatchCycles
      RETURN -30000
    END

    -- INSERT the attribute records
    INSERT BatchCycle_A(id,_Desc,_ChildEventMatrixID,_ExceptionMatrixID,_PriorityID)
      SELECT DISTINCT @base_id+row, isnull(btcy_desc,''), 0, 0, 6
      FROM opc_app_new_instance
    IF @@ERROR <> 0
    BEGIN
--      ROLLBACK TRAN _LoadOPCPlan_CreateBatchCycles
      RETURN -30000
    END

    -- UPDATE opc_app_instance to indicate the new ids
    UPDATE opc_app_instance SET btcy_id = @base_id+b.row
      FROM opc_app_new_instance b
      WHERE plan_name = @planName AND app_name = btcy_name AND btcy_id IS NULL
    IF @@ERROR <> 0
    BEGIN
--      ROLLBACK TRAN _LoadOPCPlan_CreateBatchCycles
      RETURN -30000
    END
  END

  IF @verbose = 1
  BEGIN
    SELECT 'opc_app_instance'    AS 'opc_app_instance   ', * FROM opc_app_instance WHERE plan_name = @planName
    SELECT 'opc_app_path_aggr'   AS 'opc_app_path_aggr  ', * FROM opc_app_path_aggr
    SELECT 'opc_app_path_detail' AS 'opc_app_path_detail', * FROM opc_app_path_detail
  END

  -- We have to update existing containment records for batch cycles that are 
  -- moving and create containment records for any instances that are new.
  SELECT
    c.src_id  AS old_src_id,
    c.deleted AS deleted,
    a.obj_id  AS new_src_id,
    i.btcy_id AS dst_id
  INTO #new_bcys_btcy
  FROM opc_app_path_aggr a 
    JOIN opc_app_instance i ON i.app_name = a.app_name
    LEFT OUTER JOIN containment AS c ON c.src_cid = 'BCYS' AND c.dst_cid = 'BTCY' AND c.dst_id = i.btcy_id
  WHERE
    a.obj_cid = 'BCYS'
    AND i.plan_name = @planName

  IF @verbose = 1
    SELECT '#new_bcys_btcy', * FROM #new_bcys_btcy

  -- Update previous records
  UPDATE old SET src_id = new.new_src_id
    FROM BCYS_BTCY_AGGC AS old JOIN #new_bcys_btcy AS new ON new.dst_id = old.dst_id
    WHERE new.old_src_id IS NOT NULL AND new.new_src_id <> new.old_src_id

  -- Create new records
  INSERT INTO BCYS_BTCY_AGGC(src_id, dst_id) 
    SELECT DISTINCT new_src_id, dst_id
    FROM #new_bcys_btcy
    WHERE old_src_id IS NULL AND new_src_id IS NOT NULL

  IF @@ROWCOUNT > 0
  BEGIN
    -- Insert the create notifications
    INSERT PendingNotification(type, fields, detail)
      SELECT
        EventType.type,
        CONVERT(NVARCHAR(5), BTCY.cno) + '|' + CONVERT(NVARCHAR(9), l.dst_id) + '|BTCY|AGGC|' +
        CONVERT(NVARCHAR(5), BCYS.cno) + '|' + CONVERT(NVARCHAR(9), l.new_src_id) AS fields,
        i.app_name AS detail
      FROM
        #new_bcys_btcy l JOIN opc_app_instance i ON i.btcy_id = l.dst_id,
        EventType,
        obj_class AS BTCY,
        obj_class AS BCYS
      WHERE
        l.old_src_id IS NULL
        AND EventType.name = 'ObjectCreated'
        AND BTCY.cid = 'BTCY'
        AND BCYS.cid = 'BCYS'
  END

  -- Allocate variable for 'discovery' time values we will set
  DECLARE @disctime DATETIME
  SELECT @disctime = GETDATE()

  -- Update the opc_app_link.rtime column for all rediscovered links
  IF @verbose = 1
  BEGIN
    SELECT @msg = 'Updating opc_app_link.rtime for rediscovered links'
    EXEC Object..asisp_checkpointmessage @msg, 1, @chkpt
  END
  SELECT @chkpt = @chkpt + 1
  UPDATE l SET rtime = @disctime 
    FROM opc_app_link l, opc_app_path_aggr a, opc_app_instance i
    WHERE
      i.plan_name = @planName
      AND a.app_name = i.app_name
      AND a.obj_cid = 'BCYS'
      AND a.obj_id = l.src_id
      AND i.btcy_id = l.btcy_id
  
  -- Insert newly discovered opc_app_link records that are needed.
  IF @verbose = 1
  BEGIN
    SELECT @msg = 'Inserting newly discovered opc_app_link records'
    EXEC Object..asisp_checkpointmessage @msg, 1, @chkpt
  END
  SELECT @chkpt = @chkpt + 1
  INSERT opc_app_link(plan_name, app_name, link_type, src_cid, src_id, btcy_cid, btcy_id)
    SELECT i.plan_name, i.app_name, 'AGGC', 'BCYS', l.new_src_id, 'BTCY', l.dst_id
    FROM opc_app_path_aggr AS a
      JOIN opc_app_instance AS i ON i.app_name = a.app_name
      JOIN #new_bcys_btcy AS l ON l.new_src_id = a.obj_id AND l.dst_id = i.btcy_id
    WHERE
      i.plan_name = @planName
      AND a.obj_cid = 'BCYS'
      AND l.old_src_id IS NULL

  -- For all the aggregation points that are LOBs, we have to create LOBs under those
  -- points that are linked to the BTCY instances.
  CREATE TABLE #lob_link
  (
    src_id    ObjID,
    dst_id    ObjID,
    btcy_id   ObjID,
    btcy_name ObjName
  )
  INSERT #lob_link
    SELECT a.obj_id, 0, i.btcy_id, a.app_name
    FROM opc_app_path_aggr a, opc_app_instance i
    WHERE i.plan_name = @planName AND i.app_name = a.app_name AND a.obj_cid = 'LOB'

  -- Lookup the ids for any previously existing LOB links and set the #lob_link.dst_id
  -- to indicate these.
  IF @verbose = 1
  BEGIN
    SELECT @msg = 'Looking up previously existing LOB links'
    EXEC Object..asisp_checkpointmessage @msg, 1, @chkpt
  END
  SELECT @chkpt = @chkpt + 1
  UPDATE l SET dst_id = c.dst_id
  FROM
    #lob_link l,
    containment c,
    LineOfBusiness_V lob
  WHERE
    c.src_cid = 'LOB'
    AND c.src_id = l.src_id
    AND c.dst_cid = 'LOB'
    AND lob.id = c.dst_id
    AND lob.name = l.btcy_name

  -- Update the opc_app_link.rtime column for all rediscovered links
  IF @@ROWCOUNT > 0
    UPDATE a SET rtime = @disctime 
    FROM opc_app_link a, #lob_link l
    WHERE
      l.dst_id <> 0
      AND a.plan_name = @planName
      AND a.app_name = l.btcy_name
      AND a.src_cid = 'LOB' 
      AND a.src_id = l.dst_id -- For LOBs, the source in opc_app_link is the LOB that has a direct
                              -- reference to the BTCY, not the containing LOB.

  -- Any #lob_link rows with dst_id=0 now indicate LOBs that need to be created AND
  -- linked to the BTCY instances.
  IF EXISTS (SELECT * FROM #lob_link WHERE dst_id = 0)
  BEGIN
    -- Create a temp table to hold all rows for the LOBs that need to be created
    CREATE TABLE #new_lob
    (
      src_id    ObjID,    -- containing LOB - needed to map back to #lob_link
      btcy_id   ObjID,
      btcy_name ObjName,
      btcy_desc NVARCHAR(255),
      row       INT IDENTITY(0,1)
    )
    DECLARE @nlob INT
    INSERT #new_lob 
      SELECT DISTINCT l.src_id, l.btcy_id, l.btcy_name, ISNULL(a.app_desc, '')
      FROM #lob_link l, opc_application a
      WHERE 
        l.dst_id = 0
        AND a.plan_id = @plan_id AND a.app_name = l.btcy_name

    EXEC @r = _LoadOPCPlan_CreateLOBs @planName, @chkpt OUT
    IF @r < 0
      RETURN @r
  END

  -- Update the btcy_id of the operations
  IF @verbose = 1
  BEGIN
    EXEC Object..asisp_checkpointmessage 'Updating opc_operation.btcy_id', 1, @chkpt
    SET @chkpt = @chkpt + 1
  END
  UPDATE o SET
    btcy_id = i.btcy_id
  FROM
    opc_operation AS o 
      JOIN opc_daily_plan AS p ON p.id = o.plan_id
      JOIN opc_app_instance AS i ON i.plan_name = p.name AND i.app_name = o.app_name
  WHERE
    o.plan_id = @plan_id

  EXEC @r = _LoadOPCPlan_CreateBCJs @plan_id, @chkpt OUTPUT, @verbose
  IF @r < 0
    RETURN @r

  -- Load the new plan state to the rule engine
  -- EXEC _SendOPCPlanFacts @plan_id

  SET @msg = 'OPC backend done with plan ' + @planName + ' (' + CAST(@plan_id AS NVARCHAR) + ')'

  EXEC asisp_checkpointmessage @msg, 1, @chkpt
  SELECT @chkpt = @chkpt + 1
GO
