-- $Header:   //ARCHIVE/Archive_NLS/DB/SQL/_LoadOPCPlan_CreateBCJs.sqv   1.3   Aug 09 2001 10:43:58   BRENT  $
--
-- _LoadOPCPlan_CreateBCJs.sql
--
-- Module Description
--   Defines the _LoadOPCPlan_CreateBCJs stored procedure


USE Object
GO

---------------------------------------------------------------------------------------------------
-- _LoadOPCPlan_CreateBCJs
PRINT 'Creating procedure _LoadOPCPlan_CreateBCJs'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = '_LoadOPCPlan_CreateBCJs' AND type = 'P')
  DROP PROCEDURE _LoadOPCPlan_CreateBCJs
GO

-- $Archive:   //ARCHIVE/Archive_NLS/DB/SQL/_LoadOPCPlan_CreateBCJs.sqv  $
-- $Revision:   1.3  $
-- $Date:   Aug 09 2001 10:43:58  $
-- $Author:   BRENT  $
-- Supporting procedure for _LoadOPCPlan that updates/creates the required batch job objects
CREATE PROCEDURE _LoadOPCPlan_CreateBCJs
  @plan_id    INT,
  @chkpt      INT OUTPUT,
  @verbose    TINYINT = 0
AS
  DECLARE @err INT, @rc INT

  -- Check assertions
  IF EXISTS (SELECT * FROM opc_operation WHERE plan_id = @plan_id AND btcy_id IS NULL)
  BEGIN
    SET NOCOUNT OFF
    SELECT 'null_btcy_id', * FROM opc_operation WHERE plan_id = @plan_id AND btcy_id IS NULL
    RAISERROR('_LoadOPCPlan_CreateBCJs @plan_id=%d: FATAL ERROR: opc_operation rows where btcy_id is null', 11, 1, @plan_id)
    RETURN -30000
  END

  -- Transaction logic
  DECLARE @tcount INT
  SELECT @tcount = @@TRANCOUNT
  IF @tcount = 0
    BEGIN TRAN _LoadOPCPlan_CreateBCJs
  ELSE
    SAVE TRAN _LoadOPCPlan_CreateBCJs

  DECLARE @msg NVARCHAR(255)

  -- Create a table that contains a join of all of the BATCH records and opc_operation records for this
  -- plan to be used as a source of the remaining queries that are performed.
  CREATE TABLE #all_batc
  (
    -- Data from BATCH_V
    id              ObjID NOT NULL PRIMARY KEY (id, op_id),
    deleted         TINYINT NULL,
    _ParentID       ObjID NULL,
    name            NVARCHAR(8) NULL,
    -- Data from opc_operation
    op_id           INT NOT NULL,
    description     NVARCHAR(255) NULL,
    _OPCOpNum       NVARCHAR(8) NULL,
    _SchedStartTime DATETIME NULL,
    _SchedDuration  INT NULL,
    _SchedStopTime  DATETIME NULL,
    instanceno      INT NULL,
    status          CHAR(1) NULL
  )
  INSERT INTO #all_batc
    SELECT
      ISNULL(b.id,0),
      b.deleted,
      b._ParentID,
      b.name, 
      ISNULL(o.op_id,0),
      o.op_desc,
      o.op_name,
      o.est_start,
      o.duration,
      o.est_stop,
      o.instanceno,
      o.status
    FROM BATCH_V AS b
      LEFT OUTER JOIN opc_operation AS o ON 
        o.btcy_id = b._ParentID
        AND o.job_name = b.name
        AND ( o.op_name = b._OPCOpNum OR b._OPCOpNum IS NULL )
        AND o.plan_id = @plan_id
        AND o.instanceno = 0
    WHERE
      b._ParentCID = 'BTCY'
      AND b._ParentID IN (SELECT DISTINCT btcy_id FROM opc_operation WHERE plan_id = @plan_id)
  SET @rc = @@ROWCOUNT
  IF @verbose = 1
  BEGIN
    PRINT '_LoadOPCPlan_CreateBCJs: inserted ' + CAST(@rc AS NVARCHAR) + ' row(s) into #all_batc for previously existing jobs'
  END

  -- Insert records into #all_batc for operations that do not have batch jobs yet
  ALTER TABLE #all_batc ADD CONSTRAINT all_batc_op_id_id_UQ UNIQUE (op_id, id)
  INSERT INTO #all_batc
    SELECT
      0,
      0,
      o.btcy_id,
      o.job_name, 
      o.op_id,
      o.op_desc,
      o.op_name,
      o.est_start,
      o.duration,
      o.est_stop,
      o.instanceno,
      o.status
    FROM
      opc_operation AS o
    WHERE
      o.plan_id = @plan_id
      AND o.instanceno = 0
      AND ISNULL(o.job_name, '') <> ''
      AND o.op_id NOT IN (SELECT DISTINCT op_id FROM #all_batc)
  SET @rc = @@ROWCOUNT

  IF @verbose = 1
  BEGIN
    PRINT '_LoadOPCPlan_CreateBCJs: inserted ' + CAST(@rc AS NVARCHAR) + ' row(s) into #all_batc for new jobs'
    SELECT '#all_batc' AS '#all_batc', * FROM #all_batc
  END

  IF EXISTS (SELECT * FROM #all_batc WHERE id <> 0)
  BEGIN  
    -- Update BATCH schedule attributes
    UPDATE b SET
      _SchedStartTime = a._SchedStartTime,
      _SchedStatusFlags = 1,
      _SchedDuration = a._SchedDuration,
      _SchedStopTime = a._SchedStopTime,
      _InitEstComplete = a._SchedStopTime,
      _ImmedOverrideDuration = NULL,
      _ImmedOverrideStartTime = NULL,
      _OPCOpNum = a._OPCOpNum,
      deleted = 0
    FROM
      BATCH_V AS b JOIN #all_batc AS a ON a.id = b.id
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @verbose = 1
      PRINT '_LoadOPCPlan_CreateBCJs: Update to BATCH attributes affected ' + CAST(@rc AS NVARCHAR) + ' row(s)'
    IF @err <> 0
    BEGIN
      RAISERROR('Aborting due to error updating BATCH_V from opc_operation', 3, 1)
      ROLLBACK TRAN _LoadOPCPlan_CreateBCJs
      RETURN -30000
    END
  END

  -- Create any batch jobs required
  DECLARE @nbcj INT
  SELECT DISTINCT @nbcj = COUNT(*) 
    FROM #all_batc
    WHERE id = 0
  IF @nbcj > 0
  BEGIN
    IF @verbose = 1
    BEGIN
      PRINT '_LoadOPCPlan_CreateBCJs: creating ' + CAST(@nbcj AS NVARCHAR) + ' new job(s)'
      SELECT DISTINCT 'new BATCH', COUNT(*), _ParentID, name, _OPCOpNum, _SchedStartTime
      FROM #all_batc AS a
      WHERE a.id = 0 AND a.op_id <> 0
      GROUP BY _ParentID, name, _OPCOpNum, _SchedStartTime
      ORDER BY COUNT(*) DESC, _ParentID, name, _OPCOpNum, _SchedStartTime
    END

    CREATE TABLE #new_job
    (
      op_id INT NOT NULL,
      rowno INT IDENTITY(0, 1)
    )

    -- Create the new records
    INSERT INTO #new_job
    SELECT DISTINCT a.op_id
    FROM #all_batc AS a
    WHERE a.op_id <> 0 AND a.id = 0
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('Error computing new BATCH records, aborting', 3, 1)
      ROLLBACK TRAN _LoadOPCPlan_CreateBCJs
      RETURN -30000
    END

    SELECT @nbcj = COUNT(*) FROM #new_job
    DECLARE @base_id ObjID
    EXEC @base_id = alloc_BATC @nbcj

    IF @verbose = 1
      PRINT 'Creating ' + CAST(@nbcj AS NVARCHAR) + ' BATCH_C ' + CASE WHEN @nbcj=1 THEN 'row' ELSE 'rows' END

    INSERT INTO BATCH_C(id, name, description, _OPCOpNum, _OPCOpInstanceNo, _SchedStartTime, _SchedStatusFlags, _SchedDuration, _SchedStopTime, _InitEstComplete, _ParentCID, _ParentID, _CurrentStateID, _DesiredStateID)
      SELECT @base_id+j.rowno, a.name, a.description, a._OPCOpNum, 0, a._SchedStartTime, 1, a._SchedDuration, a._SchedStopTime, a._SchedStopTime, 'BTCY', a._ParentID, 376, 376 /* 376=Pending */
      FROM
        #new_job AS j JOIN #all_batc AS a ON a.op_id = j.op_id
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('Error creating new BATCH records, aborting', 3, 1)
      ROLLBACK TRAN _LoadOPCPlan_CreateBCJs
      RETURN -30000
    END

    UPDATE a SET id = @base_id+j.rowno
      FROM #all_batc AS a JOIN #new_job AS j ON j.op_id = a.op_id
  END

  -- Update opc_operation for the jobs.
  IF @verbose = 1
  BEGIN
    EXEC asisp_checkpointmessage 'Updating opc_operation.job_id from newly created BATCH records for operations with names', 1, @chkpt
  END
  UPDATE o SET job_id = a.id
    FROM opc_operation AS o JOIN #all_batc AS a ON a.op_id = o.op_id
    WHERE o.plan_id = @plan_id AND o.job_id IS NULL
  IF @@ERROR <> 0
  BEGIN
    ROLLBACK TRAN _LoadOPCPlan_CreateBCJs
    RETURN -30000
  END
  IF @verbose = 1
    SELECT @chkpt AS chkpt, * FROM opc_operation WHERE plan_id = @plan_id
  SET @chkpt = @chkpt + 1

  -- Reset the _SchedStatusFlags and deleted columns for all existing jobs in the applications for this plan
  UPDATE b SET
    _SchedStatusFlags = 0
  FROM BATCH_V AS b
    JOIN #all_batc AS a ON a.id = b.id
  WHERE
    a._SchedStartTime IS NULL
  SET @rc = @@ROWCOUNT
  IF @verbose = 1
    PRINT '_LoadOPCPlan_CreateBCJs: reset _SchedStatusFlags updated ' + CAST(@rc AS NVARCHAR) + ' row(s)'

  -- Now make sure that there are LOB analogs to all BATCH objects in the plan based upon
  -- the opc_app_link table
  DECLARE @r INT
  EXEC @r = _LoadOPCPlan_CreateBCJLOBs @plan_id, @chkpt OUTPUT, @verbose
  IF @r < 0
  BEGIN
    ROLLBACK TRAN _LoadOPCPlan_CreateBCJs
    RETURN @r
  END

  IF @tcount = 0
    COMMIT TRAN _LoadOPCPlan_CreateBCJs
GO
