-- File: archive\db\sql\asisp_processstagedopcevents.sql
-- SID: 1.15
-- Date Modified: 01/11/30 12:09:05
--
-- Module Description
--   Defines the asisp_processStagedOPCEvents stored procedure

USE Object
GO


---------------------------------------------------------------------------------------------------
-- asisp_processStagedOPCEvents
PRINT 'Creating procedure asisp_processStagedOPCEvents'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'asisp_processStagedOPCEvents' AND type = 'P')
  DROP PROCEDURE asisp_processStagedOPCEvents
GO

-- File: archive\db\sql\asisp_processstagedopcevents.sql
-- SID: 1.15
-- Date Modified: 01/11/30 12:09:05
-- Processes the records in the StagedOPCEvent table
CREATE PROCEDURE asisp_processStagedOPCEvents
  @maxrows INT = 0,      -- maximum number of rows to process. 0=process all rows
  @verbose TINYINT = 0   -- generate progress output during execution
AS
  SET NOCOUNT ON

  DECLARE @err INT, @rc INT, @step INT, @errmsg NVARCHAR(255)
  SET @step = 1

/*  -- Diagnostic code to save events in the asisp_processStagedOPCEvents_Audit table.
  DECLARE @starttime DATETIME
  SET @starttime = CURRENT_TIMESTAMP
  INSERT INTO asisp_processStagedOPCEvents_Audit(starttime, endtime, maxrows, result)
    VALUES ( @starttime, NULL, @maxrows, NULL )
*/

  SET ROWCOUNT @maxrows

  IF @verbose = 1
    PRINT 'asisp_processStagedOPCEvents: Step ' + CAST(@step AS NVARCHAR(9)) + ': creating #StagedOPCInProgress'
  SET @step = @step + 1

  CREATE TABLE #StagedOPCInProgress
  (
    timestamp       BINARY(8) PRIMARY KEY,
    ctime           DATETIME,
    src_cid         ClassID,
    src_id          ObjID,
    TIME            DATETIME NULL,
    SEQ_NO          NVARCHAR(50) NULL,
    MESSAGE         NVARCHAR(255) NULL,
    MESSAGE_TEXT    NVARCHAR(255) NULL,
    APPID           NVARCHAR(16) NULL,
    WKSTNID         NVARCHAR(8) NULL,
    OPID            NVARCHAR(10) NULL,
    JOBNAME         NVARCHAR(8) NULL,
    PRTY            INT NULL,
    IA              NVARCHAR(16) NULL,
    EC              NVARCHAR(8) NULL,
    RESOURCE        NVARCHAR(40) NULL,
    WAIT_MINUTES    INT NULL,
    app_id          ObjID NULL,
    job_cid         ClassID NULL,
    job_id          ObjID NULL,
    EventCID        ClassID NULL,
    EventText       NVARCHAR(255) NULL,
    AlertStateID    INT NULL,
    PriorityID      INT NULL,
    StateID         INT NULL,
    op_id           INT IDENTITY(1,1)
  )

  DECLARE @nEvents INT
  INSERT INTO #StagedOPCInProgress SELECT *, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL FROM StagedOPCEvent
  SELECT @nEvents = @@ROWCOUNT

  IF @verbose = 1
  BEGIN
    PRINT 'asisp_processStagedOPCEvents: Step ' + CAST(@step AS NVARCHAR(9)) + ': Inserted ' + CAST(@nEvents AS NVARCHAR(9)) + ' row(s)' 
    SELECT @step AS Step, * FROM #StagedOPCInProgress ORDER BY timestamp
  END
  SET @step = @step + 1

  SET ROWCOUNT 0

  IF @nEvents = 0
  BEGIN
/*  -- Diagnostic code to save events in the asisp_processStagedOPCEvents_Audit table.
    UPDATE asisp_processStagedOPCEvents_Audit SET endtime = CURRENT_TIMESTAMP, result = 0 WHERE starttime = @starttime
*/
    RETURN 0
  END

  -- Change NULL OPID and WKSTNID to empty string to make comparisons easier
  UPDATE #StagedOPCInProgress SET OPID = '' WHERE OPID IS NULL
  UPDATE #StagedOPCInProgress SET WKSTNID = '' WHERE WKSTNID IS NULL

  DECLARE @tcount INT
  SELECT @tcount = @@TRANCOUNT
  IF @tcount = 0
    BEGIN TRANSACTION asisp_processStagedOPCEvents
  ELSE
    SAVE TRANSACTION asisp_processStagedOPCEvents

  -- Store the timestamp of the last StagedOPCEvent record that we're processing. If the transaction
  -- is successful we delete all StagedOPCEvent records up to this one.
  DECLARE @max_timestamp BINARY(8)
  SELECT @max_timestamp = MAX(timestamp) FROM #StagedOPCInProgress

  -- Delete records for null named jobs
  IF @verbose = 1
    PRINT 'asisp_processStagedOPCEvents: Step ' + CAST(@step AS NVARCHAR(9)) + ': Removing events with null job names'
  SET @step = @step + 1
  DELETE FROM #StagedOPCInProgress WHERE ISNULL(JOBNAME,'') = ''

  IF (SELECT COUNT(*) FROM #StagedOPCInProgress) = 0
  BEGIN
    PRINT 'Step ' + CAST(@step AS NVARCHAR(9)) + ': All events have NULL job names, nothing to process.'
    GOTO EndTran
  END

  -- Build a temp table that matches all events to schedules/jobs
  IF @verbose = 1
    PRINT 'asisp_processStagedOPCEvents: Step ' + CAST(@step AS NVARCHAR(9)) + ': Build a temp table that matches all events to schedules/jobs'
  CREATE TABLE #job1
  (
    APPID           NVARCHAR(16) NOT NULL,
    JOBNAME         NVARCHAR(8) NOT NULL,
    OPID            NVARCHAR(10) NOT NULL,
    WKSTNID         NVARCHAR(8) NOT NULL,
    btcy_id         ObjID NOT NULL,
    batc_id         ObjID NOT NULL
  )
  INSERT INTO #job1
    SELECT DISTINCT s.APPID, s.JOBNAME, s.WKSTNID, s.OPID, btcy.id, batc.id
    FROM
      #StagedOPCInProgress AS s
      JOIN BATCH_C AS batc ON batc.name = s.JOBNAME 
        -- event may not have OPID, so the join must handle this case
        AND (s.OPID='' OR batc._OPCOpNum = s.WKSTNID + '_' + s.OPID)
      JOIN BatchCycle_C AS btcy ON btcy.name = s.APPID AND btcy.id = batc._ParentID
      JOIN containment AS c ON c.dst_cid = 'BTCY' AND c.dst_id = btcy.id
    WHERE
      batc.deleted = 0
      AND c.deleted = 0
  SET @step = @step + 1

  IF @verbose = 1
    PRINT 'asisp_processStagedOPCEvents: Step ' + CAST(@step AS NVARCHAR(9)) + ': Looking for unique schedule/job mappings'
  CREATE TABLE #job1_distinct
  (
    APPID           NVARCHAR(16) NOT NULL,
    JOBNAME         NVARCHAR(8) NOT NULL,
    WKSTNID         NVARCHAR(8) NOT NULL,
    OPID            NVARCHAR(10) NOT NULL
  )
  INSERT INTO #job1_distinct
    SELECT DISTINCT APPID, JOBNAME, WKSTNID, OPID
    FROM #job1
    GROUP BY APPID, JOBNAME, WKSTNID, OPID
    HAVING COUNT(*) = 1
  IF @@ROWCOUNT > 0
  BEGIN
    SET @step = @step + 1
    IF @verbose = 1
      PRINT 'asisp_processStagedOPCEvents: Step ' + CAST(@step AS NVARCHAR(9)) + ': updating unique mappings'

    UPDATE s SET app_id = j.btcy_id, job_cid = 'BATC', job_id = j.batc_id
    FROM #StagedOPCInProgress AS s
      JOIN #job1_distinct AS jd ON jd.APPID = s.APPID AND jd.JOBNAME = s.JOBNAME AND jd.OPID = s.OPID
      JOIN #job1 AS j ON j.APPID = jd.APPID AND j.JOBNAME = jd.JOBNAME AND j.JOBNAME = jd.OPID
  END
  SET @step = @step + 1

  -- Create temp table that lets us find out all batch schedule sets that contain a schedule that
  -- matches @APPID
  CREATE TABLE #bcys_btcy
  (
     bcys_src_cid ClassID NULL,
     bcys_src_id  ObjID NULL,
     bcys_id      ObjID NULL,
     btcy_id      ObjID NOT NULL,
     is_default   TINYINT NOT NULL
  )

  -- Table to hold the ancestory of a given OS
  CREATE TABLE #os_ancestor
  (
     os_id        ObjID NOT NULL PRIMARY KEY (os_id, src_cid, src_id, gen),
     src_cid      ClassID NOT NULL,
     src_id       ObjID NOT NULL,
     gen          INT NOT NULL
  )

  -- Try to match the application by various means
  DECLARE appc INSENSITIVE CURSOR FOR
    SELECT DISTINCT src_cid, src_id, APPID
    FROM #StagedOPCInProgress
    WHERE app_id IS NULL
  OPEN appc
  DECLARE @src_cid ClassID, @src_id ObjID, @APPID NVARCHAR(16), @app_id ObjID
  FETCH NEXT FROM appc INTO @src_cid, @src_id, @APPID
  WHILE @@FETCH_STATUS <> -1
  BEGIN
    SET @app_id = NULL
    IF @verbose = 1
      PRINT 'src_cid=[' + @src_cid + '] src_id=[' + CAST(@src_id AS NVARCHAR) + '] APPID=[' + @APPID + ']'

    -- Populate temp table that lets us find out all batch schedule sets that contain a schedule that
    -- matches @APPID
    TRUNCATE TABLE #bcys_btcy
    INSERT INTO #bcys_btcy
      SELECT
        NULL,
        NULL,
        bcys_btcy.src_id,
        btcy.id,
        0
      FROM
        BatchCycle_C AS btcy
          JOIN containment AS bcys_btcy ON bcys_btcy.src_cid = 'BCYS' AND bcys_btcy.dst_cid = 'BTCY' AND bcys_btcy.dst_id = btcy.id
      WHERE
        btcy.name = @APPID
        AND bcys_btcy.deleted = 0
    SELECT @rc = @@ROWCOUNT, @err = @@ERROR
    IF @err <> 0
    BEGIN
      RAISERROR('asisp_processStagedOPCEvents: Error inserting into #bcys_btcy', 11, 1)
      DEALLOCATE appc
      RETURN -30000
    END
    IF @rc = 1
    BEGIN
      SELECT @app_id = btcy_id FROM #bcys_btcy
      IF @verbose = 1 
        PRINT 'Found uniquely named schedule [' + @APPID + ']. id=' + CAST(@app_id AS NVARCHAR)
      GOTO ScheduleFound
    END
    ELSE IF @rc = 0
    BEGIN
      IF @verbose = 1 
        PRINT 'No schedules named [' + @APPID + ']'
      GOTO NextApp
    END
    ELSE IF @rc > 1
    BEGIN
      IF @verbose = 1 
        PRINT 'Found ' + CAST(@rc AS VARCHAR) + ' schedules named [' + @APPID + ']'
      -- Fill in parent information for the matched schedule sets
      UPDATE bb SET 
        bcys_src_cid = c.src_cid,
        bcys_src_id  = c.src_id
      FROM
        #bcys_btcy AS bb
        JOIN containment AS c ON c.dst_cid = 'BCYS' AND c.dst_id = bb.bcys_id
      WHERE
        c.dst_cid = 'BCYS' AND c.deleted = 0

      -- Any schedule under a schedule set that is the default for the OS is marked
      UPDATE bb SET is_default = 1
      FROM #bcys_btcy AS bb
        JOIN BatchCycles_C AS bcys ON bcys.id = bb.bcys_id
        JOIN OS_C AS os ON os.id = bb.bcys_src_id
      WHERE
        bb.bcys_src_cid = 'OS'
        AND os.name = bcys.name

      IF @verbose = 1
        SELECT '#bcys_btcy' AS '#bcys_btcy', * FROM #bcys_btcy

      -- If there are two matching schedules and one of them is located under the default schedule
      -- set for the current OS we use the other one.
      IF @rc = 2
      BEGIN
        IF @verbose = 1
          PRINT 'asisp_processStagedOPCEvents: found 2 schedules named [' + @APPID + ']'
        SELECT @app_id = btcy_id FROM #bcys_btcy WHERE is_default = 0
        IF @@ROWCOUNT = 1
        BEGIN
          SELECT @app_id = btcy_id FROM #bcys_btcy
          IF @verbose = 1 
            PRINT 'Found unique non-default schedule [' + @APPID + ']. id=' + CAST(@app_id AS NVARCHAR)
        END
        ELSE GOTO CheckMultiples
      END
      ELSE
      BEGIN
CheckMultiples:
        -- Try for the simple case that there is an instance of this application under a schedule set
        -- in this OS (not the default).
        SELECT @app_id = bb.btcy_id
        FROM #bcys_btcy AS bb
        WHERE bcys_src_cid = 'OS' AND bcys_src_id = @src_id AND is_default = 0
        IF @@ROWCOUNT = 1
        BEGIN
          IF @verbose = 1
            PRINT 'asisp_processStagedOPCEvents: matching [' + @APPID + '] to schedule ' + CAST(@app_id AS NVARCHAR)
                  + ' since schedule exists under OS ' + CAST(@src_id AS NVARCHAR)
        END
        ELSE
        BEGIN
          -- We now try to determine the closest match based on the OS (current src_id) that sent the events.
          SELECT @rc = COUNT(*) FROM #bcys_btcy
          IF @verbose = 1
            PRINT 'asisp_processStagedOPCEvents: trying to find closest match for [' + @APPID + '] among ' + CAST(@rc AS NVARCHAR) + ' schedules'

          -- If the #os_ancestor table has not been filled in for this OS yet, we do so now
          IF NOT EXISTS (SELECT * FROM #os_ancestor WHERE os_id = @src_id)
          BEGIN
            IF @verbose = 1
              PRINT 'asisp_processStagedOPCEvents: populating ancestor table for OS ' + CAST(@src_id AS NVARCHAR)
            DECLARE @gen INT
            SET @gen = 1
            INSERT INTO #os_ancestor
              SELECT @src_id, c.src_cid, c.src_id, @gen
              FROM containment AS c
              WHERE dst_cid = 'OS' AND dst_id = @src_id AND deleted = 0
            WHILE @@ROWCOUNT > 0
            BEGIN
              SET @gen = @gen + 1
              INSERT INTO #os_ancestor
                SELECT @src_id, c.src_cid, c.src_id, @gen
                FROM #os_ancestor AS a
                  JOIN containment AS c ON c.dst_cid = a.src_cid AND c.dst_id = a.src_id
                WHERE c.deleted = 0 AND a.os_id = @src_id AND a.gen = @gen - 1 AND c.src_cid <> 'ROOT'
            END
          END

          -- Try to find a non-default schedule
          DECLARE @nMatches INT, @min_gen INT
          SET @min_gen = NULL
          SELECT @min_gen = MIN(gen)
          FROM #bcys_btcy AS bb
            JOIN #os_ancestor AS a ON a.src_cid = bb.bcys_src_cid AND a.src_id = bb.bcys_src_id
          WHERE a.os_id = @src_id AND bb.is_default = 0
          SET @rc = @@ROWCOUNT
          IF @rc = 1 AND @min_gen IS NOT NULL -- There is a schedule that is clearly closest to these events
          BEGIN
            SELECT @app_id = bb.btcy_id
            FROM #bcys_btcy AS bb
              JOIN #os_ancestor AS a ON a.src_cid = bb.bcys_src_cid AND a.src_id = bb.bcys_src_id
            WHERE a.os_id = @src_id AND a.gen = @min_gen AND bb.is_default = 0
            IF @verbose = 1
              PRINT 'asisp_processStagedOPCEvents: found non-default schedule [' + CAST(@app_id AS NVARCHAR) + '] at generation [' + CAST(@min_gen AS NVARCHAR) + ']'
            GOTO ScheduleFound
          END
          ELSE
            IF @verbose = 1
              PRINT 'asisp_processStagedOPCEvents: did not find non-default schedule'

          -- Try to find the default schedule
          SELECT @min_gen = MIN(gen)
          FROM #bcys_btcy AS bb
            JOIN #os_ancestor AS a ON a.src_cid = bb.bcys_src_cid AND a.src_id = bb.bcys_src_id
          WHERE a.os_id = @src_id AND bb.is_default = 1
          SET @rc = @@ROWCOUNT
          IF @rc = 1 AND @min_gen IS NOT NULL -- There is a default schedule
          BEGIN
            SELECT @app_id = bb.btcy_id
            FROM #bcys_btcy AS bb
              JOIN #os_ancestor AS a ON a.src_cid = bb.bcys_src_cid AND a.src_id = bb.bcys_src_id
            WHERE a.os_id = @src_id AND a.gen = @min_gen AND bb.is_default = 1
            IF @verbose = 1
              PRINT 'asisp_processStagedOPCEvents: found default schedule [' + CAST(@app_id AS NVARCHAR) + '] at generation [' + CAST(@min_gen AS NVARCHAR) + ']'
            GOTO ScheduleFound
          END
          ELSE
            IF @verbose = 1
              PRINT 'asisp_processStagedOPCEvents: did not find default schedule'

--SELECT '#bcys_btcy' AS '#bcys_btcy', * FROM #bcys_btcy
--SELECT '#os_ancestor' AS '#os_ancestor', * FROM #os_ancestor WHERE os_id = @src_id

          -- Punt because as this point we really can't tell of the multiple schedules
          -- that match this event which one it should go to.
          RAISERROR('asisp_processStagedOPCEvents: Found %d schedules named [%s], cannot disambiguate. Number of potential matches = %d', 3, 1, @rc, @APPID, @nMatches)
          SET @errmsg = 'Found ' + CAST(@rc AS NVARCHAR) + ' schedules named [' + @APPID + '], cannot disambiguate'
          INSERT INTO UnhandledOPCEvent
            SELECT CURRENT_TIMESTAMP, @errmsg, * FROM #StagedOPCInProgress WHERE APPID = @APPID
          DELETE FROM #StagedOPCInProgress WHERE APPID = @APPID
          GOTO NextApp
        END
      END
    END
    ELSE
    BEGIN
      RAISERROR('asisp_processStagedOPCEvents: NULL ROWCOUNT for APPID [%s]', 3, 1, @APPID)
      INSERT INTO UnhandledOPCEvent
        SELECT CURRENT_TIMESTAMP, 'NULL ROWCOUNT', * FROM #StagedOPCInProgress WHERE APPID = @APPID
      DELETE FROM #StagedOPCInProgress WHERE APPID = @APPID
      GOTO NextApp
    END

ScheduleFound:
    -- if found, update the temp table to reflect the correct object
    IF @app_id IS NOT NULL
    BEGIN
      UPDATE #StagedOPCInProgress
      SET app_id = @app_id
      WHERE src_cid = @src_cid AND src_id = @src_id AND APPID = @APPID
      IF @verbose = 1
        PRINT '  src_cid=' + @src_cid + ' src_id=' + CAST(@src_id AS NVARCHAR(9)) + ' app_id=' + CAST(@app_id AS NVARCHAR(9))
    END

NextApp:
    FETCH NEXT FROM appc INTO @src_cid, @src_id, @APPID
  END
  CLOSE appc
  DEALLOCATE appc

  IF @verbose = 1
    PRINT 'asisp_processStagedOPCEvents: Step ' + CAST(@step AS NVARCHAR(9)) + ': checking for new applications'
  SET @step = @step + 1

  -- If we get to this point and there are still applications that have not been found
  -- then we create them under the OS.
  IF EXISTS (SELECT * FROM #StagedOPCInProgress WHERE app_id IS NULL)
  BEGIN
    IF @verbose = 1
      PRINT 'asisp_processStagedOPCEvents: Creating Batch Schedule Sets for OSes'
    -- There has to be a Batch Schedule Set under each OS with the name of the OS
    CREATE TABLE #bcys
    (
      os_id   ObjID,
      bcys_id ObjID NULL
    )
    INSERT INTO #bcys(os_id, bcys_id)
      SELECT DISTINCT s.src_id, NULL
      FROM #StagedOPCInProgress AS s
    IF @verbose = 1
      SELECT 'BCYS PRE-UPDATE', * FROM #bcys
    UPDATE b SET bcys_id = dst_id 
      FROM #bcys AS b JOIN containment AS c ON c.dst_cid = 'BCYS' AND c.src_cid = 'OS' AND c.src_id = os_id
        JOIN BatchCycles_C AS bcys ON bcys.id = c.dst_id
        JOIN OS_C AS os ON os.id = os_id
      WHERE
        c.deleted = 0
        AND bcys.name = os.name
    IF @verbose = 1
      SELECT 'BCYS POST-UPDATE', * FROM #bcys
    DECLARE @nNewBCYS INT
    SELECT @nNewBCYS = COUNT(*) FROM #bcys WHERE bcys_id IS NULL
    IF @nNewBCYS > 0
    BEGIN
      IF @verbose = 1
        PRINT 'asisp_processStagedOPCEvents: Creating ' + CAST(@nNewBCYS AS NVARCHAR(9)) + ' Batch schedule sets'
      DECLARE @base_id ObjID
      EXEC @base_id = alloc_BCYS @nNewBCYS
      CREATE TABLE #new_bcys ( os_id ObjID, rownum INT IDENTITY(0,1) )
      IF @verbose = 1
        PRINT 'asisp_processStagedOPCEvents: Creating #new_bycs'
      INSERT INTO #new_bcys(os_id) SELECT DISTINCT os_id FROM #bcys WHERE bcys_id IS NULL
      IF @@ERROR<>0
      BEGIN
        ROLLBACK TRANSACTION asisp_processStagedOPCEvents
        RETURN -30000
      END
      IF @verbose = 1
        SELECT 'NEW BCYS' AS 'NEW BCYS', * FROM #new_bcys
      IF @verbose = 1
        PRINT 'asisp_processStagedOPCEvents: INSERT BatchCycles_C'
      INSERT INTO BatchCycles_C(id,name) SELECT @base_id+rownum, os.name FROM #new_bcys JOIN OS_C AS os ON os.id = os_id
      IF @verbose = 1
        PRINT 'asisp_processStagedOPCEvents: INSERT BatchCycles_A'
      INSERT INTO BatchCycles_A(id,_ChildEventMatrixID,_ExceptionMatrixID) SELECT @base_id+rownum, 0, 0 FROM #new_bcys
      IF @verbose = 1
        PRINT 'asisp_processStagedOPCEvents: INSERT OS_BCYS_PHYC'
      INSERT INTO OS_BCYS_PHYC(src_id, dst_id) SELECT nb.os_id, @base_id+rownum FROM #new_bcys AS nb
      IF @verbose = 1
        PRINT 'asisp_processStagedOPCEvents: INSERT PendingNotification'
      INSERT INTO PendingNotification(type,fields,detail)
        SELECT 
          et.type,
          CONVERT(NVARCHAR(5), c.cno) + '|' +
          CONVERT(NVARCHAR(9), @base_id+rownum) + '|' +
          'BCYS' + '|' +
          '|' +
          'PHYC' + '|' +
          CONVERT(NVARCHAR(5), p.cno) + '|' +
          CONVERT(NVARCHAR(9), nb.os_id),
          os.name
        FROM
          EventType AS et,
          #new_bcys AS nb
            JOIN obj_class AS p ON p.cid = 'OS'
            JOIN obj_class AS c ON c.cid = 'BCYS'
            JOIN OS_C AS os ON os.id = nb.os_id
        WHERE
          et.name = 'ObjectCreated'
      IF @verbose = 1
      BEGIN
        PRINT '  UPDATE b SET bcys_id = @base_id + nb.rownum'
        PRINT '    FROM #bcys AS b JOIN #new_bcys AS nb ON nb.os_id = b.os_id'
      END     
      UPDATE b SET bcys_id = @base_id + nb.rownum
        FROM #bcys AS b JOIN #new_bcys AS nb ON nb.os_id = b.os_id
    END

    -- Now create the Batch Schedules for the applications
    CREATE TABLE #btcy
    (
      src_cid    ClassID,
      src_id     ObjID,
      APPID      NVARCHAR(16),
      bcys_id    ObjID,
      rownum     INT IDENTITY(0,1)
    )
    INSERT INTO #btcy(src_cid, src_id, APPID, bcys_id)
      SELECT DISTINCT src_cid, src_id, APPID, bcys_id
      FROM #StagedOPCInProgress AS s JOIN #bcys AS b ON b.os_id = s.src_id
      WHERE app_id IS NULL
    SET @rc = @@ROWCOUNT
    IF @verbose = 1
      SELECT 'BTCY', * FROM #btcy
    IF @rc > 0
    BEGIN
      DECLARE @nNewBTCY INT
      SELECT @nNewBTCY = COUNT(*) FROM #btcy
      EXEC @base_id = alloc_BTCY @nNewBTCY
      INSERT INTO BatchCycle_C(id, name) SELECT @base_id + rownum, APPID FROM #btcy
      INSERT INTO BatchCycle_A(id, _ChildEventMatrixID, _ExceptionMatrixID) SELECT @base_id + rownum, 0, 0 FROM #btcy
      INSERT INTO BCYS_BTCY_AGGC(src_id, dst_id) SELECT bcys_id, @base_id+rownum FROM #btcy
      INSERT INTO PendingNotification(type,fields,detail)
        SELECT 
          et.type,
          CONVERT(NVARCHAR(5), c.cno) + '|' +
          CONVERT(NVARCHAR(9), @base_id+rownum) + '|' +
          'BTCY' + '|' +
          '|' +
          'PHYC' + '|' +
          CONVERT(NVARCHAR(5), p.cno) + '|' +
          CONVERT(NVARCHAR(9), nb.bcys_id),
          nb.APPID
        FROM
          EventType AS et,
          #btcy AS nb
            JOIN obj_class AS p ON p.cid = 'BCYS'
            JOIN obj_class AS c ON c.cid = 'BTCY'
        WHERE
          et.name = 'ObjectCreated'
      UPDATE s SET app_id = @base_id+rownum
        FROM #StagedOPCInProgress AS s 
          JOIN #btcy AS b ON s.src_cid = b.src_cid AND s.src_id = b.src_id AND s.APPID = b.APPID
    END
  END

  IF EXISTS (SELECT * FROM #StagedOPCInProgress WHERE app_id IS NULL)
  BEGIN
    RAISERROR('asisp_processStagedOPCEvents: Still have staged events with null app_id, aborting!', 3, 1)
    IF @verbose = 1
      SELECT 'NULL APPID' AS 'NULL APPID', * FROM #StagedOPCInProgress WHERE app_id IS NULL
    ROLLBACK TRAN asisp_processStagedOPCEvents
    RETURN -30000
  END

  IF @verbose = 1
    PRINT 'asisp_processStagedOPCEvents: Step ' + CAST(@step AS NVARCHAR(9)) + ': looking up job ids'
  IF @verbose = 1
    SELECT @step AS step, * FROM #StagedOPCInProgress
  SET @step = @step + 1

  -- For all applications that have been successfully looked up we now do the same with the jobs
  CREATE TABLE #jobmatch
  (
    btcy_id         ObjID NOT NULL  PRIMARY KEY (timestamp, btcy_id, job_name, WKSTNID, OPID, job_id),
    job_id          ObjID NOT NULL,
    job_name        NVARCHAR(8) NOT NULL,
    SchedStartTime  DATETIME NULL,
    WKSTNID         NVARCHAR(8) NOT NULL,
    OPID            NVARCHAR(8) NOT NULL,
    TIME            DATETIME NOT NULL,
    timestamp       BINARY(8) NOT NULL
  )
  INSERT INTO #jobmatch
    SELECT
      b._ParentID,
      b.id,
      b.name,
      b._SchedStartTime,
      s.WKSTNID,
      s.OPID,
      s.TIME,
      s.timestamp
    FROM
      #StagedOPCInProgress AS s
        JOIN BATCH_V AS b ON 
          b._ParentCID = 'BTCY' 
          AND b.deleted = 0
          AND b._ParentID = s.app_id 
          AND b.name = s.JOBNAME
          AND (s.OPID = '' OR b._OPCOpNum IS NULL OR b._OPCOpNum = RTRIM(s.WKSTNID) + '_' + s.OPID)
-- Note: consider the case where there are two jobs that match by name, one with null _OPCOpNum and one
-- with _OPCOpNum that matches the event. In this case we should give preference to the better match,
-- perhaps even to the point of removing the job with the NULL _OPCOpNum.
    WHERE
      s.app_id IS NOT NULL
  IF @@ERROR <> 0
  BEGIN
    PRINT 'Error computing #jobmatch'
    ROLLBACK TRANSACTION asisp_processStagedOPCEvents
    RETURN -30000
  END

/*
  IF (SELECT COUNT(*) FROM #jobmatch) = 0
  BEGIN
    PRINT 'Debug break on no #jobmatch rows'
    ROLLBACK TRANSACTION asisp_processStagedOPCEvents
    RETURN -30000
  END
*/

  IF @verbose = 1
    SELECT '#jobmatch' AS '#jobmatch', * FROM #jobmatch
  
  CREATE TABLE #jobmatchcount
  (
    btcy_id   ObjID NOT NULL PRIMARY KEY (timestamp, btcy_id, job_name, WKSTNID, OPID),
    job_name  NVARCHAR(8) NOT NULL,
    WKSTNID   NVARCHAR(8) NOT NULL,
    OPID      NVARCHAR(8) NOT NULL,
    timestamp BINARY(8) NOT NULL,
    nmatch    INT NOT NULL
  )
  INSERT INTO #jobmatchcount
    SELECT DISTINCT btcy_id, job_name, WKSTNID, OPID, timestamp, COUNT(*) 
    FROM #jobmatch
    GROUP BY btcy_id, job_name, WKSTNID, OPID, timestamp

  IF @verbose = 1
    SELECT '#jobmatchcount' AS '#jobmatchcount', * FROM #jobmatchcount

/*
  IF (SELECT COUNT(*) FROM #jobmatchcount) = 0
  BEGIN
    PRINT 'Debug break on no #jobmatchcount rows'
    ROLLBACK TRANSACTION asisp_processStagedOPCEvents
    RETURN -30000
  END
*/
  -- For those jobs that match unique the update is straightforward
  IF EXISTS
  (
    SELECT * FROM #jobmatchcount WHERE nmatch = 1
  )
  BEGIN
    IF @verbose = 1
      PRINT 'Mapping unique jobs'
    UPDATE s SET 
      job_cid = 'BATC',
      job_id  = j.job_id
    FROM
      #StagedOPCInProgress AS s
      JOIN #jobmatch AS j ON
        j.timestamp = s.timestamp
        AND j.btcy_id = s.app_id
        AND j.job_name = s.JOBNAME
        AND j.WKSTNID = s.WKSTNID
        AND j.OPID = s.OPID
      JOIN #jobmatchcount AS c ON
         c.timestamp = j.timestamp
         AND c.btcy_id = j.btcy_id
         AND c.job_name = j.job_name
         AND c.WKSTNID = j.WKSTNID
         AND c.OPID = j.OPID
    WHERE
      s.app_id IS NOT NULL
      AND c.nmatch = 1
  END

  -- For those jobs where the match is not unique we have to guess. We do so by trying to match
  -- based on the minimum difference between the scheduled start time and the event time
  IF EXISTS
  (
    SELECT * FROM #jobmatchcount WHERE nmatch > 1
  )
  BEGIN
    IF @verbose = 1
    BEGIN
      PRINT 'Selecting match table'
      SELECT
        'match table',
        *
      FROM
        #StagedOPCInProgress AS s
        JOIN #jobmatch AS j ON
          j.timestamp = s.timestamp
          AND j.btcy_id = s.app_id
          AND j.job_name = s.JOBNAME
          AND j.WKSTNID = s.WKSTNID
          AND j.OPID = s.OPID
        JOIN #jobmatchcount AS c ON
          c.timestamp = j.timestamp
          AND c.btcy_id = j.btcy_id
          AND c.job_name = j.job_name
          AND c.WKSTNID = j.WKSTNID
          AND c.OPID = j.OPID
      WHERE
        s.app_id IS NOT NULL
        AND c.nmatch > 1
        AND 
        (
          j.OPID <> ''
          OR 
          NOT EXISTS
          (
            SELECT * FROM #jobmatch AS j2
            WHERE
              j2.timestamp = j.timestamp
              AND j2.btcy_id = j.btcy_id
              AND j2.job_name = j.job_name
              AND j2.job_id <> j.job_id
              AND ABS(DATEDIFF(SECOND, j2.SchedStartTime, j2.TIME)) < ABS(DATEDIFF(SECOND, j.SchedStartTime, j.TIME))
          )
        )
    END  
    UPDATE s SET 
      job_cid = 'BATC',
      job_id  = j.job_id
    FROM
      #StagedOPCInProgress AS s
      JOIN #jobmatch AS j ON
        j.timestamp = s.timestamp
        AND j.btcy_id = s.app_id
        AND j.job_name = s.JOBNAME
        AND j.WKSTNID = s.WKSTNID
        AND j.OPID = s.OPID
      JOIN #jobmatchcount AS c ON
        c.timestamp = j.timestamp
        AND c.btcy_id = j.btcy_id
        AND c.job_name = j.job_name
        AND c.WKSTNID = j.WKSTNID
        AND c.OPID = j.OPID
    WHERE
      s.app_id IS NOT NULL
      AND c.nmatch > 1
      AND 
      (
        j.OPID <> ''
        OR
        NOT EXISTS
        (
          SELECT * FROM #jobmatch AS j2
          WHERE
            j2.timestamp = j.timestamp
            AND j2.btcy_id = j.btcy_id
            AND j2.job_name = j.job_name
            AND j2.job_id <> j.job_id
            AND ABS(DATEDIFF(SECOND, j2.SchedStartTime, j2.TIME)) < ABS(DATEDIFF(SECOND, j.SchedStartTime, j.TIME))
        )
      )
  END

  -- Update the SchedStartTime on the matched jobs from the IA field of the events
  UPDATE b SET _SchedStartTime =
      '20' + SUBSTRING(IA, 1, 2) + '-' -- year
      + SUBSTRING(IA, 3, 2) + '-'      -- month
      + SUBSTRING(IA, 5, 2) + ' '      -- day
      + SUBSTRING(IA, 7, 2) + ':'      -- hour
      + SUBSTRING(IA, 9, 2)            -- minute
  FROM
    BATCH_C AS b
    JOIN #StagedOPCInProgress AS s ON s.job_cid = 'BATC' AND s.job_id = b.id
  WHERE
    s.IA LIKE '[0-9][0-9][0-1][0-9][0-3][0-9][0-2][0-9][0-5][0-9]'
    AND b._SchedStartTime <> 
      '20' + SUBSTRING(IA, 1, 2) + '-' -- year
      + SUBSTRING(IA, 3, 2) + '-'      -- month
      + SUBSTRING(IA, 5, 2) + ' '      -- day
      + SUBSTRING(IA, 7, 2) + ':'      -- hour
      + SUBSTRING(IA, 9, 2)            -- minute

  IF @verbose = 1
    PRINT 'asisp_processStagedOPCEvents: Step ' + CAST(@step AS NVARCHAR(9)) + ': creating any new jobs required'
  SET @step = @step + 1

  -- Create any new batch jobs required.
  DECLARE @nNewBATC INT  
  SELECT @nNewBATC = COUNT(*) FROM #StagedOPCInProgress WHERE job_cid IS NULL
  IF @nNewBATC > 0
  BEGIN
    CREATE TABLE #batc
    (
      btcy_id   ObjID,
      jobname   NVARCHAR(8),
      WKSTNID   NVARCHAR(8) NULL,
      OPID      NVARCHAR(8) NULL,
      IA        NVARCHAR(16) NULL,
      rownum    INT IDENTITY(0,1)
    )
    INSERT INTO #batc(btcy_id, jobname, WKSTNID, OPID, IA) 
      SELECT DISTINCT app_id, JOBNAME, WKSTNID, OPID, IA
      FROM #StagedOPCInProgress 
      WHERE JOBNAME IS NOT NULL AND job_cid IS NULL
    IF @verbose = 1
      SELECT 'BATC', * FROM #batc
    EXEC @base_id = alloc_BATC @nNewBATC
    INSERT INTO BATCH_C(id, name, description, _ChildEventMatrixID, _ExceptionMatrixID, _ConsoleTrapStatus, _SysplexResource, _ParentCID, _ParentID, _OPCOpNum, _SchedStartTime)
      SELECT
        @base_id+rownum,
        jobname,
        'AUTO-CREATED FROM MESSAGE AT ' + CONVERT(NVARCHAR(40), GETDATE(), 120),
        0 AS _ChildEventMatrixID,
        0 AS _ExceptionMatrixID,
        0 AS _ConsoleTrapStatus,
        0 AS _SysplexResource,
        'BTCY' AS _ParentCID,
        btcy_id AS _ParentID,
        CASE WHEN OPID = '' THEN NULL ELSE RTRIM(WKSTNID) + '_' + OPID END AS _OPCOpNum,
        CASE 
          WHEN IA IS NULL THEN NULL
          WHEN IA NOT LIKE '[0-9][0-9][0-1][0-9][0-3][0-9][0-2][0-9][0-5][0-9]' THEN NULL
          ELSE
            '20' + SUBSTRING(IA, 1, 2) + '-' -- year
            + SUBSTRING(IA, 3, 2) + '-'      -- month
            + SUBSTRING(IA, 5, 2) + ' '      -- day
            + SUBSTRING(IA, 7, 2) + ':'      -- hour
            + SUBSTRING(IA, 9, 2)            -- minute
          END
      FROM #batc
    INSERT INTO PendingNotification(type,fields,detail)
      SELECT 
        et.type,
        CONVERT(NVARCHAR(5), c.cno) + '|' +
        CONVERT(NVARCHAR(9), @base_id+rownum) + '|' +
        'BATC' + '|' +
        '|' +
        'PHYC' + '|' +
        CONVERT(NVARCHAR(5), p.cno) + '|' +
        CONVERT(NVARCHAR(9), nb.btcy_id),
        nb.jobname
      FROM
        EventType AS et,
        #batc AS nb
          JOIN obj_class AS p ON p.cid = 'BTCY'
          JOIN obj_class AS c ON c.cid = 'BATC'
      WHERE
        et.name = 'ObjectCreated'
    UPDATE s SET job_cid = 'BATC', job_id = @base_id+rownum
      FROM #StagedOPCInProgress AS s 
        JOIN #batc AS b ON s.app_id = b.btcy_id AND b.jobname = s.JOBNAME AND b.WKSTNID = s.WKSTNID AND b.OPID = s.OPID
      WHERE s.job_cid IS NULL
  END

  -- Remove any records that did not map to jobs - there should never be any!
  DELETE FROM #StagedOPCInProgress WHERE job_id IS NULL
  SET @rc = @@ROWCOUNT
  IF @rc > 0
    RAISERROR('Step %d: deleted %d records that did not map to jobs', 3, 1, @step, @rc)
  SET @step = @step + 1

  IF @verbose = 1
  BEGIN
    PRINT 'asisp_processStagedOPCEvents: Step ' + CAST(@step AS NVARCHAR(9)) + ': Creating events'
    SELECT @step AS step, * FROM #StagedOPCInProgress
  END
  SET @step = @step + 1

  -- Now that all objects have been found/created we use the OPCEventTypeMap table to generate
  -- events of the appropriate type.
  DECLARE etmc INSENSITIVE CURSOR FOR
    SELECT DISTINCT CAST(m.id AS NVARCHAR(9)), m.EventName, m.WhereClause, m.MapPriority
    FROM OPCEventTypeMap AS m JOIN #StagedOPCInProgress AS s ON s.MESSAGE LIKE m.EventName
    WHERE m.Enabled = 1
    ORDER BY m.MapPriority DESC
  OPEN etmc
  DECLARE @MapId NVARCHAR(9), @MapEventName NVARCHAR(16), @MapWhereClause NVARCHAR(500), @MapPriority INT
  FETCH NEXT FROM etmc INTO @MapId, @MapEventName, @MapWhereClause, @MapPriority
  WHILE @@FETCH_STATUS <> -1
  BEGIN
    IF @verbose = 1
      PRINT 'asisp_processStagedOPCEvents: Checking OPCEventTypeMap ' + CAST(@MapId AS NVARCHAR) + ': EventName=''' + @MapEventName + ''''
    IF @MapWhereClause <> '' SET @MapWhereClause = ' AND (' + @MapWhereClause + ')'
    EXEC("
    UPDATE s SET 
      EventCID = m.EventCID,
      EventText = CASE WHEN m.EventTextPat IS NULL THEN s.MESSAGE_TEXT ELSE m.EventTextPat END,
      AlertStateID = m.AlertStateID,
      PriorityID = m.PriorityID,
      StateID = m.StateID
    FROM
      #StagedOPCInProgress AS s,
      OPCEventTypeMap AS m
    WHERE
      s.MESSAGE LIKE '" + @MapEventName + "' AND s.EventCID IS NULL AND m.id = " + @MapId + @MapWhereClause
    )
    FETCH NEXT FROM etmc INTO @MapId, @MapEventName, @MapWhereClause, @MapPriority
  END
  DEALLOCATE etmc

  -- Remove any records that did not map to exceptions or messages
  IF @verbose = 1
  BEGIN
    PRINT 'Step ' + CAST(@step AS NVARCHAR(9)) + ': deleting records that did not map to exceptions or messages'
    SELECT @step AS step, * FROM #StagedOPCInProgress WHERE EventCID IS NULL
  END
  DELETE FROM #StagedOPCInProgress WHERE EventCID IS NULL
  SET @step = @step + 1

  IF (SELECT COUNT(*) FROM #StagedOPCInProgress) = 0
  BEGIN
    PRINT 'Step ' + CAST(@step AS NVARCHAR(9)) + ': no events to process.'
    GOTO EndTran
  END

  -- We now go into a loop inserting StagedEXCP and StagedMESG records for the events. This
  -- needs to be a loop because the StagedEXCP and StagedMESG records must be inserted in the
  -- correct order.
  DECLARE @nEventsInserted INT, @nEventsTotal INT, @timestamp BINARY(8)
  SELECT @nEventsTotal = COUNT(*) FROM #StagedOPCInProgress
  SET @nEventsInserted = 0
  WHILE @nEventsInserted < @nEventsTotal
  BEGIN
    IF @verbose = 1
      PRINT 'asisp_processStagedOPCEvents: Step ' + CAST(@step AS NVARCHAR(9)) + ', Events Inserted: ' + CAST(@nEventsInserted AS NVARCHAR(9)) + ', Events Total: ' + CAST(@nEventsTotal AS NVARCHAR(9))
    SET @step = @step + 1

    -- Lookup latest exception that does not have an earlier unprocessed message
    SET @timestamp = NULL
    SELECT @timestamp = MAX(e.timestamp)
    FROM #StagedOPCInProgress AS e
    WHERE e.EventCID = 'EXCP'
      AND NOT EXISTS
      (
        SELECT * FROM #StagedOPCInProgress AS m 
        WHERE m.EventCID = 'MESG' AND m.timestamp < e.timestamp
      )

    IF @timestamp IS NOT NULL -- there is at least one exception that is next sequentially
    BEGIN
      IF @verbose = 1
        PRINT 'asisp_processStagedOPCEvents: Inserting StagedEXCP records'

      INSERT INTO StagedEXCP(parent_cid, parent_id, TIME, SEQ_NO, OBJECT_NAME, NATIVE_KEY, EXCP_NAME, EXCP_CD, AlertState, Priority, nEXCH)
        SELECT e.job_cid, e.job_id, e.TIME, e.SEQ_NO, e.JOBNAME, '', e.MESSAGE, ISNULL(e.EventText,''), e.AlertStateID, e.PriorityID, 1
        FROM #StagedOPCInProgress AS e
        WHERE EventCID = 'EXCP' AND e.timestamp <= @timestamp
        ORDER BY e.timestamp
      SELECT @err = @@ERROR, @rc = @@ROWCOUNT
      IF @err <> 0
      BEGIN
        ROLLBACK TRANSACTION asisp_processStagedOPCEvents
       RETURN -30000
      END
      PRINT 'asisp_processStagedOPCEvents: Inserted ' + CAST(@rc AS NVARCHAR(9)) + ' StagedEXCP records'
      SET @nEventsInserted = @nEventsInserted + @rc
     IF @nEventsInserted = @nEventsTotal
        BREAK
      DELETE FROM #StagedOPCInProgress WHERE EventCID = 'EXCP' AND timestamp <= @timestamp
    
      IF @verbose = 1
        PRINT 'Step ' + CAST(@step AS NVARCHAR(9))
      SET @step = @step + 1
    END

    -- Lookup latest message that does not have an earlier unprocessed exception
    SET @timestamp = NULL
    SELECT @timestamp = MAX(timestamp)
    FROM #StagedOPCInProgress AS m
    WHERE m.EventCID = 'MESG'
      AND NOT EXISTS
      (
        SELECT * FROM #StagedOPCInProgress AS e 
        WHERE e.EventCID = 'EXCP' AND e.timestamp < m.timestamp
      )

    IF @timestamp IS NOT NULL -- there is at least one message that is next sequentially
    BEGIN
      IF @verbose = 1
        PRINT 'asisp_processStagedOPCEvents: Inserting StagedMESG records'

      INSERT INTO StagedMESG(parent_cid, parent_id, TIME, SEQ_NO, OBJECT_NAME, NATIVE_KEY, MESSAGE, STATE, REASON_CD, AlertState, Priority, StateID, DISCOVERY_IND)
        SELECT m.job_cid, m.job_id, m.TIME, m.SEQ_NO, m.JOBNAME, '', m.MESSAGE, '', ISNULL(m.EventText,''), m.AlertStateID, m.PriorityID, m.StateID, 'MON'
        FROM #StagedOPCInProgress AS m
        WHERE m.EventCID = 'MESG' AND m.timestamp <= @timestamp
        ORDER BY m.timestamp
      SELECT @err = @@ERROR, @rc = @@ROWCOUNT
      IF @err <> 0
      BEGIN
        ROLLBACK TRANSACTION asisp_processStagedOPCEvents
        RETURN -30000
      END
      IF @verbose = 1
        PRINT 'asisp_processStagedOPCEvents: Inserted ' + CAST(@rc AS NVARCHAR(9)) + ' StagedMESG records'
      SET @nEventsInserted = @nEventsInserted + @rc
      IF @nEventsInserted = @nEventsTotal
        BREAK
      DELETE FROM #StagedOPCInProgress WHERE EventCID = 'MESG' AND timestamp <= @timestamp
    END
  END

  IF @verbose = 1
    PRINT 'asisp_processStagedOPCEvents: Step ' + CAST(@step AS NVARCHAR(9))
  SET @step = @step + 1

EndTran:
  IF @tcount = 0
    COMMIT TRANSACTION asisp_processStagedOPCEvents

  -- We should only get here if the records were successfully processed, so get rid of the
  -- StagedOPCEvent records whose timestamps are less than or equal to the last one we processed.
  DELETE FROM StagedOPCEvent WHERE timestamp <= @max_timestamp

/*  -- Diagnostic code to save events in the asisp_processStagedOPCEvents_Audit table.
  DECLARE @endtime DATETIME
  SET @endtime = CURRENT_TIMESTAMP
  UPDATE asisp_processStagedOPCEvents_Audit SET endtime = @endtime, result = @nEvents WHERE starttime = @starttime
  PRINT CONVERT(NVARCHAR(20), CURRENT_TIMESTAMP, 121) + ': asisp_processStagedOPCEvents:'
    + ' Elapsed time: ' + CAST(DATEDIFF(MS, @starttime, @endtime) AS NVARCHAR) + 'ms'
    + ' #Events: ' + CAST(@nEvents AS NVARCHAR) 
    + ' Events/Sec.: ' + CAST(CAST( (@nEvents * 1000.0 / DATEDIFF(MS, @starttime, @endtime)) AS REAL) AS NVARCHAR)
*/
   
  RETURN @nEvents
GO

-- Truncate the auditing table
TRUNCATE TABLE asisp_processStagedOPCEvents_Audit
GO
