-- File: archive/db/sql/asisp_processstageddb2_sub_downs.sql
-- SID: 1.1
-- Date Modified: 01/11/05 13:55:41
-- 
-- Module Description:
--   creates the asisp_processStagedDB2_SUB_DOWNs stored procedure

Use Object
GO

----------------------------------------------------------------------
PRINT 'Creating procedure asisp_processStagedDB2_SUB_DOWNs'
GO

IF EXISTS ( SELECT * FROM sysobjects WHERE name = 'asisp_processStagedDB2_SUB_DOWNs' AND type = 'P' )
  DROP PROCEDURE asisp_processStagedDB2_SUB_DOWNs
GO

-- $Archive:   //ARCHIVE/PVCSROOT/Archive_NLS/DB/sql/asisp_processStagedDB2_SUB_DOWNs.sqv  $
-- $Revision:   1.1  $
-- $Date:   Sep 06 2001 17:25:56  $
-- $Author:   Forthman  $

-- File: archive/db/sql/asisp_processstageddb2_sub_downs.sql
-- SID: 1.1
-- Date Modified: 01/11/05 13:55:41
-- Process 17\11 events for db2 objects
CREATE PROCEDURE asisp_processStagedDB2_SUB_DOWNs
  @maxrows INT = 0
AS
  -- exit quickly IF there IS no data to process
  IF NOT EXISTS (SELECT * FROM StagedDB2_SUBSYS_DOWN)
  BEGIN
    RAISERROR('asisp_processStagedDB2_SUB_DOWNs: no staged events to process', 0, 1)
    RETURN 0
  END

  RAISERROR('asisp_processStagedDB2_SUB_DOWNs: rowcount=%d', 0, 1, @maxrows)
  SET ROWCOUNT @maxrows

  -- Standard TRANSACTION logic
  DECLARE @tcount INT
  SELECT @tcount = @@TRANCOUNT
  IF @tcount = 0
    BEGIN TRAN asisp_processStagedDB2_SUB_DOWNs
  ELSE
    SAVE TRAN asisp_processStagedDB2_SUB_DOWNs

  -- Get the @timestamp of the last object inserted from StagedDB2_SUBSYS_DOWN
  --   This will be used to limit the items that will be deleted from StagedDB2_SUBSYS_DOWN
  --   Cannot do this directly because a SELECT MAX statement ignores ROWCOUNT
  DECLARE @timestamp BINARY(8)

  CREATE TABLE #timestamp
  (
     timestamp binary(8) PRIMARY KEY
  )

  INSERT INTO #timestamp SELECT timestamp FROM StagedDB2_SUBSYS_DOWN
  SELECT @timestamp = MAX( timestamp ) FROM #timestamp
  
  DROP TABLE #timestamp


  -- cleanup duplicates FROM StagedDB2_SUBSYS_DOWN.  Keep the first message.
  -- create unique list of objects with StagedDB2_SUBSYS_DOWN event within the timestamp (maxrows) limitation
  CREATE TABLE #DB2_SUB_DOWN_UNIQ
  ( 
    obj_cid ClassID,
    obj_id  ObjID,
    timestamp BINARY(8)
  )
  CREATE CLUSTERED INDEX #DB2_SUB_DOWN_UNIQ_all ON #DB2_SUB_DOWN_UNIQ ( obj_cid, obj_id, timestamp )

  INSERT INTO #DB2_SUB_DOWN_UNIQ 
  ( obj_cid, obj_id, timestamp )
  SELECT DISTINCT src_cid, src_id, MIN( timestamp )
  FROM StagedDB2_SUBSYS_DOWN
  WHERE 
        timestamp <= @timestamp
  GROUP BY src_cid, src_id

  -- Remove the ROWCOUNT limitation to allow for large number of deletions and descendents
  SET ROWCOUNT 0

  -- delete duplicate StagedDB2_SUBSYS_DOWN events from StagedDB2_SUBSYS_DOWN table ( this event should only be applied once per obj_cid, obj_id combination)
  DELETE FROM StagedDB2_SUBSYS_DOWN
  FROM StagedDB2_SUBSYS_DOWN AS s1 
  JOIN #DB2_SUB_DOWN_UNIQ AS s2 ON s1.src_cid=s2.obj_cid AND s1.src_id=s2.obj_id AND s1.timestamp<>s2.timestamp
  WHERE
        s1.timestamp <= @timestamp

  -- Goals: Minimize propagation processing:
  -- The subsystem gets a message that uses propagation
  -- The subsystem's descendents get a green ignore message directly inserted ( audit ), no propagation
  -- The subsystem's descendents's current state gets change to match the message ( no state = object's orignal state )
  -- The subsystem's descendents that are in LOBs get a exception that uses propagation
  -- This is done by:
  -- Create SUBSYSTEM_DOWN message for each subsystem listed in #DB2_SUB_DOWN_UNIQ 
  -- Directly insert SUBSYSTEM_DOWN message for all the children of the subsystem
  -- Directly update Current State for all the children of the subsystem to match SUBSYSTEM_DOWN message if _StateID defined
  -- Create SCHV Exception for each child of the subsystem that is in an LOB

  DECLARE @num_events_processed INT

  -- Create SUBSYSTEM_DOWN message for each subsystem listed in #DB2_SUB_DOWN_UNIQ 
  INSERT StagedMESG(parent_cid, parent_id, TIME, SEQ_NO, OBJECT_NAME, NATIVE_KEY, STATE, MESSAGE, 
                    REASON_CD, DISCOVERY_IND)
    SELECT s.src_cid, s.src_id, DATE_TIME, SEQ_NO, ISNULL(s.obj_name,''), NATIVE_KEY, 'DB2_Down', 'SUBSYSTEM_DOWN', 
                    FORMATMESSAGE(59100, s.obj_name, s.src_id ),
                    'MON'
      FROM StagedDB2_SUBSYS_DOWN AS s
      INNER JOIN #DB2_SUB_DOWN_UNIQ AS s_uniq ON s_uniq.obj_cid = s.src_cid AND s_uniq.obj_id = s.src_id
  SELECT @num_events_processed=@@ROWCOUNT
  RAISERROR('asisp_processStagedDB2_SUB_DOWNs: inserted=%d records INTO StagedMESG', 0, 1, @num_events_processed)

  -- Directly insert SUBSYSTEM_DOWN message for all the children of the subsystem
  --   create all descendents of one or more target subsystems
  CREATE TABLE #descendents
  (
    generation      INT,
    orig_cid        ClassID,
    orig_id         ObjID,
    cid             ClassID,
    id              ObjID,
    name            ObjName,
    _CurrentStateID INT
  )
  CREATE NONCLUSTERED INDEX #descendents_generation ON #descendents ( generation )
  CREATE NONCLUSTERED INDEX #descendents_id         ON #descendents ( cid, id, orig_cid, orig_id )
  -- populate #descendents with the unique subsystem ids
  DECLARE @generation SMALLINT, @rc INT
  SELECT @generation = 1
  INSERT #descendents
    SELECT @generation, obj_cid, obj_id, obj_cid, obj_id, '', 0
      FROM #DB2_SUB_DOWN_UNIQ 
  -- populate #descendents with children from physical or AGGC containment
  SELECT @rc = @@ROWCOUNT
  WHILE @rc > 0
  BEGIN
    SELECT @generation = @generation + 1
    INSERT #descendents
    SELECT DISTINCT @generation, orig_cid, orig_id, dst_cid, dst_id, '', 0
    FROM link l, link_type lt, #descendents d
    WHERE
        l.src_cid=cid
        AND src_id=id
        AND deleted=0
        AND l.link_type=lt.link_type
        AND ( lt.flags&2=2 OR lt.link_type = 'AGGC' )
        AND d.generation=@generation-1
    SELECT @rc = @@ROWCOUNT
  END
  -- populate #descendents columns: name, _CurrentStateID 
  DECLARE @sNameTableCur NVARCHAR(50)
  DECLARE @sql           NVARCHAR(500)
  DECLARE d CURSOR FOR
    SELECT DISTINCT cname + '_C' FROM #descendents AS d
    INNER JOIN obj_class AS oc ON d.cid = oc.cid
  OPEN d
  FETCH NEXT FROM d INTO @sNameTableCur
  WHILE @@FETCH_STATUS <> -1
  BEGIN
    SET @sql = '
    UPDATE #descendents 
    SET 
      name            = o.name,
      _CurrentStateID = o._CurrentStateID 
    FROM ' + @sNameTableCur + ' AS o
    INNER JOIN #descendents AS d ON d.cid = o.cid AND d.id = o.id'
    EXEC( @sql )
    FETCH NEXT FROM d INTO @sNameTableCur
  END
  CLOSE d
  DEALLOCATE d

  -- list of descendents for whom it is ok to have a direct message and updated Current State
  CREATE TABLE #updatable
  (
    cid             ClassID,
    id              ObjID,
    rowno           INT IDENTITY( 0, 1)
  )
  CREATE CLUSTERED    INDEX #updatable_ids        ON #updatable ( cid, id, rowno )
  CREATE NONCLUSTERED INDEX #updatable_rowno      ON #updatable ( rowno )
  
  INSERT INTO #updatable (cid, id)
  SELECT cid, id FROM #descendents WHERE cid NOT IN ( 'D2SS', 'D2DD', 'D2DS', 'D2MS', 'D2PM', 'D2SP' )

  -- how many message id's need to be reserved 
  SELECT @rc = @@ROWCOUNT

  -- Directly insert messages for descendents

  DECLARE @base_id ObjID
  EXEC @base_id = alloc_MESG @rc

  DECLARE @min_rowno INT
  SELECT @min_rowno = MIN(rowno) FROM #updatable 

  INSERT INTO Message_A(id,_DATE,_TIME,_OBJECT_NAME,_REASON_CD,_AlertState,_Priority,_StateID,_ParentCID,_ParentID,ctime,
                        _MESSAGE, _NATIVE_KEY, _OBJ_TYPE_ID, _SEQ_NO, _STATE )
    SELECT
      @base_id+u.rowno-@min_rowno         AS id,
      s.DATE_TIME                         AS _DATE,
      s.DATE_TIME                         AS _TIME,
      d.name                              AS _OBJECT_NAME,
      FORMATMESSAGE(59100, p.name, p.id ) AS _REASON_CD,
      1                                   AS _AlertState, -- Green
      5                                   AS _Priority,   -- Ignore, these are audit messages
      ISNULL( m._StateID, 0 )             AS _StateID,
      d.cid                               AS _ParentCID,
      d.id                                AS _ParentID,
      CURRENT_TIMESTAMP                   AS ctime,
      'SUBSYSTEM_DOWN'                    AS _MESSAGE,
      ''                                  AS _NATIVE_KEY,
      d.cid                               AS _OBJ_TYPE_ID,
      s.SEQ_NO                            AS _SEQ_NO,
      ''                                  AS _STATE
    FROM #descendents AS d
    INNER JOIN      #updatable            AS u ON u.cid        = d.cid      AND u.id      = d.id
    LEFT OUTER JOIN MessageDescription_V  AS m ON m._ParentCID = d.cid      AND m._MsgID  = 'SUBSYSTEM_DOWN' AND m.deleted = 0
    INNER JOIN      #descendents          AS p ON p.cid        = d.orig_cid AND p.id      = d.orig_id
    INNER JOIN      StagedDB2_SUBSYS_DOWN AS s ON s.src_cid    = d.orig_cid AND s.src_id  = d.orig_id
    -- HACK: need to handle the m._ParentID = d.id join, to pick up information specific to a single object + event combination

  SELECT @num_events_processed=@@ROWCOUNT
  RAISERROR('asisp_processStagedDB2_SUB_DOWNs: inserted=%d records INTO Message_A', 0, 1, @num_events_processed)

  INSERT INTO Message_C(id,name,ctime)
    SELECT 
      id,
      'SUBSYSTEM_DOWN',
      ctime
    FROM
      Message_A
    WHERE
      id BETWEEN @base_id AND @base_id + @rc

  SELECT @num_events_processed=@@ROWCOUNT
  RAISERROR('asisp_processStagedDB2_SUB_DOWNs: inserted=%d records INTO Message_C', 0, 1, @num_events_processed)

  -- Directly update Current State for all the children of the subsystem to match SUBSYSTEM_DOWN message if _StateID defined
  DECLARE d CURSOR FOR
    SELECT DISTINCT cname + '_C' FROM #updatable AS d
    INNER JOIN obj_class AS oc ON d.cid = oc.cid
  OPEN d
  FETCH NEXT FROM d INTO @sNameTableCur
  WHILE @@FETCH_STATUS <> -1
  BEGIN
    SET @sql = '
    UPDATE ' + @sNameTableCur + '
    SET 
      _CurrentStateID = m._StateID 
    FROM Message_A AS m
    INNER JOIN ' + @sNameTableCur + ' AS o ON o.cid = m._ParentCID AND o.id = m._ParentID AND o._CurrentStateID <> m._StateID
    WHERE
          m.id BETWEEN ' + CONVERT(NVARCHAR(20), @base_id) + ' AND ' + CONVERT( NVARCHAR(20), @base_id + @rc ) + '
      AND m._StateID <> 0
    '
    EXEC( @sql )
    SELECT @num_events_processed=@@ROWCOUNT
    RAISERROR('asisp_processStagedDB2_SUB_DOWNs: updated=%d records in %s', 0, 1, @num_events_processed, @sNameTableCur)
    FETCH NEXT FROM d INTO @sNameTableCur
  END
  CLOSE d
  DEALLOCATE d


  -- Create SCHV Exception for each child of the subsystem that is in an LOB
  INSERT INTO StagedEXCP(parent_cid, parent_id, TIME, SEQ_NO, OBJECT_NAME, NATIVE_KEY, EXCP_OBJ_TYPE, EXCP_OBJ_NAME, EXCP_NAME, 
                         EXCP_CD, nEXCH)
    SELECT DISTINCT d.cid, d.id, p.DATE_TIME, p.SEQ_NO, ISNULL(d.name,''), '', '', '', e.Exception, 
        FORMATMESSAGE(59101, p.obj_name, p.src_id ), 1
    FROM #descendents AS d
    INNER JOIN #DB2_SUB_DOWN_UNIQ    AS u ON u.obj_cid = d.orig_cid AND u.obj_id = d.orig_id 
    INNER JOIN StagedDB2_SUBSYS_DOWN AS p ON p.src_cid = u.obj_cid  AND p.src_id = u.obj_id  AND p.timestamp = u.timestamp
    INNER JOIN link                  AS l ON l.dst_cid = d.cid      AND l.dst_id = d.id
    INNER JOIN ExceptionType_T       AS e ON e.cid     = d.cid
    WHERE  
          l.link_type  = 'LOBL'
      AND l.deleted    = 0
      AND d.generation > 1
      AND e.Exception  = 'SCHV'
  SELECT @num_events_processed=@@ROWCOUNT
  RAISERROR('asisp_processStagedDB2_SUB_DOWNs: inserted=%d records INTO StagedEXCP', 0, 1, @num_events_processed)


  -- cleanup the staging TABLE
  DROP TABLE #updatable
  DROP TABLE #descendents
  DROP TABLE #DB2_SUB_DOWN_UNIQ

  DELETE FROM StagedDB2_SUBSYS_DOWN WHERE timestamp <= @timestamp
  SELECT @num_events_processed=@@ROWCOUNT
  RAISERROR('asisp_processStagedDB2_SUB_DOWNs: deleted=%d records FROM StagedDB2_SUBSYS_DOWN', 0, 1, @num_events_processed)

  -- Finish up TRANSACTION
  IF @tcount = 0
    COMMIT TRAN asisp_processStagedDB2_SUB_DOWNs

  RETURN @num_events_processed
GO
