-- archive/db/sql/detachdatabases.sql
-- SID: 1.9
-- Date Modified: 01/09/28 15:16:20
--
-- detachdatabases.sql
--
-- Module Description
--   Process for detaching TBSM databases so that they can be transported to a different
--   server.

USE master
GO
SET NOCOUNT ON
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'asisp_rename_sqlserver' AND type = 'P')
  DROP PROCEDURE asisp_rename_sqlserver
GO

CREATE PROCEDURE asisp_rename_sqlserver
  @db_host        NVARCHAR(255),
  @pa_host        NVARCHAR(255)=NULL,
  @websvr_host    NVARCHAR(255)=NULL
AS
  DECLARE @local_host NVARCHAR(255), @sql NVARCHAR(500)

  IF @db_host IS NULL
    SELECT @db_host=LOWER(srvname) FROM master..sysservers WHERE srvstatus=65
  IF @pa_host IS NULL
    SELECT @pa_host=@db_host
  IF @websvr_host IS NULL
    SELECT @websvr_host=@db_host

  SELECT @local_host=srvname FROM master..sysservers WHERE srvstatus=65
  
  RAISERROR('asisp_rename_sqlserver: renaming SQLServer internal hostname to host=[%s]', 0, 1, @db_host)

  RAISERROR('asisp_rename_sqlserver: dropping server definition %s', 0, 1, @local_host)
  EXEC sp_dropserver @local_host
  IF @@ERROR <> 0
  BEGIN
    RAISERROR('asisp_rename_sqlserver: Could NOT DROP old server definition %s', 3, 1, @local_host)
    RETURN -30000
  END

  RAISERROR('asisp_rename_sqlserver: adding server definition %s', 0, 1, @db_host)
  EXEC sp_addserver @db_host, LOCAL
  IF @@ERROR <> 0
  BEGIN
    RAISERROR('asisp_rename_sqlserver: Could NOT ADD new server definition %s', 3, 1, @db_host)
    RETURN -30000
  END

  DECLARE @tcount INT
  SET @tcount = @@TRANCOUNT
  IF @tcount = 0
    BEGIN TRAN asisp_rename_sqlserver
  ELSE
    SAVE TRAN asisp_rename_sqlserver

  IF EXISTS (SELECT * FROM sysdatabases WHERE name = 'Object')
  BEGIN
    RAISERROR('asisp_rename_sqlserver: setting propagation agent host=[%s]', 0, 1, @pa_host)
    EXEC("
    UPDATE Object..PropagationAgent_A SET _Host = '" + @pa_host + "'
    ")
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('asisp_rename_sqlserver: Could NOT UPDATE Object..PropagationAgent_A, aborting', 3, 1)
      ROLLBACK TRAN asisp_rename_sqlserver
      RETURN -30000
    END

    RAISERROR('asisp_rename_sqlserver: changing MenuItem URLs to USE host=[%s]', 0, 1, @websvr_host)
    EXEC("
    UPDATE Object..MenuItem SET control_id='http://' + '" + @websvr_host + "' + SUBSTRING(substring(control_id,8,255),charindex('/',SUBSTRING(control_id,8,255)),255)
      FROM Object..MenuItem WHERE LOWER(control_id) like 'http://%'
    ")
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('asisp_rename_sqlserver: Could NOT UPDATE Object..MenuItem, aborting', 3, 1)
      ROLLBACK TRAN asisp_rename_sqlserver
      RETURN -30000
    END

    RAISERROR('asisp_rename_sqlserver: truncating application_session and workstation_session tables', 0, 1)
    EXEC("
    TRUNCATE TABLE Object..application_session
    TRUNCATE TABLE Object..workstation_session
    ")
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('asisp_rename_sqlserver: Could NOT TRUNCATE Object database tables application_session AND workstation_session, aborting', 3, 1)
      ROLLBACK TRAN asisp_rename_sqlserver
      RETURN -30000
    END


    -- Table to hold replacement host records for Installation Management / Version Utility 
    CREATE TABLE #replacementHost
    (
      host       NVARCHAR(64),
      pseudoHost NVARCHAR(64)
    )
    
    -- Insert as many records as necessary to translate real host names to replacement host names
    -- All host names entered here should be upper case
    INSERT #replacementHost ( host, pseudoHost ) VALUES ( 'TBSM01', 'S1' )
    INSERT #replacementHost ( host, pseudoHost ) VALUES ( 'TBSM02', 'S2' )
    INSERT #replacementHost ( host, pseudoHost ) VALUES ( 'TBSM03', 'S3' )
    INSERT #replacementHost ( host, pseudoHost ) VALUES ( 'TBSM04', 'S4' )
    INSERT #replacementHost ( host, pseudoHost ) VALUES ( 'TBSM05', 'S5' )
    
    -- Prepare Installation Management / Version Utility data to be detached
    UPDATE Object..imObjectList
      SET host = r.pseudoHost
      FROM Object..imObjectList AS o
      INNER JOIN #replacementHost AS r ON o.host = r.host
    
    PRINT 'UPDATE Object..imObjectList'
    
    UPDATE Object..imExclusionPattern
      SET host = r.pseudoHost
      FROM Object..imExclusionPattern AS o
      INNER JOIN #replacementHost AS r ON o.host = r.host
    
    PRINT 'UPDATE Object..imExclusionPattern'
    
    UPDATE Object..imObjectSeed
      SET host = r.pseudoHost
      FROM Object..imObjectSeed AS o
      INNER JOIN #replacementHost AS r ON o.host = r.host
    
    PRINT 'UPDATE Object..imObjectSeed'
    
    -- Cleanup
    DROP TABLE #replacementHost
    
    
  END

  IF EXISTS (SELECT * FROM sysdatabases WHERE name = 'ASIRuleSvc')
  BEGIN
    RAISERROR('asisp_rename_sqlserver: truncating ASIRuleSvc..ObjectSync', 0, 1)
    EXEC("
    TRUNCATE TABLE ASIRuleSvc..ObjectSync
    ")
    IF @@ERROR <> 0
    BEGIN
      RAISERROR('asisp_rename_sqlserver: Could NOT TRUNCATE ASIRuleSvc..ObjectSync', 3, 1)
      ROLLBACK TRAN asisp_rename_sqlserver
      RETURN -30000
    END
  END

  IF EXISTS (SELECT * FROM sysdatabases WHERE name = 'ObjectQueues')
  BEGIN
    -- Reset event processing tables IN the ObjectQueues database
    EXEC("
    DECLARE eventc INSENSITIVE CURSOR FOR 
      SELECT name FROM ObjectQueues..sysobjects 
      WHERE type = 'U' AND (name like 'Pending%' OR name like 'Processed%' OR name like 'LastProcessed%')
      ORDER BY name desc
    ")
    OPEN eventc
    DECLARE @table sysname
    FETCH NEXT FROM eventc INTO @table
    WHILE @@FETCH_STATUS <> -1
    BEGIN
      IF @table like 'LastProcessed%'
        SELECT @sql = 'UPDATE ObjectQueues..' + @table + ' SET eventno=0'
      ELSE
        SELECT @sql = 'TRUNCATE TABLE ObjectQueues..' + @table
      PRINT @sql
      EXEC(@sql)
      IF @@ERROR <> 0
      BEGIN
        RAISERROR('asisp_rename_sqlserver: Operation failed, aborting', 3, 1)
        ROLLBACK TRAN asisp_rename_sqlserver
        DEALLOCATE eventc
        RETURN -30000
      END
      FETCH NEXT FROM eventc INTO @table
    END
    DEALLOCATE eventc
  END

  IF @tcount = 0
    COMMIT TRAN asisp_rename_sqlserver
GO

EXEC sp_defaultdb 'BUILTIN\Administrators', 'master'
EXEC sp_defaultdb 'sa', 'master'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE type = 'X' AND name = 'xp_rule_stoppulsing')
  EXEC xp_rule_stoppulsing
GO

EXEC asisp_rename_sqlserver 'XXXX', 'XXXX', 'XXXX'
GO


checkpoint
GO

IF EXISTS (SELECT * FROM master..sysdatabases WHERE name='Meta')
  EXEC sp_detach_db Meta
IF EXISTS (SELECT * FROM master..sysdatabases WHERE name='Object')
  EXEC sp_detach_db Object
IF EXISTS (SELECT * FROM master..sysdatabases WHERE name='ObjectEvents')
  EXEC sp_detach_db ObjectEvents
IF EXISTS (SELECT * FROM master..sysdatabases WHERE name='ObjectQueues')
  EXEC sp_detach_db ObjectQueues
IF EXISTS (SELECT * FROM master..sysdatabases WHERE name='ASIRuleSvc')
  EXEC sp_detach_db ASIRuleSvc
IF EXISTS (SELECT * FROM master..sysdatabases WHERE name='WebServer')
  EXEC sp_detach_db WebServer
IF EXISTS (SELECT * FROM master..sysdatabases WHERE name='msdb')
  EXEC sp_detach_db msdb
IF EXISTS (SELECT * FROM master..sysdatabases WHERE name='RODM')
  EXEC sp_detach_db RODM
IF EXISTS (SELECT * FROM master..sysdatabases WHERE name='RODMLoad')
  EXEC sp_detach_db RODMLoad
IF EXISTS (SELECT * FROM master..sysdatabases WHERE name='OPCLoad')
  EXEC sp_detach_db OPCLoad
IF EXISTS (SELECT * FROM master..sysdatabases WHERE name='EventHistory')
  EXEC sp_detach_db EventHistory
GO

-- obsolete database
IF EXISTS (SELECT * FROM master..sysdatabases WHERE name='ASIWebServer')
  EXEC sp_detach_db ASIWebServer
GO
IF EXISTS (SELECT * FROM master..sysdevices WHERE name='ASIWebServer')
  EXEC sp_dropdevice 'ASIWebServer', 'DELFILE'
GO

PRINT 'Done.'
GO
