-- File: archive\DB\sql\tu_batch_c.sql
-- SID: 1.2
-- Date Modified: 01/10/24 12:29:51
--
-- Module Description
--   Defines the tU_BATCH_C trigger

USE Object
GO


---------------------------------------------------------------------------------------------------
-- tU_BATCH_C
PRINT 'Creating trigger tU_BATCH_C '
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'tU_BATCH_C ' AND type = 'TR')
  DROP TRIGGER tU_BATCH_C
GO

-- File: archive\DB\sql\tu_batch_c.sql
-- SID: 1.2
-- Date Modified: 01/10/24 12:29:51
-- Custom update trigger for BATCH_C table. Does the following:
--
-- 1. KeyStatus changes cause the Priority of the updated rows to be changed as follows:
--    New KeyStatus  ->  New Priority      
--         0             5 (Ignore)
--         1             6 (InheritFromEvent)
--
-- 2. All jobs in a Completed or Unknown state whose SchedStartTime is changing to a non-null value
--    are reset to the Pending state. If the jobs are in a Completed state then we only perform the
--    update if current StartTime of the job is closer to the new SchedStartTime than the old. 
--    Audit messages are also generated on the affected jobs.
CREATE TRIGGER tU_BATCH_C ON BATCH_C FOR UPDATE
AS
  DECLARE @rc INT

  IF UPDATE(_KeyStatus)
  BEGIN
    -- Jobs that are changing from key to nonkey which have InheritFromEvent priority
    -- are changed to Ignore priority
    UPDATE b SET b._PriorityID = 5 FROM BATCH_C AS b
      JOIN inserted AS i ON b.id = i.id AND i._KeyStatus = 0 
      JOIN deleted AS d ON i.id = d.id AND d._PriorityID = 6
    -- Jobs that are changing from nonkey to key which have Ignore priority
    -- are change to InheritFromEvent priority
    UPDATE b SET b._PriorityID = 6 FROM BATCH_C AS b
      JOIN inserted AS i ON b.id = i.id AND i._KeyStatus = 1
      JOIN deleted AS d ON b.id = d.id AND d._PriorityID = 5
  END
  IF UPDATE(_SchedStartTime)
  BEGIN
    -- All jobs in a Completed or Unknown state whose scheduled start times are changing
    -- to a non-null value are reset to the Pending state. If the jobs are in a Completed
    -- state then we only perform the update if current StartTime of the job is closer
    -- to the new SchedStartTime than the old.

    DECLARE @ts TIMESTAMP
    SET @ts = @@DBTS

    INSERT INTO RowAllocator(cid, id, timestamp)
      SELECT
        'BATC' AS _ParentCID,
        b.id AS _ParentID,
        @ts
      FROM
        BATCH_C AS b
        JOIN deleted AS d ON d.id = b.id
        JOIN inserted AS i ON i.id = d.id
      WHERE
        i._CurrentStateID IN ( 0 /* Unknown */, 83 /* Completed */ )
        AND i._SchedStartTime IS NOT NULL
        AND
        (
          d._SchedStartTime IS NULL OR 
          d._StartTime IS NULL OR
          (
            d._SchedStartTime <> i._SchedStartTime AND 
            ABS(DATEDIFF(SECOND, d._StartTime, d._SchedStartTime)) < 
            ABS(DATEDIFF(SECOND, d._StartTime, i._SchedStartTime))
          )
        )
    SET @rc = @@ROWCOUNT

    IF @rc > 0
    BEGIN
      UPDATE b SET
        _StartTime = NULL,
        _StopTime = NULL,
        _CurrentStateID = 376 /* Pending */
      FROM
        BATCH_C AS b
        JOIN RowAllocator AS r ON r.cid = 'BATC' AND r.id = b.id
      WHERE
        r.timestamp = @ts

      -- We also create audit messages on the affected jobs
      DECLARE @base_id ObjID
      EXEC @base_id = alloc_MESG @rc

      DECLARE @min_rowno INT
      SELECT @min_rowno = MIN(rowno) FROM RowAllocator WHERE timestamp = @ts

      INSERT INTO Message_A(id,_DATE,_TIME,_OBJECT_NAME,_REASON_CD,_AlertState,_Priority,_StateID,_ParentCID,_ParentID,ctime)
        SELECT 
          @base_id+r.rowno-@min_rowno AS id,
          CURRENT_TIMESTAMP AS _DATE,
          CURRENT_TIMESTAMP AS _TIME,
          i.name AS _OBJECT_NAME,
          FORMATMESSAGE(58001, i.name, CONVERT(NVARCHAR(16),d._SchedStartTime,120), CONVERT(NVARCHAR(16),i._SchedStartTime,120)) AS _REASON_CD,
          0 AS _AlertState,
          5 AS _Priority,
          376 AS _StateID,
          'BATC' AS  _ParentCID,
          i.id AS _ParentID,
          CURRENT_TIMESTAMP AS ctime
        FROM
          deleted AS d
          JOIN inserted AS i ON i.id = d.id
          JOIN RowAllocator AS r ON r.cid = 'BATC' AND r.id = i.id
        WHERE
          r.timestamp = @ts
      INSERT INTO Message_C(id,name,ctime)
        SELECT 
          id,
          '',
          ctime
        FROM
          Message_A
        WHERE
          id BETWEEN @base_id AND @base_id + @rc
      DELETE RowAllocator WHERE timestamp = @ts
    END
  END
GO
